/***************************************************************************
                    interface_net.cpp  -  client connexion for GUI
                             -------------------
    begin                : Mon Aug 14 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
  #include <config.h>
#endif

#include "interface_net.h"

// =======================================================
CInterfaceNet::CInterfaceNet(BOOL bUseSSL): CNetClient(bUseSSL)
{
  // multi-thread
  //pthread_mutex_init(&m_mutexNetwork, &m_mutexNetworkAttr);
}

// =======================================================
CInterfaceNet::~CInterfaceNet()
{
  lock(__LINE__);
  m_msg.setMessage(MG1_QUIT);
  unlock();

  // multi-thread
  //pthread_mutex_destroy(&m_mutexNetwork);
}

// =======================================================
void CInterfaceNet::lock(int n) // Multi-thread -> avoid problems
{
  showDebug(5, "[#%d] required mutex\n", n);
  //pthread_mutex_lock(&m_mutexNetwork);
  showDebug(5, "got mutex\n");
}

// =======================================================
void CInterfaceNet::unlock() // Multi-thread -> avoid problems
{
  showDebug(5, "release mutex\n");
  //pthread_mutex_unlock(&m_mutexNetwork);
}

// =======================================================
void CInterfaceNet::Connect(char *szServerName, WORD wServerPort)
{
  lock(__LINE__);
  try { CNetClient::Connect(szServerName, wServerPort); }
  catch ( CExceptions * excep )
    {
      unlock();
      throw excep;
    }

  unlock();
  showDebug(9, "return from SendPass: success\n");
}

void CInterfaceNet::set_message(DWORD number)
{
  m_msg.setMessage(number);
  SendMsg(&m_msg);
}
void CInterfaceNet::set_message(DWORD number, DWORD size)
{
  m_msg.setMessage(number, size);
  SendMsg(&m_msg);
}

// =======================================================
void CInterfaceNet::set_options(COptions options)
{
  m_msg.setMessage(MG1_SET_OPTIONS);
  showDebug(3, "ENCRPYT?: %d\n",options.wEncryptAlgo); 
  SendMsg(&m_msg);
  Send(&options, sizeof(COptions));
}

// =======================================================
void CInterfaceNet::set_param(WORD wOperation, char *szDevice,
   char *szImageFile, char *szImageFile2)
{
  WORD wDevice, wImageFile, wImageFile2;
  wDevice = strlen(szDevice)+1;           // +1 needed because of endding \0
  wImageFile = strlen(szImageFile)+1;     // +1 needed because of endding \0
  wImageFile2 = strlen(szImageFile2)+1;   // +1 needed because of endding \0

  m_msg.setMessage(MG1_SET_PARAM, wOperation);
  SendMsg(&m_msg);
  switch(wOperation)
    {
    case OPERATION_SAVE: // szDevice + szImageFile
    case OPERATION_RESTORE:
      Send(&wDevice, sizeof(wDevice));
      Send(szDevice, wDevice);
      Send(&wImageFile, sizeof(wImageFile));
      Send(szImageFile, wImageFile);
      break;
      
    case OPERATION_RESTMBR: // szImageFile
    case OPERATION_IMGINFO:
      Send(&wImageFile, sizeof(wImageFile));
      Send(szImageFile, wImageFile);
      break;

    case OPERATION_IMGCONV: // szImageFile + szImageFile2
      Send(&wImageFile, sizeof(wImageFile));
      Send(szImageFile, wImageFile);
      Send(&wImageFile2, sizeof(wImageFile2));
      Send(szImageFile2, wImageFile2);
      break;
    }
}

// =======================================================
void CInterfaceNet::set_description(char *szPartDescription)
{
  WORD wDescLength;
  
  wDescLength = strlen(szPartDescription)+1;
  m_msg.setMessage(MG1_SET_DESCRIPTION, wDescLength);
  SendMsg(&m_msg);
  Send(szPartDescription, wDescLength);
}
