#include "formDialogPartedCreate.h"
#include "pipushbutton.h"
#include "common.h"

#include <qslider.h>
#include <qlabel.h>
#include <qpushbutton.h>
#include <qlayout.h>
#include <qtooltip.h>
#include <qwhatsthis.h>
#include <qcombobox.h>
#include <qlineedit.h>

#define POSITION_BEGIN       0
#define POSITION_END         1

#define MAX_PARTSIZE_SECTORS 2048 // 1 MB = 512*2048
#define SECTOR_SIZE          512LL

#define PARTSTRING_PRIMARY   "Primary"
#define PARTSTRING_EXTENDED  "Extended"
#define PARTSTRING_LOGICAL   "Logical (inside extended)"

//------------------------------------------------
CPartedCreateDialog::CPartedCreateDialog(char *szDevice, QStrList *strlFs, DWORD dwMinStart, DWORD dwMaxEnd, BOOL bCanMakeExtended, 
					 BOOL bCanMakePrimary, BOOL bCanMakeLogical, QWidget* parent,  const char* name, 
					 bool modal, WFlags fl) : QDialog(parent, name, modal, fl)
{
   // init
   m_strlFs = strlFs;
   m_strDevice.sprintf("%s", szDevice);
   m_dwMinStart = m_dwStart = dwMinStart;
   m_dwMaxEnd = m_dwEnd = dwMaxEnd;
   m_dwPartType = PARTTYPE_PRIMARY;

   initDialog(bCanMakeExtended, bCanMakePrimary, bCanMakeLogical);
   showInfos();

   connect(m_comboPartType, SIGNAL(activated(int)), this, SLOT(slotChanged()));
   connect(m_comboPartPos, SIGNAL(activated(int)), this, SLOT(slotChanged()));
   connect(m_slider, SIGNAL(valueChanged(int)), this, SLOT(slotChanged()));
   connect(m_btnOk, SIGNAL(clicked()), this, SLOT(slotOk()));
   connect(m_editLabel, SIGNAL(textChanged(const QString &)), this, SLOT(slotChanged()));
   connect(m_btnCancel, SIGNAL(clicked()), this, SLOT(reject()));

   slotChanged();
}

//------------------------------------------------
void CPartedCreateDialog::initDialog(BOOL bCanMakeExtended, BOOL bCanMakePrimary, BOOL bCanMakeLogical)
{
   DWORD dwMaxSize;
   QString strTemp;

   setMinimumSize(300, 200);
   resize(500, 300);
   setCaption( trUtf8( "Create a new partition" ) );
   dwMaxSize = m_dwMaxEnd - m_dwMinStart;

   m_lay = new QGridLayout (this, 11/*row*/, 4/*col*/, 0, 10);
   m_lay -> setMargin(15);

   m_labelDevice1 = new QLabel(this, "m_labelDevice1");
   m_labelDevice1 -> setText("Device:");
   m_lay -> addMultiCellWidget (m_labelDevice1, 0, 0, 0, 0);

   m_labelDevice2 = new QLabel(this, "m_labelDevice2");
   m_labelDevice2 -> setText(m_strDevice);
   m_lay -> addMultiCellWidget (m_labelDevice2, 0, 0, 1, 3);

   m_labelType = new QLabel(this, "m_labelType");
   m_labelType -> setText("Partition type:");
   m_lay -> addMultiCellWidget (m_labelType, 1, 1, 0, 0);

   m_comboPartType = new QComboBox(this, "m_comboPartType");
   if (bCanMakePrimary)
      m_comboPartType -> insertItem(PARTSTRING_PRIMARY);
   if (bCanMakeLogical)
      m_comboPartType -> insertItem(PARTSTRING_LOGICAL);
   if (bCanMakeExtended)
      m_comboPartType -> insertItem(PARTSTRING_EXTENDED);
   m_lay -> addMultiCellWidget (m_comboPartType, 1, 1, 1, 3);

   m_labelPos = new QLabel(this, "m_labelPos");
   m_labelPos -> setText("Partition position:");
   m_lay -> addMultiCellWidget (m_labelPos, 2, 2, 0, 0);

   m_comboPartPos = new QComboBox(this, "m_comboPartPos");
   m_comboPartPos -> insertItem("Begin of free space");
   m_comboPartPos -> insertItem("End of free space");
   m_lay -> addMultiCellWidget (m_comboPartPos, 2, 2, 1, 3);

   m_labelFs = new QLabel(this, "m_labelFs");
   m_labelFs -> setText("File system:");
   m_lay -> addMultiCellWidget (m_labelFs, 3, 3, 0, 0);

   m_comboFs = new QComboBox(this, "m_comboFs");
   for (strTemp = m_strlFs->first(); strTemp; strTemp = m_strlFs->next() )
      m_comboFs -> insertItem(strTemp);
   m_lay -> addMultiCellWidget (m_comboFs, 3, 3, 1, 3);

   m_labelLabel = new QLabel(this, "m_labelLabel");
   m_labelLabel -> setText("Label:");
   m_lay -> addMultiCellWidget (m_labelLabel, 4, 4, 0, 0);

   m_editLabel = new QLineEdit(this, "m_editLabel");
   m_editLabel -> setMaxLength (11);
   m_lay -> addMultiCellWidget (m_editLabel, 4, 4, 1, 3);

   m_labelTitle1 = new QLabel(this, "m_labelTitle1");
   m_labelTitle1 -> setText("sectors");
   m_lay -> addWidget (m_labelTitle1, 5, 1);

   m_labelTitle2 = new QLabel(this, "m_labelTitle2");
   m_labelTitle2 -> setText("megabytes");
   m_lay -> addWidget (m_labelTitle2, 5, 2);

   m_labelStart1 = new QLabel(this, "m_labelStart1");
   m_labelStart1 -> setText("Start");
   m_lay -> addWidget (m_labelStart1, 6, 0);

   m_labelStart2 = new QLabel(this, "m_labelStart2");
   m_lay -> addWidget (m_labelStart2, 6, 1);

   m_labelStart3 = new QLabel(this, "m_labelStart3");
   m_lay -> addWidget (m_labelStart3, 6, 2);

   m_labelEnd1 = new QLabel(this, "m_labelEnd1");
   m_labelEnd1 -> setText("End");
   m_lay -> addWidget (m_labelEnd1, 7, 0);

   m_labelEnd2 = new QLabel(this, "m_labelEnd2");
   m_lay -> addWidget (m_labelEnd2, 7, 1);

   m_labelEnd3 = new QLabel(this, "m_labelEnd3");
   m_lay -> addWidget (m_labelEnd3, 7, 2);

   m_labelSize1 = new QLabel(this, "m_labelSize1");
   m_labelSize1 -> setText("Size");
   m_lay -> addWidget (m_labelSize1, 8, 0);
   
   m_labelSize2 = new QLabel(this, "m_labelSize2");
   m_lay -> addWidget (m_labelSize2, 8, 1);

   m_labelSize3 = new QLabel(this, "m_labelSize3");
   m_lay -> addWidget (m_labelSize3, 8, 2);

   m_labelSize4 = new QLabel(this, "m_labelSize4");
   m_labelSize4 -> setText("Partition size");
   m_lay -> addWidget (m_labelSize4, 9, 0);
   
   m_slider = new QSlider(Qt::Horizontal, this,"m_slider");
   m_lay -> addMultiCellWidget (m_slider, 9, 9, 1, 3);
   m_slider -> setMaxValue (dwMaxSize);
   m_slider -> setMinValue (MAX_PARTSIZE_SECTORS);
   m_slider -> setValue (dwMaxSize);

   m_btnOk = new COkButton(this,"m_btnOk");
   m_lay -> addMultiCellWidget (m_btnOk, 10, 10, 0, 1);
  
   m_btnCancel = new CCancelButton(this,"m_btnCancel");
   m_lay -> addMultiCellWidget (m_btnCancel, 10, 10, 2, 3);
}

//------------------------------------------------
CPartedCreateDialog::~CPartedCreateDialog()
{
    // no need to delete child widgets, Qt does it all for us
}

//------------------------------------------------
void CPartedCreateDialog::slotOk()
{
  // Check the Search text is not empty
  /*if (not good)
    {
    error
    return;
    }*/
  
  // Copy text into variables
  //m_strSearch = m_editSearch -> text();
  //m_strReplace = m_editReplace -> text();
  
  accept();
}

//------------------------------------------------
void CPartedCreateDialog::slotChanged()
{
   DWORD dwSliderVal;
   DWORD dwPosition;
   QString strPrimary, strExtended, strLogical;

   strPrimary.sprintf("%s", PARTSTRING_PRIMARY);
   strExtended.sprintf("%s", PARTSTRING_EXTENDED);
   strLogical.sprintf("%s", PARTSTRING_LOGICAL);

   dwSliderVal = m_slider->value();
   dwPosition = m_comboPartPos->currentItem();
   //m_dwPartType = m_comboPartType->currentItem();

   if (m_comboPartType->currentText() == strPrimary)
      m_dwPartType = PARTTYPE_PRIMARY;
   else if (m_comboPartType->currentText() == strExtended)
      m_dwPartType = PARTTYPE_EXTENDED;
   else if (m_comboPartType->currentText() == strLogical)
      m_dwPartType = PARTTYPE_LOGICAL;
   else // unknown
   {
      m_dwPartType = PARTTYPE_PRIMARY; // default
      pi_printf ("Error\n");
      return ;
   }

   // gray comboFs is need
   m_comboFs->setEnabled(m_dwPartType != PARTTYPE_EXTENDED);
   m_editLabel->setEnabled(m_dwPartType != PARTTYPE_EXTENDED);
 
   if (dwPosition == POSITION_BEGIN) // begin of free space
   {
      m_dwStart = m_dwMinStart;
      m_dwEnd = m_dwMinStart + dwSliderVal;
   }
   else // end of free space
   {
      m_dwEnd = m_dwMaxEnd;
      m_dwStart = m_dwMaxEnd - dwSliderVal;
   }

   m_strFs = m_comboFs->currentText();
   m_strLabel = m_editLabel->text();

   showInfos();
}

//------------------------------------------------
int CPartedCreateDialog::getResults(DWORD *dwSectStart, DWORD *dwSectEnd, DWORD *dwPartType, char *szFs, int nMaxFsLen, char *szLabel, int nMaxLabelLen) 
{
   *dwSectStart = m_dwStart; 
   *dwSectEnd = m_dwEnd; 
   *dwPartType = m_dwPartType;
   snprintf(szFs, nMaxFsLen, "%s", m_strFs.latin1());
   snprintf(szLabel, nMaxLabelLen, "%s", m_strLabel.latin1());

   return 0;
}

//------------------------------------------------
void CPartedCreateDialog::showInfos()
{
   DWORD dwSize;
   QString strTemp;
   char szTemp[4096];

   dwSize = m_dwEnd - m_dwStart;

   strTemp.sprintf("%lu", m_dwStart);
   m_labelStart2 -> setText(strTemp);

   strTemp.sprintf("%s", formatSize((QWORD)m_dwStart*SECTOR_SIZE, szTemp));
   m_labelStart3 -> setText(strTemp);

   strTemp.sprintf("%lu", m_dwEnd);
   m_labelEnd2 -> setText(strTemp);

   strTemp.sprintf("%s", formatSize((QWORD)m_dwEnd*SECTOR_SIZE, szTemp));
   m_labelEnd3 -> setText(strTemp);

   strTemp.sprintf("%lu", dwSize);
   m_labelSize2 -> setText(strTemp);

   strTemp.sprintf("%s", formatSize((QWORD)dwSize*SECTOR_SIZE, szTemp));
   m_labelSize3 -> setText(strTemp);
}

#include "formDialogPartedCreate.moc"
