#include "formDialogPiSave.h"
#include "pipushbutton.h"
#include "common.h"

#include <qslider.h>
#include <qlabel.h>
#include <qpushbutton.h>
#include <qlayout.h>
#include <qtooltip.h>
#include <qwhatsthis.h>
#include <qcombobox.h>
#include <qbuttongroup.h>
#include <qradiobutton.h>
#include <qcheckbox.h> 
#include <qhbox.h>
#include <qspinbox.h>
#include <qfiledialog.h>
#include <qlineedit.h>
#include <qtextedit.h>

//------------------------------------------------
CPiSaveDialog::CPiSaveDialog(char *szDevice, COptions *options, QWidget* parent,  const char* name, bool modal, WFlags fl ) : QDialog(parent, name, modal, fl)
{
   // init
   m_options = options;
   m_strDevice.sprintf("%s", szDevice);
   initDialog();

   connect(m_btnSelectFile, SIGNAL(clicked()), this, SLOT(slotSelect()));
   connect(m_btnOk, SIGNAL(clicked()), this, SLOT(slotOk()));
   connect(m_btnCancel, SIGNAL(clicked()), this, SLOT(reject()));
   connect(m_radioSplit1, SIGNAL(clicked()), this, SLOT(slotSelectSplit1()));
   connect(m_radioSplit2, SIGNAL(clicked()), this, SLOT(slotSelectSplit2()));
}

//------------------------------------------------
CPiSaveDialog::~CPiSaveDialog()
{
    // no need to delete child widgets, Qt does it all for us
}

//------------------------------------------------
void CPiSaveDialog::initDialog()
{
   setCaption( trUtf8( "Save partition to image" ) );

   m_lay = new QGridLayout (this, 7/*row*/, 4/*col*/, 0, 10);
   m_lay -> setMargin(15);

   m_labelDevice1 = new QLabel(this, "m_labelDevice1");
   m_labelDevice1 -> setText("Device:");
   m_lay -> addMultiCellWidget (m_labelDevice1, 0, 0, 0, 0);

   m_labelDevice2 = new QLabel(this, "m_labelDevice2");
   m_labelDevice2 -> setText(m_strDevice);
   m_lay -> addMultiCellWidget (m_labelDevice2, 0, 0, 1, 3);

   m_labelImagefile1 = new QLabel(this, "m_labelImagefile1");
   m_labelImagefile1 -> setText("Image file:");
   m_lay -> addMultiCellWidget (m_labelImagefile1, 1, 1, 0, 0);
   
   m_labelImagefile2 = new QLabel(this, "m_labelImagefile2");
   m_labelImagefile2 -> setText("You have to select an image file");
   m_lay -> addMultiCellWidget (m_labelImagefile2, 1, 1, 1, 2);

   m_btnSelectFile = new QPushButton("Select", this,"m_btnSelectFile");
   m_btnSelectFile->setText("Select");
   m_lay -> addMultiCellWidget (m_btnSelectFile, 1, 1, 3, 3);

   // ------------
   m_bgCompress = new QButtonGroup(1, Qt::Horizontal, "Compress level", this, "m_bgCompress");
   m_lay -> addMultiCellWidget (m_bgCompress, 2, 2, 0, 1);
   m_bgCompress -> setExclusive(true);

   m_radioCompress1 = new QRadioButton("None (very big image file)", m_bgCompress, "m_radioCompress1");
   m_radioCompress1 -> setChecked(m_options->dwCompression == COMPRESS_NONE);

#ifdef LEVELSUPPORT_LZO
   m_radioCompress2 = new QRadioButton("Lzo (very fast)", m_bgCompress, "m_radioCompress2");
   m_radioCompress2 -> setChecked(m_options->dwCompression == COMPRESS_LZO);
#endif // LEVELSUPPORT_LZO

#ifdef LEVELSUPPORT_GZIP
   m_radioCompress3 = new QRadioButton("Gzip (quite fast)", m_bgCompress, "m_radioCompress3");
   m_radioCompress3 -> setChecked(m_options->dwCompression == COMPRESS_GZIP);
#endif // LEVELSUPPORT_GZIP

#ifdef LEVELSUPPORT_BZIP2
   m_radioCompress4 = new QRadioButton("Bzip2 (very slow and efficient)", m_bgCompress, "m_radioCompress4");
   m_radioCompress4 -> setChecked(m_options->dwCompression == COMPRESS_BZIP2);
#endif // LEVELSUPPORT_BZIP2

   // ------------
   m_bgOptions = new QButtonGroup(1, Qt::Horizontal, "Options", this,"m_bgOptions");
   m_bgOptions -> setExclusive(false);
   m_lay -> addMultiCellWidget (m_bgOptions, 2, 2, 2, 3);
   m_checkCheckPart = new QCheckBox("Check partition before saving", m_bgOptions, "m_checkCheckPart");
   //m_checkDesc = new QCheckBox("Enter description about partition", m_bgOptions, "m_checkDesc");
   m_checkOverwrite = new QCheckBox("Overwrite image file without prompt", m_bgOptions, "m_checkOverwrite");
   m_checkCheckPart -> setChecked(!!m_options->bCheckBeforeSaving);
   //m_checkDesc -> setChecked(!!m_options->bAskDesc);
   m_checkOverwrite -> setChecked(!!m_options->bOverwrite);


   // ------------
   m_bgSplit = new QButtonGroup(1, Qt::Horizontal, "Split mode", this,"m_bgSplit");
   m_bgSplit -> setExclusive(false);
   m_lay -> addMultiCellWidget (m_bgSplit, 4, 4, 0, 3);
   m_radioSplit1 = new QRadioButton("Automatic split (when no space left)", m_bgSplit, "m_radioSplit1");
   m_radioSplit1 -> setChecked(!m_options->qwSplitSize);
   QHBox *hbox = new QHBox(m_bgSplit);
   m_radioSplit2 = new QRadioButton("Into files whose size is: ", hbox, "m_radioSplit1");
   m_radioSplit2 -> setChecked(!!m_options->qwSplitSize);
   m_spinSplitSize = new QSpinBox(hbox);
   m_spinSplitSize -> setRange(1024, 4*1024*1024);
   m_labelSplit = new QLabel(hbox, "m_labelSplit");
   m_labelSplit -> setText("  Kilo Bytes");
   slotSelectSplit1();

   // ------------
   m_bgDesc = new QButtonGroup(1, Qt::Horizontal, "Description of the image", this,"m_bgDesc");
   m_lay -> addMultiCellWidget (m_bgDesc, 5, 5, 0, 3);
   m_editDesc = new QTextEdit(m_bgDesc, "m_editDesc");

   // ------------
   m_btnOk = new COkButton(this,"m_btnOk");
   m_lay -> addMultiCellWidget (m_btnOk, 6, 6, 0, 1);
   m_btnOk -> setEnabled(!m_strImageFile.isEmpty());

   m_btnCancel = new CCancelButton(this,"m_btnCancel");
   m_lay -> addMultiCellWidget (m_btnCancel, 6, 6, 2, 3);
}

//------------------------------------------------
void CPiSaveDialog::slotSelect()
{
   QFileDialog dlg;
   QString fn;

   dlg.setViewMode( QFileDialog::Detail );
   fn = dlg.getSaveFileName("", "Partimage files (*.partimg)", this);

   if (!fn.isEmpty())
   {
      m_strImageFile = fn;
      m_labelImagefile2 -> setText(m_strImageFile);
   }

   m_btnOk -> setEnabled(!m_strImageFile.isEmpty());
}

//------------------------------------------------
void CPiSaveDialog::slotSelectSplit1()
{
   m_radioSplit1->setChecked(true);
   m_radioSplit2->setChecked(false);
   m_spinSplitSize->setEnabled(false);
   m_options->qwSplitSize = 0L;
}

//------------------------------------------------
void CPiSaveDialog::slotSelectSplit2()
{
   m_radioSplit1->setChecked(!true);
   m_radioSplit2->setChecked(!false);
   m_spinSplitSize->setEnabled(!false);
   m_options->qwSplitSize = m_spinSplitSize->value()*1024L;
}

//------------------------------------------------
void CPiSaveDialog::slotOk()
{
   if (m_radioCompress1->isChecked())
      m_options->dwCompression = COMPRESS_NONE;

#ifdef LEVELSUPPORT_LZO
   if(m_radioCompress2->isChecked())
      m_options->dwCompression = COMPRESS_LZO;
#endif // LEVELSUPPORT_LZO

#ifdef LEVELSUPPORT_GZIP
   if(m_radioCompress3->isChecked())
      m_options->dwCompression = COMPRESS_GZIP;
#endif // LEVELSUPPORT_GZIP

#ifdef LEVELSUPPORT_BZIP2
   if(m_radioCompress4->isChecked())
      m_options->dwCompression = COMPRESS_BZIP2;
#endif // LEVELSUPPORT_BZIP2

   m_options->bCheckBeforeSaving = m_checkCheckPart -> isChecked();
   m_options->bOverwrite = m_checkOverwrite -> isChecked();

   m_strDesc = m_editDesc->text();
   m_options->bAskDesc = false; //m_checkDesc -> isChecked();

   accept();
}

//------------------------------------------------
int CPiSaveDialog::getResults(char *szImagefile, int nMaxImageLen, char *szDevice, int nMaxDeviceLen, char *szDesc, int nMaxDescLen, COptions *options)
{
   snprintf(szImagefile, nMaxImageLen, "%s", m_strImageFile.latin1());
   snprintf(szDevice, nMaxDeviceLen, "%s", m_strDevice.latin1());
   snprintf(szDesc, nMaxDescLen, "%s", m_strDesc.latin1());
   options = m_options;

   return 0;
}

#include "formDialogPiSave.moc"
