#include "formDialogResize.h"
#include "pipushbutton.h"
#include "common.h"

#include <qslider.h>
#include <qlabel.h>
#include <qpushbutton.h>
#include <qlayout.h>
#include <qtooltip.h>
#include <qwhatsthis.h>
#include <qcombobox.h>
#include <qlineedit.h>

//------------------------------------------------
CResizeDialog::CResizeDialog(char *szDevice, DWORD dwStart, DWORD dwMinSize, DWORD dwMaxEnd, QWidget* parent /*= 0*/, const char* name /*= 0*/, bool modal/* = FALSE*/, WFlags fl/* = 0*/)
{
   // init
   m_strDevice.sprintf("%s", szDevice);
   m_dwMinSize = dwMinSize;
   m_dwMaxEnd = m_dwEnd = dwMaxEnd;
   m_dwStart = dwStart;

   initDialog();
   showInfos();

   connect(m_slider, SIGNAL(valueChanged(int)), this, SLOT(slotChanged()));
   connect(m_btnOk, SIGNAL(clicked()), this, SLOT(slotOk()));
   connect(m_btnCancel, SIGNAL(clicked()), this, SLOT(reject()));

   slotChanged();
}

//------------------------------------------------
void CResizeDialog::initDialog()
{
   DWORD dwMaxSize;
   QString strTemp;

   setCaption( trUtf8( "Resize a partition" ) );
   dwMaxSize = m_dwMaxEnd - m_dwStart;

   m_lay = new QGridLayout (this, 7/*row*/, 4/*col*/, 0, 10);
   m_lay -> setMargin(15);

   m_labelDevice1 = new QLabel(this, "m_labelDevice1");
   m_labelDevice1 -> setText("Device:");
   m_lay -> addMultiCellWidget (m_labelDevice1, 0, 0, 0, 0);

   m_labelDevice2 = new QLabel(this, "m_labelDevice2");
   m_labelDevice2 -> setText(m_strDevice);
   m_lay -> addMultiCellWidget (m_labelDevice2, 0, 0, 1, 3);

   m_labelTitle1 = new QLabel(this, "m_labelTitle1");
   m_labelTitle1 -> setText("sectors");
   m_lay -> addWidget (m_labelTitle1, 1, 1);

   m_labelTitle2 = new QLabel(this, "m_labelTitle2");
   m_labelTitle2 -> setText("megabytes");
   m_lay -> addWidget (m_labelTitle2, 1, 2);

   m_labelStart1 = new QLabel(this, "m_labelStart1");
   m_labelStart1 -> setText("Start");
   m_lay -> addWidget (m_labelStart1, 2, 0);

   m_labelStart2 = new QLabel(this, "m_labelStart2");
   m_lay -> addWidget (m_labelStart2, 2, 1);

   m_labelStart3 = new QLabel(this, "m_labelStart3");
   m_lay -> addWidget (m_labelStart3, 2, 2);

   m_labelEnd1 = new QLabel(this, "m_labelEnd1");
   m_labelEnd1 -> setText("End");
   m_lay -> addWidget (m_labelEnd1, 3, 0);

   m_labelEnd2 = new QLabel(this, "m_labelEnd2");
   m_lay -> addWidget (m_labelEnd2, 3, 1);

   m_labelEnd3 = new QLabel(this, "m_labelEnd3");
   m_lay -> addWidget (m_labelEnd3, 3, 2);

   m_labelSize1 = new QLabel(this, "m_labelSize1");
   m_labelSize1 -> setText("Size");
   m_lay -> addWidget (m_labelSize1, 4, 0);
   
   m_labelSize2 = new QLabel(this, "m_labelSize2");
   m_lay -> addWidget (m_labelSize2, 4, 1);

   m_labelSize3 = new QLabel(this, "m_labelSize3");
   m_lay -> addWidget (m_labelSize3, 4, 2);

   m_labelSize4 = new QLabel(this, "m_labelSize4");
   m_labelSize4 -> setText("Partition size");
   m_lay -> addWidget (m_labelSize4, 5, 0);
   
   m_slider = new QSlider(Qt::Horizontal, this,"m_slider");
   m_lay -> addMultiCellWidget (m_slider, 5, 5, 1, 3);
   m_slider -> setMaxValue (dwMaxSize);
   m_slider -> setMinValue (MAX_PARTSIZE_SECTORS);
   m_slider -> setValue (dwMaxSize);

   m_btnOk = new COkButton(this,"m_btnOk");
   m_lay -> addMultiCellWidget (m_btnOk, 6, 6, 0, 1);
  
   m_btnCancel = new CCancelButton(this,"m_btnCancel");
   m_lay -> addMultiCellWidget (m_btnCancel, 6, 6, 2, 3);
}

//------------------------------------------------
CResizeDialog::~CResizeDialog()
{
    // no need to delete child widgets, Qt does it all for us
}

//------------------------------------------------
void CResizeDialog::slotOk()
{
  // Check the Search text is not empty
  /*if (not good)
    {
    error
    return;
    }*/
  
  // Copy text into variables
  //m_strSearch = m_editSearch -> text();
  //m_strReplace = m_editReplace -> text();
  
  accept();
}

//------------------------------------------------
void CResizeDialog::slotChanged()
{
   DWORD dwSliderVal;

   dwSliderVal = m_slider->value();
   if (dwSliderVal < m_dwMinSize)
   {
      dwSliderVal = m_dwMinSize;
      m_slider->setValue(m_dwMinSize);
      slotChanged();
   }
   m_dwEnd = m_dwStart + dwSliderVal;

   showInfos();
}

//------------------------------------------------
int CResizeDialog::getResults(DWORD *dwSectEnd) 
{
   *dwSectEnd = m_dwEnd; 
   return 0;
}

//------------------------------------------------
void CResizeDialog::showInfos()
{
   DWORD dwSize;
   QString strTemp;
   char szTemp[4096];

   dwSize = m_dwEnd - m_dwStart;

   strTemp.sprintf("%lu", m_dwStart);
   m_labelStart2 -> setText(strTemp);

   strTemp.sprintf("%s", formatSize((QWORD)m_dwStart*SECTOR_SIZE, szTemp));
   m_labelStart3 -> setText(strTemp);

   strTemp.sprintf("%lu", m_dwEnd);
   m_labelEnd2 -> setText(strTemp);

   strTemp.sprintf("%s", formatSize((QWORD)m_dwEnd*SECTOR_SIZE, szTemp));
   m_labelEnd3 -> setText(strTemp);

   strTemp.sprintf("%lu", dwSize);
   m_labelSize2 -> setText(strTemp);

   strTemp.sprintf("%s", formatSize((QWORD)dwSize*SECTOR_SIZE, szTemp));
   m_labelSize3 -> setText(strTemp);
}

#include "formDialogResize.moc"
