/***************************************************************************
                          main.cpp  -  description
                             -------------------
    begin                : Wed Sep  4 19:21:54 UTC 2002
    copyright            : (C) 2002 by Francois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <qapplication.h>
#include <qfont.h>
#include <qstring.h>
#include <qtextcodec.h>
#include <qtranslator.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>

#include "workserver.h"
#include "piguicqt.h"
#include "args.h"

//------------------------------------------------
FILE *g_fDebug; // debug file
FILE *g_fLocalDebug; // debug file

//------------------------------------------------
int main(int argc, char *argv[])
{
   COptions options;
   BOOL bStartServer;
   WORD wServerPort;
   char szServerName[MAX_HOSTNAMESIZE];
   int nRes;

   // check the user is root
   if (geteuid() != 0)
   {
      fprintf (stderr, "Error: user must be root\n");
      return EXIT_FAILURE;
   }

   // init debug
   g_fDebug = fopen("/dev/null", "rw");
   g_fLocalDebug = fopen("/dev/null", "rw");

   // initialize options with defaults values
   initOptions(&options);

   // network = default values
   wServerPort = PARTIMAGE_GUI_PORT; 
   SNPRINTF(szServerName, "127.0.0.1");

   // read command line
   nRes = processCommandLine(argc, argv, &options, &bStartServer, szServerName, &wServerPort);
   if (nRes)
      return EXIT_FAILURE;

   // if the server is local
   if (strcmp(szServerName, "127.0.0.1") == 0)
   {
      // check user is root
      if (geteuid() != 0)
      {
	 pi_printf ("Error: user must be root\n");
	 return EXIT_FAILURE;
      }

      pi_printf ("running server...\n\n");
      // run the server
      pid_t nFork = fork();
      switch(nFork)
      {
	 case 0: // child = server
	    serverapp(wServerPort);
	    return EXIT_SUCCESS;
	    break;

	 case -1: // error
	    perror("fork");
	    return EXIT_FAILURE;

	 default: // parent = client
	    sleep(1); // wait the server is running
	    break;
      };
   }

#ifdef Q_WS_QWS // Frame Buffer 
   QApplication a(argc, argv, QApplication::GuiServer);
#else // X11
   QApplication a(argc, argv);
#endif // Q_WS_QWS

   a.setFont(QFont("helvetica", 12));
   QTranslator tor( 0 );
   // set the location where your .qm files are in load() below as the last parameter instead of "."
   // for development, use "/" to use the english original as
   // .qm files are stored in the base project directory.
   tor.load( QString("piguicqt.") + QTextCodec::locale(), "." );
   a.installTranslator( &tor );
    
   PiguicqtApp *piguicqt=new PiguicqtApp(&a, szServerName, wServerPort, &options);

   a.setMainWidget(piguicqt);

   piguicqt->show();

   return a.exec();
}


