/***************************************************************************
                          piguicqt.cpp  -  description
                             -------------------
    begin                : Wed Sep  4 19:21:54 UTC 2002
    copyright            : (C) 2002 by Francois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <qaccel.h>
#include <qsignalmapper.h>
#include <qstylefactory.h>
#include <qimage.h>

#include "piguicqt.h"
#include "images.h"
#include "newnet.h"

//------------------------------------------------
PiguicqtApp::PiguicqtApp(QApplication *app, char *szServerName, WORD wServerPort, COptions *options)
{
   m_options = options;
   m_app = app;
   m_view = NULL;
   
   setCaption(tr("PartGui, version " VERSION));
   
   initDoc(szServerName, wServerPort, options);
   
   // call inits to invoke all other construction parts
   initActions();
   initMenuBar();
   initToolBar();
   initStatusBar();
   
   initView();

   m_viewToolBar->setOn(true);
   m_viewStatusBar->setOn(true);
   
}

//------------------------------------------------
PiguicqtApp::~PiguicqtApp()
{
}

//------------------------------------------------
void PiguicqtApp::grayActions(BOOL bOperationInProgress/* = false*/)
{
   int nDisk, nPart;
   BOOL bUsedSpace;

   if (getView())
   {
      getView()->getCurrentDiskPart(&nDisk, &nPart);
      bUsedSpace = m_doc->isPartUsed(nDisk, nPart);
   }
   else
   {
      bUsedSpace = false;
   }

   m_operCreate -> setEnabled(!bOperationInProgress && !bUsedSpace);
   m_operFormat -> setEnabled(!bOperationInProgress && bUsedSpace);
   m_operResize -> setEnabled(false); //!bOperationInProgress && bUsedSpace);
   m_operDelete -> setEnabled(!bOperationInProgress && bUsedSpace);
   m_operPiSave -> setEnabled(false); //bUsedSpace);
   m_operPiRestore -> setEnabled(false); //bUsedSpace);
   m_operPiCopy -> setEnabled(false); //bUsedSpace);
   m_operProperties -> setEnabled(false); //bUsedSpace);
}

//------------------------------------------------
void PiguicqtApp::makeStyle(const QString &style)
{
   qApp->setStyle(style);
}

//------------------------------------------------
QPixmap loadPixmap( const QString& name )
{
   Embed *e = &embed_vec[ 0 ];
   while ( e->name ) 
   {
      if ( QString( e->name ) == name ) 
      {
	 QImage img;
	 img.loadFromData( (const uchar*)e->data, e->size );
	 QPixmap pix;
	 pix.convertFromImage( img );
	 return pix;
      }
      e++;
   }
   return QPixmap();
}

//------------------------------------------------
void PiguicqtApp::initActions()
{
   setUsesBigPixmaps(true);

  m_operLabel = new QAction(tr("New label..."), tr("&New label"), QAccel::stringToKey(tr("Ctrl+N")), this);
  m_operLabel->setStatusTip(tr("Create a new disk label"));
  connect(m_operLabel, SIGNAL(activated()), m_doc, SLOT(slotOperLabel()));

  m_operCreate = new QAction(tr("Create..."), loadPixmap("create.xpm"), tr("&Create"), QAccel::stringToKey(tr("Ctrl+C")), this);
  m_operCreate->setStatusTip(tr("Create a new partition"));
  connect(m_operCreate, SIGNAL(activated()), m_doc, SLOT(slotOperCreate()));

  m_operFormat = new QAction(tr("Format..."), loadPixmap("format.xpm"), tr("&Format"), QAccel::stringToKey(tr("Ctrl+F")), this);
  m_operFormat->setStatusTip(tr("Format an existing partition"));
  connect(m_operFormat, SIGNAL(activated()), m_doc, SLOT(slotOperFormat()));

  m_operResize = new QAction(tr("Resize..."), loadPixmap("resize.xpm"), tr("Resi&ze"), QAccel::stringToKey(tr("Ctrl+Z")), this);
  m_operResize->setStatusTip(tr("Resize an existing partition"));
  connect(m_operResize, SIGNAL(activated()), m_doc, SLOT(slotOperResize()));

  m_operDelete = new QAction(tr("Delete..."), loadPixmap("delete.xpm"), tr("&Delete"), QAccel::stringToKey(tr("Ctrl+D")), this);
  m_operDelete->setStatusTip(tr("Delete an existing partition"));
  connect(m_operDelete, SIGNAL(activated()), m_doc, SLOT(slotOperDelete()));

  m_operPiSave = new QAction(tr("Save..."), loadPixmap("save.xpm"), tr("&Save"), QAccel::stringToKey(tr("Ctrl+S")), this);
  m_operPiSave->setStatusTip(tr("Save partition to image file"));
  connect(m_operPiSave, SIGNAL(activated()), m_doc, SLOT(slotOperPiSave()));

  m_operPiRestore = new QAction(tr("Restore..."), loadPixmap("restore.xpm"), tr("&Restore"), QAccel::stringToKey(tr("Ctrl+R")), this);
  m_operPiRestore->setStatusTip(tr("Restore partition from image file"));
  connect(m_operPiRestore, SIGNAL(activated()), m_doc, SLOT(slotOperPiRestore()));

  m_operPiCopy = new QAction(tr("Copy..."), loadPixmap("copy.xpm"), tr("Cop&y"), QAccel::stringToKey(tr("Ctrl+Y")), this);
  m_operPiCopy->setStatusTip(tr("Copy partition onto another partition"));
  connect(m_operPiCopy, SIGNAL(activated()), m_doc, SLOT(slotOperPiCopy()));

  m_operRefresh = new QAction(tr("Refresh"), loadPixmap("refresh.xpm"), tr("R&efresh"), QAccel::stringToKey(tr("Ctrl+E")), this);
  m_operRefresh->setStatusTip(tr("Redraw informations about the disk layout"));
  connect(m_operRefresh, SIGNAL(activated()), m_doc, SLOT(slotOperRefresh()));

  m_operProperties = new QAction(tr("Properties..."), loadPixmap("properties.xpm"), tr("&Properties"), QAccel::stringToKey(tr("Ctrl+P")), this);
  m_operProperties->setStatusTip(tr("Give informations about an existing partition"));
  connect(m_operProperties, SIGNAL(activated()), m_doc, SLOT(slotOperProperties()));

  m_fileQuit = new QAction(tr("Exit"), loadPixmap("exit.xpm"), tr("E&xit"), QAccel::stringToKey(tr("Ctrl+Q")), this);
  m_fileQuit->setStatusTip(tr("Quits the application"));
  m_fileQuit->setWhatsThis(tr("Exit\n\nQuits the application"));
  connect(m_fileQuit, SIGNAL(activated()), this, SLOT(slotFileQuit()));

  m_viewToolBar = new QAction(tr("Toolbar"), tr("Tool&bar"), 0, this, 0, true);
  m_viewToolBar->setStatusTip(tr("Enables/disables the toolbar"));
  m_viewToolBar->setWhatsThis(tr("Toolbar\n\nEnables/disables the toolbar"));
  connect(m_viewToolBar, SIGNAL(toggled(bool)), this, SLOT(slotViewToolBar(bool)));

  m_viewStatusBar = new QAction(tr("Statusbar"), tr("&Statusbar"), 0, this, 0, true);
  m_viewStatusBar->setStatusTip(tr("Enables/disables the statusbar"));
  m_viewStatusBar->setWhatsThis(tr("Statusbar\n\nEnables/disables the statusbar"));
  connect(m_viewStatusBar, SIGNAL(toggled(bool)), this, SLOT(slotViewStatusBar(bool)));

  m_helpAboutApp = new QAction(tr("About"), tr("&About..."), 0, this);
  m_helpAboutApp->setStatusTip(tr("About the application"));
  m_helpAboutApp->setWhatsThis(tr("About\n\nAbout the application"));
  connect(m_helpAboutApp, SIGNAL(activated()), this, SLOT(slotHelpAbout()));

  m_helpAboutQt = new QAction(tr("About Qt"), tr("About &Qt..."), 0, this);
  m_helpAboutQt->setStatusTip(tr("About Qt"));
  m_helpAboutQt->setWhatsThis(tr("About Qt\n\nAbout Qt"));
  connect(m_helpAboutQt, SIGNAL(activated()), this, SLOT(slotHelpAboutQt()));

}

//------------------------------------------------
void PiguicqtApp::initMenuBar()
{
   m_fileMenu=new QPopupMenu();
   m_contextMenu = new QPopupMenu();

   m_operLabel->addTo(m_fileMenu);
   m_operLabel->addTo(m_contextMenu);
   m_operCreate->addTo(m_fileMenu);
   m_operCreate->addTo(m_contextMenu);
   m_operFormat->addTo(m_fileMenu);
   m_operFormat->addTo(m_contextMenu);
   m_operResize->addTo(m_fileMenu);
   m_operResize->addTo(m_contextMenu);
   m_operDelete->addTo(m_fileMenu);
   m_operDelete->addTo(m_contextMenu);

   m_fileMenu->insertSeparator();
   m_contextMenu->insertSeparator();

   m_operPiSave->addTo(m_fileMenu);
   m_operPiSave->addTo(m_contextMenu);
   m_operPiRestore->addTo(m_fileMenu);
   m_operPiRestore->addTo(m_contextMenu);
   m_operPiCopy->addTo(m_fileMenu);
   m_operPiCopy->addTo(m_contextMenu);

   m_fileMenu->insertSeparator();
   m_contextMenu->insertSeparator();

   m_operRefresh->addTo(m_fileMenu);
   m_operRefresh->addTo(m_contextMenu);
   m_operProperties->addTo(m_fileMenu);
   m_operProperties->addTo(m_contextMenu);

   m_fileMenu->insertSeparator();
   m_fileQuit->addTo(m_fileMenu);

   m_viewMenu=new QPopupMenu();
   m_viewMenu->setCheckable(true);
   m_viewToolBar->addTo(m_viewMenu);
   m_viewStatusBar->addTo(m_viewMenu);

   // ===================================

   QPopupMenu *style = new QPopupMenu(this);
   style->setCheckable(TRUE);
   
   style->setCheckable( TRUE );
   QActionGroup *ag = new QActionGroup(this, 0);
   ag->setExclusive( TRUE );
   QSignalMapper *styleMapper = new QSignalMapper( this );
   connect( styleMapper, SIGNAL( mapped( const QString& ) ), this, SLOT( makeStyle( const QString& ) ) );
   QStringList list = QStyleFactory::keys();
   list.sort();

   for ( QStringList::Iterator it = list.begin(); it != list.end(); ++it ) 
   {
      QString styleStr = *it;
      QString styleAccel = styleStr;
      QAction *a = new QAction( styleStr, QIconSet(), styleAccel, 0, ag, 0, ag->isExclusive() );
      connect( a, SIGNAL( activated() ), styleMapper, SLOT(map()) );
      styleMapper->setMapping( a, a->text() );
   }
   ag->addTo(style);

   m_helpMenu=new QPopupMenu();
   m_helpAboutApp->addTo(m_helpMenu);
   m_helpAboutQt->addTo(m_helpMenu);

   menuBar()->insertItem(tr("&Partition"), m_fileMenu);
   menuBar()->insertItem(tr("&View"), m_viewMenu);
   menuBar()->insertItem("&Style" , style);
   menuBar()->insertSeparator();
   menuBar()->insertItem(tr("&Help"), m_helpMenu);

}

//------------------------------------------------
void PiguicqtApp::initToolBar()
{
   m_fileToolbar = new QToolBar(this, "file operations");
   m_operCreate->addTo(m_fileToolbar);
   m_operDelete->addTo(m_fileToolbar);
   m_operFormat->addTo(m_fileToolbar);
   //m_operResize->addTo(m_fileToolbar);

   m_fileToolbar->addSeparator();

   m_operPiSave->addTo(m_fileToolbar);
   m_operPiRestore->addTo(m_fileToolbar);
   m_operPiCopy->addTo(m_fileToolbar);

   m_fileToolbar->addSeparator();

   m_operProperties->addTo(m_fileToolbar);
   m_operRefresh->addTo(m_fileToolbar);

   m_fileToolbar->addSeparator();
   m_fileQuit->addTo(m_fileToolbar);
   //m_fileToolbar->addSeparator();
   //QWhatsThis::whatsThisButton(m_fileToolbar);
}

//------------------------------------------------
void PiguicqtApp::initStatusBar()
{
  statusBar()->message(tr("Ready."), 2000);
}

//------------------------------------------------
void PiguicqtApp::initDoc(char *szServerName, WORD wServerPort, COptions *options)
{
   m_doc = new PiguicqtDoc(this, szServerName, wServerPort, options);

   // get partition/drives infos
   m_doc -> initDrives();
}

//------------------------------------------------
void PiguicqtApp::initView()
{
  // set the main widget here
  m_view=new PiguicqtView(this, m_doc);

  setCentralWidget(m_view);
}

//------------------------------------------------
bool PiguicqtApp::queryExit()
{
  int exit=QMessageBox::information(this, tr("Quit..."),
                                    tr("Do your really want to quit?"),
                                    QMessageBox::Ok, QMessageBox::Cancel);

  if (exit==1)
  {

  }
  else
  {

  };

  return (exit==1);
}

//------------------------------------------------
void PiguicqtApp::exitNow()
{
   pi_printf ("PiguicqtApp::exitNow\n");
   
   m_doc->sendExit();
   qApp->quit();
}

//------------------------------------------------
void PiguicqtApp::closeEvent(QCloseEvent * /*e*/)
{
   pi_printf ("PiguicqtApp::closeEvent()\n");
   exitNow();
}

//------------------------------------------------
void PiguicqtApp::slotFileQuit()
{
   statusBar()->message(tr("Exiting application..."));

   /*if(m_doc->isModified())
     {
     if(queryExit())
     {
     exitNow();
     }
     else
     {

     };
     }
     else
     {*/
   exitNow();
   //}

   statusBar()->message(tr("Ready."));
}

//------------------------------------------------
void PiguicqtApp::slotViewToolBar(bool toggle)
{
  statusBar()->message(tr("Toggle toolbar..."));
  // turn Toolbar on or off

  if (toggle== false)
  {
    m_fileToolbar->hide();
  }
  else
  {
    m_fileToolbar->show();
  };

  statusBar()->message(tr("Ready."));
}

//------------------------------------------------
void PiguicqtApp::slotViewStatusBar(bool toggle)
{
  statusBar()->message(tr("Toggle statusbar..."));
  //turn Statusbar on or off

  if (toggle == false)
  {
    statusBar()->hide();
  }
  else
  {
    statusBar()->show();
  }

  statusBar()->message(tr("Ready."));
}

//------------------------------------------------
void PiguicqtApp::slotHelpAbout()
{
   QMessageBox::about(this,tr("About..."), tr("<b>PartGUI: GUI for GNU Parted</b> <i></i><br><br>Version " VERSION "<br><br>"
					      "<i>based on trolltech Qt</i><br><br>"
					      "<br>(c) 2003 by Francois Dupoux<br><br>" HOMEPAGE "<br>") );
}

//------------------------------------------------
void PiguicqtApp::slotHelpAboutQt()
{
  QMessageBox::aboutQt(this);
}

//------------------------------------------------
// disable actions, and wait while there is no message on the network
void PiguicqtApp::waitDuringOperation(CNewNet *net)
{
   grayActions(true);
   
   while (net->messageToRead() == false)
   {
      getApp()->processEvents();
      usleep(50);
   }

   grayActions(false);
}
