/***************************************************************************
                          piguicqtdoc.cpp  -  description
                             -------------------
    begin                : Wed Sep  4 19:21:54 UTC 2002
    copyright            : (C) 2002 by Francois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "global.h"
#include "newnet.h"
#include "workserver.h"
#include "result.h"
#include "piguicqt.h"
#include "piguicqtdoc.h"
#include "piguicqtview.h"
#include "formDialogPiSave.h"
#include "formDialogPartedCreate.h"
#include "formDialogResize.h"
#include "formDialogFormat.h"

#include <parted/parted.h>
#include <parted/disk.h>

#include <qstrlist.h>

//------------------------------------------------
PiguicqtDoc::PiguicqtDoc(QWidget *parent, char *szServerName, WORD wServerPort, COptions *options)
{
   //modified = false;
   m_options = options;
   m_app = (PiguicqtApp *) parent;
   
   pi_printf ("--- client: Preparing connexion\n");

   m_net = new CNewNet("ClientCnx");
   m_net -> connect(szServerName, wServerPort, PIGUI_BANNER_TEXT, PIGUI_BANNER_VERSION);

   pi_printf ("--- client: Connected\n");
}

//------------------------------------------------
PiguicqtDoc::~PiguicqtDoc()
{
   sendExit();
}

//------------------------------------------------
int PiguicqtDoc::umountPart(char *szPart)
{
   u8 bResult;
   QString strText;

   m_net->send_u32(PICMD_UNMOUNTED);
   m_net->send_string(szPart);

   bResult = m_net->recv_u8();

   if (bResult != 0)
   {
      strText.sprintf("Cannot unmount %s.", szPart);
      QMessageBox::information (getApp(), "Unmount partition", strText);
      return -1;
   }

   return 0; // success
}

//------------------------------------------------
int PiguicqtDoc::checkPartNotMounted(char *szPart)
{
   u8 bIsMounted;
   char szMountPoint[4096];
   QString strText;

   m_net->send_u32(PICMD_GET_ISMOUNTED);
   m_net->send_string(szPart);

   bIsMounted = m_net->recv_u8();
   if (!bIsMounted)
      return 0; // okay

   m_net->recv_string(szMountPoint);

   strText.sprintf("%s is mounted on %s.<br>It's dangerous to work on a mounted partition.<br>"
		   "It's recommanded to unmount it before. (<b>umount %s</b>)<br>", 
		   szPart, szMountPoint, szPart);
   switch (QMessageBox::warning (getApp(), "Warning: dangerous !", strText, "Cancel", "Continue", "Try to unmount"))
   {
      case 0: // Cancel
	 return -1;
	 break;

      case 1: // Continue
	 return 0;
	 break;

      case 2: // Try to umount
	 return umountPart(szPart);
	 break;
   };

   return 0; 
}

//------------------------------------------------
void PiguicqtDoc::sendExit()
{
   u32 a;

   m_net -> send_u32(PICMD_EXIT);

   do 
   {
      a = m_net -> recv_u32();
   } while (a != PICMD_EXIT);

   pi_printf ("--- client: Disconnected\n");
}

//------------------------------------------------
int PiguicqtDoc::initDrives()
{
   DWORD dwSize;

   m_net->send_u32(PICMD_GET_HDLIST);
   m_net->send_u8(getOptions()->bShowPartDetails);
   m_dwDrivesCount=0;

   while ((dwSize = m_net->recv_u32()) != 0)
   {
      // alloc memory
      m_drives[m_dwDrivesCount] = new CDriveDoc;
      if (!m_drives[m_dwDrivesCount])
	 return -1;

      recvDiskInfos(m_drives[m_dwDrivesCount]);
      m_dwDrivesCount++;
   }

   return 0;
}

//------------------------------------------------
int PiguicqtDoc::recvDiskInfos(CDriveDoc *drives)
{
   DWORD dwPartsCount;
   DWORD i;
   CPartInfo pi;

   // get disk info
   m_net->recv(&(drives->dinfo), sizeof(CDiskInfo));

   // get parts list
   dwPartsCount = m_net->recv_u32();
      
   if (dwPartsCount > MAX_DRIVES)
   {
      pi_printf ("Error in  PiguicqtDoc::initDrives(): too many parts: %lu = %.8x\n", dwPartsCount, dwPartsCount);
      return -1;
   }

   for (i=0; i < dwPartsCount; i++)
   {
      m_net->recv(&pi, sizeof(CPartInfo));
      if (strcmp(pi.szMagic, MAGIC_INFOGUI) != 0)
      {
	 pi_printf ("Error magic: [%s]\n", pi.szMagic);
	 //return -1;
      }
      drives->piInfo[i] = new CPartInfo;
      if (!(drives->piInfo[i]))
	 return -1;
      memcpy(drives->piInfo[i], &pi, sizeof(pi));
   }
      
   drives->dwPartsCount = dwPartsCount;
   return 0;
}

//------------------------------------------------
void PiguicqtDoc::refreshDisk(int nDisk)
{
   // send command
   m_net->send_u32(PICMD_GET_HDLIST_FORONEDISK);

   // send command param
   m_net->send_u8(getOptions()->bShowPartDetails);
   m_net->send_string(m_drives[nDisk]->dinfo.szDevice);

   recvDiskInfos(m_drives[nDisk]);

   // update view
   getView()->refreshDisk(nDisk);
}

//------------------------------------------------
PiguicqtView *PiguicqtDoc::getView()
{
   return m_app->getView();
}

//------------------------------------------------
void PiguicqtDoc::slotOperLabel()
{
   int nDisk;
   int nPart;
   QString strText;
   int nRes;
   CResult res;

   pi_printf ("PiguicqtDoc::slotOperLabel()\n");
   
   strText = QString("All partitions of the hard disk will be destroyed !");
   nRes = QMessageBox::warning (getApp(), "Warning: dangerous !", strText, QMessageBox::Ok, QMessageBox::Cancel, QMessageBox::NoButton);
   if (nRes == QMessageBox::Ok)
   {
      getView()->getCurrentDiskPart(&nDisk, &nPart);
   
      m_net->send_u32(PICMD_OPERATION_PARTED_NEWLABEL);
      m_net->send_string(m_drives[nDisk]->dinfo.szDevice);
      m_net->recv_result(&res); // wait the operation is finished
      if (res.getResult() == RESULT_ERROR)
      {
	 QMessageBox::critical (getApp(), "Error", res.getError(), QMessageBox::Ok, QMessageBox::NoButton);
      }

      refreshDisk(nDisk);
      pi_printf ("--> New label done\n");
   }
   else
   {
      pi_printf ("--> New label canceled\n");
   }

}

//------------------------------------------------
void PiguicqtDoc::slotOperCreate()
{
   int nDisk;
   int nPart;
   DWORD dwPartType;
   DWORD dwSectorStart;
   DWORD dwSectorEnd;
   BOOL bCanMakeExtended;
   BOOL bCanMakePrimary;
   BOOL bCanMakeLogical;
   BOOL bSpaceInsideExtended;
   QStrList strlFs;
   char szFs[512];
   char szLabel[MAX_LABEL];
   CResult res;
   CFileSys fs;

   pi_printf ("PiguicqtDoc::slotOperCreate()\n");

   getView()->getCurrentDiskPart(&nDisk, &nPart);
   pi_printf ("CreatePart: Disk(%d)\n", nDisk);

   // test
   dwSectorStart = m_drives[nDisk]->piInfo[nPart]->qwGeomStart;
   dwSectorEnd = m_drives[nDisk]->piInfo[nPart]->qwGeomEnd;

   bSpaceInsideExtended = isSpaceInsideExtended(m_drives[nDisk]->dinfo, dwSectorStart, dwSectorEnd);
   bCanMakeExtended = (m_drives[nDisk]->dinfo.dwPrimaryCount < 4) && (m_drives[nDisk]->dinfo.bExtendedExists == false);
   bCanMakePrimary = (bSpaceInsideExtended == false) && (m_drives[nDisk]->dinfo.dwPrimaryCount < 4);
   bCanMakeLogical = (bSpaceInsideExtended == true);
   //pi_printf ("m_drives[nDisk]->dinfo.dwPrimaryCount=%d\n", m_drives[nDisk]->dinfo.dwPrimaryCount);
   pi_printf ("bCanMakeExtended=%d and bCanMakePrimary=%d and bCanMakeLogical=%d\n", bCanMakeExtended, bCanMakePrimary, bCanMakeLogical);

   // fill list of file systems
   m_net->send_u32(PICMD_GET_FSLIST);
   while (m_net->recv_u32() == 1)
   {
      m_net->recv(&fs, sizeof(CFileSys));
      if (*fs.szCmdMkfsName)
	 strlFs.append(QString(fs.szDesc));
   }

   // get infos about what to do (part_type, start, end)
   CPartedCreateDialog dlg(m_drives[nDisk]->dinfo.szDevice, &strlFs, dwSectorStart, dwSectorEnd, bCanMakeExtended, bCanMakePrimary, bCanMakeLogical);
   if (dlg.exec() == QDialog::Rejected)
      return ;

   dlg.getResults(&dwSectorStart, &dwSectorEnd, &dwPartType, szFs, sizeof(szFs), szLabel, sizeof(szLabel));

   pi_printf ("PiguicqtDoc::slotOperCreate PARTCREATE: start=%lu, end=%lu, type=%lu\n", dwSectorStart, dwSectorEnd, dwPartType);

   m_net->send_u32(PICMD_OPERATION_PARTED_CREATE);
   m_net->send_string(m_drives[nDisk]->dinfo.szDevice);
   m_net->send_u32(dwPartType);
   m_net->send_u32(dwSectorStart);
   m_net->send_u32(dwSectorEnd);
   //m_net->send_string(m_drives[nDisk]->piInfo[nPart]->szPartName);
   m_net->send_string(szFs);
   m_net->send_string(szLabel);

   m_net->recv_result(&res); // wait the operation is finished
   if (res.getResult() == RESULT_ERROR)
   {
      QMessageBox::critical (getApp(), "Error", res.getError(), QMessageBox::Ok, QMessageBox::NoButton);
   }
 
   refreshDisk(nDisk);

   pi_printf ("finished\n");
}


//------------------------------------------------
void PiguicqtDoc::slotOperResize()
{
   int nDisk;
   int nPart;
   DWORD dwSectorStart;
   DWORD dwSectorEnd;
   DWORD dwMinSize;

   pi_printf ("PiguicqtDoc::slotOperResize()\n");

   getView()->getCurrentDiskPart(&nDisk, &nPart);

   // get informations about partition
   m_net->send_u32(PICMD_OPERATION_GET_PARTDETAILS);
   m_net->send_string(m_drives[nDisk]->dinfo.szDevice);
   m_net->send_u32(m_drives[nDisk]->piInfo[nPart]->dwMinor);

   m_net->recv(&m_drives[nDisk]->piInfo[nPart]->headInfo, sizeof(CLocalHeader));


   // test
   dwSectorStart = m_drives[nDisk]->piInfo[nPart]->qwGeomStart;
   dwSectorEnd = m_drives[nDisk]->piInfo[nPart]->qwGeomEnd;
   dwMinSize = (SGET(m_drives[nDisk]->piInfo[nPart]->headInfo,qwBlockSize) * SGET(m_drives[nDisk]->piInfo[nPart]->headInfo,qwUsedBlocks)) / SECTOR_SIZE;

   // get infos about what to do (part_type, start, end)
   CResizeDialog dlg(m_drives[nDisk]->piInfo[nPart]->szPartName, dwSectorStart, dwMinSize, dwSectorEnd);
   if (dlg.exec() == QDialog::Rejected)
      return ;

   dlg.getResults(&dwSectorEnd);
}

//------------------------------------------------
void PiguicqtDoc::slotOperFormat()
{
   int nDisk;
   int nPart;
   int nRes;
   QStrList strlFs;
   QString strText;
   CFileSys fs;
   char szResFs[128];
   char szResLabel[MAX_LABEL];
   CResult res;

   pi_printf ("PiguicqtDoc::slotOperFormat()\n");
   
   getView()->getCurrentDiskPart(&nDisk, &nPart);

   if (checkPartNotMounted(m_drives[nDisk]->piInfo[nPart]->szPartName) < 0)
      return;

   // fill list of file systems
   m_net->send_u32(PICMD_GET_FSLIST);
   while (m_net->recv_u32() == 1)
   {
      m_net->recv(&fs, sizeof(CFileSys));
      if (*fs.szCmdMkfsName)
	 strlFs.append(QString(fs.szDesc));
   }

   CFormatDialog dlg(m_drives[nDisk]->piInfo[nPart]->szPartName, &strlFs, m_drives[nDisk]->piInfo[nPart]->szFilesysName);
   if (dlg.exec() == QDialog::Rejected)
      return ;

   strText = QString("All files of this partition will be destroyed !");
   nRes = QMessageBox::warning (getApp(), "Warning: dangerous !", strText, QMessageBox::Ok, QMessageBox::Cancel, QMessageBox::NoButton);
   if (nRes == QMessageBox::Ok)
   {
      pi_printf ("Format partition: Disk(%d), part(%d) = [%s]->[%s]\n", 
		 nDisk, nPart, m_drives[nDisk]->piInfo[nPart]->szPartName, m_drives[nDisk]->dinfo.szDiskName);

      dlg.getResults(szResFs, sizeof(szResFs), szResLabel, sizeof(szResLabel));
      
      m_net->send_u32(PICMD_OPERATION_FORMAT);
      m_net->send_string(m_drives[nDisk]->piInfo[nPart]->szPartName);
      m_net->send_string(szResFs);
      m_net->send_string(szResLabel);

      getApp()->waitDuringOperation(m_net);

      m_net->recv_result(&res); // wait the operation is finished
      if (res.getResult() != RESULT_SUCCESS)
      {
	 QMessageBox::critical (getApp(), "Error", res.getError(), QMessageBox::Ok, QMessageBox::NoButton);
      }

      pi_printf("slotOperFormat() -> call refreshDisk()\n");
      refreshDisk(nDisk);

      pi_printf ("--> Format part done\n");
   }
   else
   {
      pi_printf ("--> Format part canceled\n");
   }
   
}

//------------------------------------------------
void PiguicqtDoc::slotOperDelete()
{
   QString strText;
   CResult res;
   int nDisk;
   int nPart;
   int nRes;

   pi_printf ("PiguicqtDoc::slotOperDelete()\n");

   getView()->getCurrentDiskPart(&nDisk, &nPart);

   if (checkPartNotMounted(m_drives[nDisk]->piInfo[nPart]->szPartName) < 0)
      return;

   strText = QString("All files of this partition will be destroyed !");
   nRes = QMessageBox::warning (getApp(), "Warning: dangerous !", strText, QMessageBox::Ok, QMessageBox::Cancel, QMessageBox::NoButton);
   if (nRes == QMessageBox::Ok)
   {
      pi_printf ("Delete Disk(%d), part(%d) = [%s]->[%s]\n", nDisk, nPart, m_drives[nDisk]->piInfo[nPart]->szPartName, m_drives[nDisk]->dinfo.szDiskName);

      m_net->send_u32(PICMD_OPERATION_PARTED_DELETE);
      m_net->send_string(m_drives[nDisk]->dinfo.szDevice);
      m_net->send_u32(m_drives[nDisk]->piInfo[nPart]->dwMinor);
      m_net->recv_result(&res); // wait the operation is finished
      if (res.getResult() == RESULT_ERROR)
      {
	 QMessageBox::critical (getApp(), "Error", res.getError(), QMessageBox::Ok, QMessageBox::NoButton);
      }

      refreshDisk(nDisk);

      pi_printf ("--> Delete part done\n");
   }
   else
   {
      pi_printf ("--> Delete part canceled\n");
   }

   pi_printf ("finished\n");
}

//------------------------------------------------
void PiguicqtDoc::slotOperPiSave()
{
   CResult res;
   char szImageFile[4096];
   char szDevice[4096];
   char szDesc[8192];
   int nDisk;
   int nPart;

   pi_printf ("PiguicqtDoc::slotOperPiSave()\n");
   getView()->getCurrentDiskPart(&nDisk, &nPart);

   CPiSaveDialog dlg(m_drives[nDisk]->piInfo[nPart]->szPartName, getApp()->getOptions());
   if (dlg.exec() == QDialog::Rejected)
      return ;

   dlg.getResults(szImageFile, sizeof(szImageFile), szDevice, sizeof(szDevice), szDesc, sizeof(szDesc), getApp()->getOptions());

   m_net->send_u32(PICMD_OPERATION_PI_SAVE);
   m_net->send(getApp()->getOptions(), sizeof(COptions));
   m_net->send_string(szImageFile);
   m_net->send_string(szDevice);  
   m_net->recv_result(&res); // wait the operation is finished
   if (res.getResult() == RESULT_ERROR)
   {
      QMessageBox::critical (getApp(), "Error", res.getError(), QMessageBox::Ok, QMessageBox::NoButton);
   }
}

//------------------------------------------------
void PiguicqtDoc::slotOperPiRestore()
{
   pi_printf ("PiguicqtDoc::slotOperPiRestore()\n");
   //m_net->send_u32(PICMD_OPERATION_RESTORE);
}

//------------------------------------------------
void PiguicqtDoc::slotOperPiCopy()
{
   pi_printf ("PiguicqtDoc::slotOperPiCopy()\n");
   //m_net->send_u32(PICMD_OPERATION_COPY);
}

//------------------------------------------------
void PiguicqtDoc::slotOperProperties()
{
   pi_printf ("PiguicqtDoc::slotOperProperties()\n");
}

//------------------------------------------------
void PiguicqtDoc::slotOperRefresh()
{
   pi_printf ("PiguicqtDoc::slotOperRefresh()\n");

   int nDisk;
   int nPart;
   getView()->getCurrentDiskPart(&nDisk, &nPart);
   refreshDisk(nDisk);
}

//------------------------------------------------
BOOL PiguicqtDoc::isPartUsed(int nDrive, int nPart)
{
   return (m_drives[nDrive]->piInfo[nPart]->dwType != PED_PARTITION_FREESPACE);
}
