/***************************************************************************
                          piguicqtview.cpp  -  description
                             -------------------
    begin                : Wed Sep  4 19:21:54 UTC 2002
    copyright            : (C) 2002 by Francois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "piguicqtview.h"
#include "pipushbutton.h"
#include "piguicqt.h"

#include <qvbox.h> 
#include <qlistview.h> 
#include <qimage.h>
#include <qpixmap.h>
#include <qbuttongroup.h> 
#include <qprogressbar.h>
#include <qlabel.h>

#include <parted/parted.h>

char *g_szColumns[] = {"Position", "Minor", "Device", "Flags", "File System", "Size", 
		       "Usage", "Type", "Start", "End", "Label", "Used", "Free", "Mount point", "*"};

//------------------------------------------------
COptions *PiguicqtView::getOptions()
{
   return m_app->getOptions();
}

//------------------------------------------------
PiguicqtView::PiguicqtView(QWidget *parent, PiguicqtDoc *doc) : QTabWidget(parent)
{
   m_app = (PiguicqtApp *) parent;
   m_doc = doc;

   DWORD dwDrivesCount=0;
   DWORD i;

   m_nDrive = m_nPart = 0;
   dwDrivesCount = m_doc->getDrivesCount();

   for (i=0; i < dwDrivesCount; i++)
   {
      // init device
      m_drive[i] = new CDriveView;
      if (!m_drive[i])
	 return ;

      createTab(m_drive[i], i);
   }

   setCurrentPart(0, 0);
}

//------------------------------------------------
int PiguicqtView::createTab(CDriveView *drive, int nTabPos)
{
   DWORD j, k;
   DWORD dwPartsCount=0;
   CPartInfo *pi;
   QString strTemp;
   QString strFlags;
   QString strType;
   QString strMountPoint;
   QString strMinor;
   QString strUsage;
   QString strGeomStart;
   QString strGeomEnd;
   QString strSize;
   QString strPos;
   QString strUsageUsed;
   QString strUsageFree;
   QString strLabel;
   char szTemp[1024];
   QWORD qwSummaryUsed;
   QWORD qwSummaryTotal;

   // init
   QImage img(2*16, 16, 8, 256);
   img.create(2*16, 16, 8, 256);
   m_nPart = 0;
   
   dwPartsCount = m_doc->getPartsCount(nTabPos);

   drive->vbox = new QVBox((QWidget *)this);
   drive->hBoxBtnBar = new QHBox((QWidget *)drive->vbox, "QHBoxBtnBar");
   if (getOptions()->bShowPartDetails)
      drive->hBoxSummary = new QHBox((QWidget *)drive->vbox, "QHBoxSummary");
   drive->lv = new QListView ((QWidget *)drive->vbox);
   connect(drive->lv, SIGNAL(selectionChanged(QListViewItem *)), this, SLOT(slotListViewClickedLeft(QListViewItem *)));
   drive->hBoxBtnBar -> setMinimumHeight(60);
   drive->lv -> setAllColumnsShowFocus(true);
   drive->m_btnList = new CPiPushButtons(dwPartsCount);

   for (j=0; *g_szColumns[j] != '*'; j++)
      drive -> lv -> addColumn(g_szColumns[j]);

   // add partitions
   qwSummaryUsed = 0;
   for (j=0; j < dwPartsCount; j++)
   {
      pi = m_doc->getPartInfo(nTabPos, j);
      strType.sprintf("%s ", ped_partition_type_get_name((PedPartitionType)pi->dwType));
      if (pi->dwType != PED_PARTITION_FREESPACE)
      {
	 strUsage.sprintf ("%lu %%", (u32)pi->dwPctUsage);
	 strMinor.sprintf ("%.2lu", (u32)pi->dwMinor);
	 strMountPoint.sprintf ("%s", pi->szMountPoint);
	 strUsageUsed.sprintf("%s", formatSize(SGET(pi->headInfo, qwUsedBlocks)*SGET(pi->headInfo, qwBlockSize), szTemp));
	 strUsageFree.sprintf("%s", formatSize((SGET(pi->headInfo, qwBlocksCount)-SGET(pi->headInfo, qwUsedBlocks))
					       *SGET(pi->headInfo, qwBlockSize), szTemp));
	 strLabel.sprintf("%s", pi->headInfo.val.szLabel);
	 qwSummaryUsed += SGET(pi->headInfo, qwUsedBlocks)*SGET(pi->headInfo, qwBlockSize);
      }
      else
      {
	 strUsage = strMinor = strMountPoint = strUsageUsed = strUsageFree = strLabel = QString("");
	 if (pi->bInsideExtended)
	    strType += QString(" (extended)");
      }
      strGeomStart.sprintf ("%.10lu", (u32)(u64)pi->qwGeomStart);
      strGeomEnd.sprintf ("%.10lu", (u32)(u64)pi->qwGeomEnd);
      strSize.sprintf("%s (%.2f %%)", formatSize(pi->qwSize, szTemp), ((float)pi->dwP1000Disk)/10.0); 
      strPos.sprintf ("%.2lu", (u32)pi->dwPosition);

      //strTemp.sprintf(" [bInExt=%d]", pi->bInsideExtended);
      //strPos += strTemp;

      // button
      drive->m_pbItem[j]  = new CPiPushButton(*pi, drive->hBoxBtnBar);
      drive->m_pbItem[j] -> setToggleButton (true);
      drive->m_btnList -> insert(j, drive->m_pbItem[j]);
      connect(drive->m_pbItem[j], SIGNAL(clicked()), this, SLOT(slotButtonBarClicked()));

      drive->hBoxBtnBar -> setStretchFactor(drive->m_pbItem[j], (int)(pi->dwP1000Disk/5));

      // flags
      strFlags = QString("");
      for (k=PED_PARTITION_FIRST_FLAG; k < PED_PARTITION_LAST_FLAG; k++)
      {	    
	 if ((pi->dwFlags & (1 << k)) != 0)
	 {
	    strTemp.sprintf("%s ", ped_partition_flag_get_name((PedPartitionFlag)k));
	    strFlags += strTemp;
	 }
      }

      if (pi->dwType == PED_PARTITION_FREESPACE)
	 img.setColor (2, QColor(0xFF, 0xFF, 0xFF).rgb());
      else
	 img.setColor (2, QColor(pi->cColorRed, pi->cColorGreen, pi->cColorBlue).rgb());
	    
      img.fill (2);
      QPixmap pm(img);

      drive->m_lviItem[j]  = new QListViewItem(drive->lv, strPos, strMinor);
      drive->m_lviItem[j] -> setText (2, pi->szPartName);
      drive->m_lviItem[j] -> setText (3, strFlags);
      drive->m_lviItem[j] -> setText (4, pi->szFilesysName);
      drive->m_lviItem[j] -> setText (5, strSize);
      drive->m_lviItem[j] -> setText (7, strType);
      drive->m_lviItem[j] -> setText (8, strGeomStart);
      drive->m_lviItem[j] -> setText (9, strGeomEnd);
      drive->m_lviItem[j] -> setText (10, strLabel);
      drive->m_lviItem[j] -> setText (13, strMountPoint);
      drive->m_lviItem[j] -> setPixmap(0, pm);
      if (getOptions()->bShowPartDetails)
      {
	 drive->m_lviItem[j] -> setText (6, strUsage);
	 drive->m_lviItem[j] -> setText (11, strUsageUsed);
	 drive->m_lviItem[j] -> setText (12, strUsageFree);
      }
   }
      
   // summary about hard disk
   if (getOptions()->bShowPartDetails)
   {
      QString strTotal, strUsed, strFree;
      QProgressBar *prog;
      QLabel *label1, *label2, *label3;

      qwSummaryTotal = m_doc->getDriveSize(nTabPos);
      prog = new QProgressBar(drive->hBoxSummary);
      label1 = new QLabel(drive->hBoxSummary);
      label2 = new QLabel(drive->hBoxSummary);
      label3 = new QLabel(drive->hBoxSummary);

      strTotal.sprintf("  Total: %s  ", formatSize(qwSummaryTotal, szTemp));
      strUsed.sprintf("  Used: %s  ", formatSize(qwSummaryUsed, szTemp));
      strFree.sprintf("  Free: %s  ", formatSize(qwSummaryTotal-qwSummaryUsed, szTemp));

      DWORD dwPct=(DWORD)((100LL * qwSummaryUsed) / qwSummaryTotal);
      prog -> setProgress(dwPct);
      label1 -> setText(strTotal);
      label2 -> setText(strUsed);
      label3 -> setText(strFree);

      drive->hBoxSummary -> setStretchFactor(prog, 2);
      drive->hBoxSummary -> setStretchFactor(label1, 3);
      drive->hBoxSummary -> setStretchFactor(label2, 3);
      drive->hBoxSummary -> setStretchFactor(label3, 3);
   }

   // add tab 
   insertTab((QWidget *)drive->vbox, m_doc->getDriveName(nTabPos), nTabPos);

   return 0;
}

//------------------------------------------------
int PiguicqtView::refreshDisk(int nDisk)
{
   // delete old tab
   removePage ((QWidget *)m_drive[nDisk]->vbox);

   // make new tab
   m_drive[nDisk] = new CDriveView;
   createTab(m_drive[nDisk], nDisk);

   // select first part in the disk
   setCurrentPart(nDisk, 0);

   return 0;
}

//------------------------------------------------
int PiguicqtView::setCurrentPart(int nDrive, int nPart)
{
   DWORD i;

   m_nDrive = nDrive;
   m_nPart = nPart;

   //pi_printf ("select Drive(%d) and Part(%d)\n", nDrive, nPart);
   setCurrentPage (nDrive);

   // select item in the list
   m_drive[nDrive]->lv -> setCurrentItem(m_drive[nDrive]->m_lviItem[nPart]);

   // unselect all buttons, except the good one
   for (i=0; i < m_drive[nDrive]->m_btnList->count(); i++)
      (*(m_drive[nDrive]->m_btnList))[i]->setOn((int)i == nPart);

   // gray / ungray actions
   //bUsedSpace = m_doc->isPartUsed(nDrive, nPart);
   getApp() -> grayActions();

   return 0;
}
 
//------------------------------------------------
void PiguicqtView::slotListViewClickedLeft(QListViewItem *lvi)
{
   int i, j;

   i = currentPageIndex();
   j = atoi(lvi->text(0).data());

   setCurrentPart(i, j);
}

//------------------------------------------------
void PiguicqtView::slotButtonBarClicked()
{
   int i, j;

   i = currentPageIndex();

   j = m_drive[i]->m_btnList -> findRef((CPiPushButton *)sender());
   
   setCurrentPart(i, j);
}

//------------------------------------------------
void PiguicqtView::mouseReleaseEvent(QMouseEvent *e)
{
   /*if (e->button() == QMouseEvent::RightButton)
      ((PiguicqtApp *)m_app)->getContextMenu()->exec(e->globalPos());*/
}

//------------------------------------------------
PiguicqtView::~PiguicqtView()
{
   //delete m_menu;
}

//------------------------------------------------
void PiguicqtView::slotDocumentChanged()
{
  //TODO update the view
}

//------------------------------------------------
void PiguicqtView::getCurrentDiskPart(int *nDisk, int *nPart)
{
   *nDisk = m_nDrive;
   *nPart = m_nPart;
}

