/***************************************************************************
                  piguictext-main.cpp  -  Text GUI for client
                             -------------------
    begin                : Thu Aug 30 16:32:34 CEST 2001
    copyright            : (C) 2001 by Francois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifdef HAVE_CONFIG_H
  #include <config.h>
#endif

#ifdef HAVE_GETOPT_H
  #include <getopt.h>
#endif

#ifdef HAVE_SYS_PARAM_H
  #include <sys/param.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>        
#include <sys/types.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "common.h"
#include "partimage.h"
#include "messages.h"
#include "netclient.h"
#include "netserver.h"
#include "args.h"

#include "interface_net.h"

// time
time_t g_timeBegin=0; // Beginning of main()
time_t g_timeStart=0; // Beginning of the copy, after confirmation
time_t g_timeEnd=0;   // End of the operation

FILE *g_fDebug; // debug file
FILE *g_fLocalDebug; // debug file
BOOL g_bSigInt = false;
BOOL g_bSigKill = false;

void catch_sigint(int signo);
int operationSave(CInterfaceNet *net, char *szDevice, char *szImageFile, COptions *options);
int operationRestore(CInterfaceNet *net, char *szDevice, char *szImageFile, COptions *options);
int operationConvert(CInterfaceNet *net, char *szOldImage, char *szNewImage, COptions *options);
BOOL checkForInterruption(CInterfaceNet *net);

// =======================================================
void closeDebugFiles()
{
  if (g_fDebug && (g_fDebug != stderr))
    fclose(g_fDebug);
}

// =======================================================
int main(int argc, char *argv[])
{
  COptions options;
  char szDevice[MAX_DEVICENAMELEN];
  char szImageFile[MAXPATHLEN];
  char szImageFile2[MAXPATHLEN];
  char szTemp[4096];
  char szAux[MAXPATHLEN+1];
  char szAux2[MAXPATHLEN+1];
  int nOperation;
  char szServerName[MAX_HOSTNAMESIZE];
  WORD wServerPort;
  CInterfaceNet *net;
  int nRes;

  BOOL bStartServer = false;

  // network = default values
  wServerPort = PARTIMAGE_GUI_PORT; //4890;
  SNPRINTF(szServerName, "127.0.0.1");

  // initialize options with defaults values
  initOptions(&options);

  // initialize language for i18n
  setlocale(LC_ALL, "");
  bindtextdomain (PACKAGE, LOCALEDIR);
  textdomain (PACKAGE);
  
  // read command line
  nRes = processCommandLine(argc, argv, &options, &bStartServer, szServerName,
     &wServerPort);
  if (nRes)
    return EXIT_FAILURE;

  // start the server
#if 0
  if (bStartServer)
    {
      switch(fork())
	{
	case -1:
	  perror("fork");
	  exit(-1);
	  
	case 0: // child -> server
	  system("../../client/partimage 1>/dev/null 2>/dev/null");
	  return 0;

	default: // current process
	  printf("fork ok\n");
	  sleep(1);
	}
    }

  // signal
  signal(SIGTERM, catch_sigint);
  signal(SIGINT, catch_sigint);
  //signal(SIGSEGV, catch_sigint); // segmentation fault
#endif

  // debug file
  if (options.dwDebugLevel)
    {
      SNPRINTF(szAux2, "%s_latest", PIGUICNEWT_LOG);
#ifdef APPEND_PID // append pid at the end of PARTIMAGE_LOG filename
      SNPRINTF(szAux, "%s_%d", PIGUICNEWT_LOG, getpid());
#else
      SNPRINTF(szAux, "%s", PIGUICNEWT_LOG);
#endif 
      unlink(szAux2);
      symlink(szAux, szAux2);
#ifdef O_NOFOLLOW
      g_fDebug = openFileDescriptorSecure(szAux, "a", O_WRONLY | O_CREAT | O_NOFOLLOW | O_TRUNC, S_IRUSR | S_IWUSR);
#else
      g_fDebug = openFileDescriptorSecure(szAux, "a", O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
#endif

      g_nDebugThreadMain = getpid();
      g_dwDebugLevel = options.dwDebugLevel;
      pthread_mutex_init(&g_mutexDebug, &g_mutexDebugAttr);

      if (!g_fDebug)
	{	
	  fprintf(stderr, i18n("Cannot open debug file.\n"));
	  g_fDebug = stderr;
	  //return EXIT_FAILURE;
	}
      else
	{
          showDebug(1, "%s: partimage-piguicnewt version %s (DebugLevel %lu used, "
		    "MainThread=%d)\n",PIGUICNEWT_LOG, VERSION, options.dwDebugLevel, 
		    getpid());
	  showDebug(1, "======================================================="
		    "================================\n\n\n");

	  formatCompilOptions(szTemp, sizeof(szTemp));
	  showDebug(1, "%s\n\n", szTemp);
	}
    }

  nOperation = -1;

  memset(szDevice, 0, MAX_DEVICENAMELEN);
  memset(szImageFile, 0, MAXPATHLEN);
  memset(szImageFile2, 0, MAXPATHLEN);

  // run operations if the command line is full
  if (argc - optind == 3) // commands with 2 parameters
    {
      showDebug(8, "full cmdline with 2 param\n"); 

      if (strcmp(argv[optind], "save")==0) // save
	nOperation = OPERATION_SAVE;
      else if (strcmp(argv[optind], "restore")==0) // restore
	nOperation = OPERATION_RESTORE;
      else if (strcmp(argv[optind], "imgconv")==0) // restore
	nOperation = OPERATION_IMGCONV;

      if ((nOperation == OPERATION_SAVE) || (nOperation == OPERATION_RESTORE))
	{
	  strncpy(szDevice, argv[optind+1], MAX_DEVICENAMELEN);
	  strncpy(szImageFile, argv[optind+2], MAXPATHLEN);
	}
      else // OPERATION_IMGCONV
	{
	  strncpy(szImageFile, argv[optind+1], MAXPATHLEN);
	  strncpy(szImageFile2, argv[optind+2], MAXPATHLEN);
	}
    }

  if (argc - optind == 2) // commands with 1 parameter
    {
      showDebug(8, "full cmdline with 1 param\n"); 
      strncpy(szImageFile, argv[optind+1], MAXPATHLEN);
      
      if (strcmp(argv[optind], "restmbr")==0) // restore an MBR
	nOperation = OPERATION_RESTMBR;
      else if (strcmp(argv[optind], "imginfo")==0) // infos about image
	nOperation = OPERATION_IMGINFO;
    }

  // if no command line or invalid action
  if (nOperation == -1)
    {
      printf("erreur ligne de commande\n");
      showDebug(8, "weird cmdline\n"); 
      exit(1);
    }
  
  // connect to pi-work server
  net = new CInterfaceNet(false);

  try { net->Connect(szServerName, wServerPort); }
  catch (CExceptions *excep)
    {
      fprintf(stderr, "cannot connect to server\n");
      fprintf(stderr, "->%u\n", excep->GetExcept());
      closeDebugFiles();
      return EXIT_FAILURE;
    }

  options.wOperation = (WORD)nOperation;

  showDebug(8, "ok for options\n");

// send first message, skip: we don't want to receive partition list
// since partition to use is get from command line
  net->set_message(MG1_SKIP);

// send parameters (image file, device, ...)
  net->set_param(options.wOperation, szDevice, szImageFile, szImageFile2);

  nRes = 0;
  switch(options.wOperation)
    {
    case OPERATION_SAVE:
      nRes = operationSave(net, szDevice, szImageFile, &options);
      break;

    case OPERATION_RESTORE:
      nRes = operationRestore(net, szDevice, szImageFile, &options);
      break;
      
    case OPERATION_RESTMBR:
      break;

    case OPERATION_IMGINFO:
      break;

    case OPERATION_IMGCONV:
      nRes = operationConvert(net, szImageFile, szImageFile2, &options);
      break;
    }

  // check options
  nRes = checkOptions(options, szDevice, szImageFile, szImageFile2);
  if (nRes == -1)
    {
      closeDebugFiles();
      return EXIT_FAILURE;
    }

  showDebug(8, "go on\n");

  closeDebugFiles();
  return EXIT_SUCCESS;
}

// =======================================================
int operationSave(CInterfaceNet *net, char *szDevice, char *szImageFile, COptions *options)
{
  BEGIN;

  CMessages msg;
  char szPartDescription[MAX_DESCRIPTION];
  char szTemp[4096];
  QWORD qwTemp;
  time_t timeTemp;

  memset(szPartDescription, 0, sizeof(szPartDescription));
  net->set_options(*options);
  net->set_description(szPartDescription); 

  // RUN operation (now, the GUI will wait for answers to this message)
  net->set_message(MG1_OPERATION_BEGIN);

  // wait for answers
  do
    {
      net->RecvMsg(&msg);	// read message

      showDebug(3, "RecvMsg()=%lu\n",(DWORD)msg.getCmd());
      switch(msg.getCmd())
	{
	case MG2_SET_PARTINFO_SAVE: // only at startup
	  net->RECVC(qwTemp); // qwPartSize
	  net->RecvString(szTemp, sizeof(szTemp)); // szFileSystem
	  break;
	  
	case MG2_UPDATE_PARTITION_SAVE: // many times during the operation
	  QWORD qwBlockSize, qwBlocksDone, qwBlocksTotal;
	  QWORD qwFreeSpace, qwImageSize;
	  net->Recv(&timeTemp, sizeof(timeTemp));
//	  net->Recv(&qwBlockSize, sizeof(qwBlockSize));
          qwBlockSize = 512LL;
	  net->Recv(&qwBlocksDone, sizeof(qwBlocksDone));
	  net->Recv(&qwBlocksTotal, sizeof(qwBlocksTotal));
          msg.setMessage(MG3_OKAY);
          net->SendMsg(&msg);
          break;

        case MG2_UPDATE_IMAGE_SAVE: // many times during the operation
	  net->RecvString(szTemp, sizeof(szTemp)); // szImageFile
	  net->Recv(&qwFreeSpace, sizeof(qwFreeSpace));
	  net->Recv(&qwImageSize, sizeof(qwImageSize));
          msg.setMessage(MG3_OKAY);
          net->SendMsg(&msg);
	  break;

        case MG2_WAIT_REQUIRED:
	  printf("volume was splitted\n");
	  net->set_message(MG3_OKAY);
	  break;

	case MG2_GET_NEWPATH:
	  printf("new path required for splitting -> /tmp");
	  net->set_message(MG3_OKAY, 5);
          net->Send("/tmp", 5);
	  break;

        case MG2_ERR_COMPRESSFAILED:
          printf("erreur lors de la compression->send SKIP\n");
          net->set_message(MG1_SKIP);
	  break;

        case MG2_ERR_NOMEM:
          printf("server ran out of memory\n");
          net->set_message(MG1_QUIT);
	  break;

        default: 
	  printf("message recu: %ld\n", msg.getCmd());
	  break;
	}

    } while (msg.getCmd() != MG2_END);
  printf("pi-work finished\n");
  net->set_message(MG3_OKAY);

  RETURN_int(0);
}

// =======================================================
int operationRestore(CInterfaceNet *net, char *szDevice, char *szImageFile, COptions *options)
{
  BEGIN;

  CMessages msg;
  QWORD qwTemp;
  BOOL bTemp;
  time_t timeTemp;
  char szTemp[4096];
  char szTemp2[4096];
  char szTemp3[4096];

  // initialize interface
  net->set_options(*options);
  net->set_description(""); 

  // RUN operation (now, the GUI will wait for answers to this message)
  net->set_message(MG1_OPERATION_BEGIN);

  // --------- initialize info window -------------
  // wait for answers
  do
    {
      printf("attente message\n");
      net->RecvMsg(&msg);	// read message

      showDebug(3, "RecvMsg()=%lu\n",(DWORD)msg.getCmd());
      switch(msg.getCmd())
	{
	case MG2_SET_PARTINFO_RESTORE: // only at startup
	  net->RECVC(qwTemp); // qwPartSize
	  net->RecvString(szTemp, sizeof(szTemp)); // szImageFile
	  net->RecvString(szTemp2, sizeof(szTemp2)); // szOriginalDevice
	  net->RecvString(szTemp3, sizeof(szTemp3)); // szFileSystem
	  break;

	case MG2_UPDATEWIN_RESTORE:  // many times during the operation
	  QWORD qwBlockSize, qwBlocksDone, qwBlocksTotal, qwBlocksUsed;
	  net->RECVC(timeTemp); // g_timeStart
	  net->RECVC(qwBlockSize);
	  net->RECVC(qwBlocksDone);
	  net->RECVC(qwBlocksUsed);
	  net->RECVC(qwBlocksTotal);
	  net->RECVC(bTemp);
	  break;

        case MG2_ERR_DECOMPRESSFAILED:
          printf("erreur lors de la decompression->send SKIP\n");
          net->set_message(MG1_SKIP);
	  break;

        case MG2_ERR_NOMEM:
          printf("server ran out of memory\n");
          net->set_message(MG1_QUIT);
	  break;
 
        case MG2_ERR_CRCMISMATCH:
          printf("crc check failed -> ignored\n");
          net->set_message(MG1_IGNORE);
          break;

        default: 
	  printf("message recu: %ld\n", msg.getCmd());
	  break;
	}

    } while (msg.getCmd() != MG2_END);

  RETURN_int(0);
}

// =======================================================
int operationConvert(CInterfaceNet *net, char *szOldImage, char *szNewImage, COptions *options)
{
  CMessages msg;
  
  // wait for answers
  do
    {
      net->RecvMsg(&msg);	// read message

      showDebug(3, "RecvMsg()=%lu\n",(DWORD)msg.getCmd());
      switch(msg.getCmd())
      {
	case MG2_UPDATEWIN_IMGCONV:
	  break;
      }

  } while (msg.getCmd() != MG2_END);	      

  RETURN_int(0);
}

// =======================================================
int Mg2_UpdateWin_Convert(int nClient, CNetServer *net, QWORD qwBlockSize, QWORD qwBlocksDone, QWORD qwBlocksTotal, char *szImageFile, QWORD qwFreeSpace, QWORD qwImageSize)
{
/*
  CMessages msg;

  msg.setMessage(MG2_UPDATEWIN_SAVE); net->SendMsg(nClient, &msg);
  net->SENDS(nClient, g_timeStart);
  net->SENDS(nClient, qwBlockSize);
  net->SENDS(nClient, qwBlocksDone);
  net->SENDS(nClient, qwBlocksTotal);
  net->SendString(nClient, szImageFile, MAXPATHLEN);
  net->SENDS(nClient, qwFreeSpace);
  net->SENDS(nClient, qwImageSize);
  net->RecvMsg(nClient, &msg);
  return msg.getCmd(); // return(MG3_INTERRUPTED || MG3_OKAY)
*/
   return 0;
}


// =======================================================
void catch_sigint(int signo)
{
  if (signo == SIGTERM)
    {
      g_bSigKill = true;
      exit(0);
    }
  else if (signo == SIGINT)
    {
      g_bSigInt = true;
    }
  /*else if (signo == SIGSEGV)
    {
      delete g_interface;
      fprintf(stderr, i18n("Segmentation fault. Please report the bug and send the /var/log/partimage-debug.log file to authors\n"));
      exit(0);      
    }*/
}

// =======================================================
BOOL checkForInterruption(CInterfaceNet *net)
{
  if (g_bSigInt == true)
    {
      net->set_message(MG3_INTERRUPTED);
      RETURN_int(true);
    }
  
  net->set_message(MG3_OKAY);
  RETURN_int(false);
}

