/***************************************************************************
                          buffer.h  -  description
                             -------------------
    begin                : Web Dec 13 2000
    copyright            : (C) 2000 by Franck Ladurelle
    email                : ladurelf@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef _BUFFER_H_
#define _BUFFER_H_

#include <pthread.h>

#include "global.h"
#include "misc.h"

class CImage;

// MULTITHREAD
extern int g_nThreadReadState; // is it the end of the operation
extern int g_nThreadWriteState; // is it the end of the operation
extern pthread_t g_threadReadBuffer;
extern pthread_t g_threadWriteBuffer;
extern int g_nThreadReadError;
extern int g_nThreadWriteError;
extern BOOL g_bSpaceOnDisk;

// for time report only
extern DWORD g_dwTimeThreadUser;
extern DWORD g_dwTimeThreadSys;

// function (work thread) which will write the buffer 
void procWriteBufferToImage(void *ptr);

// function (work thread) to fill the buffer
void procReadBufferFromImage(void *ptr); // Thread which will read data from image file

#define KIND_HEADER     1
#define KIND_BLOCKS     2
#define KIND_EOF        3
#define KIND_VOLHEAD    4 
#define KIND_MAINTAIL   5
#define KIND_MAINHEAD   6
#define KIND_LOCALHEAD  7

// ================================================
struct CChunkHeader
{
  QWORD qwPos; // number of the first block of the chunk
  DWORD dwCRC; // CRC of the CHECK_FREQUENCY blocks
  DWORD dwCompress;
  DWORD dwChunkCompressedSize;
  WORD  wKind;
  char cMagic[3]; // must be 'C','H','K'
};

struct CSenderHeader
{
  DWORD dwDataSize;
  DWORD dwCompression;
  WORD wKind;
};


#define BUFFER_SIZE           131072
#define BUFFER_BLOCK_SIZE     14336
#define ENDVOL_MARGIN         65536
#define SOME_SPACE            (1*1024*1024) // 1 MB
#define MAX_CHUNKSIZE         (1024*1024) // 1 MB

// Must be a 65 KB multiple size
#define CHUNKSIZE_NONE (65536) // 65 KB
#define CHUNKSIZE_GZIP (65536) // 65 KB
#define CHUNKSIZE_LZO (65536*4) // 256 KB
#define CHUNKSIZE_BZIP2 (65536*14) // 900 KB = default bzip2 block size

#define THREAD_RUNNING   0
#define THREAD_ASKEXIT   1
#define THREAD_FINISHED  2
#define THREAD_EOF       3

// =======================================================
struct CParam
{
  COptions options;
  CImage *image;
  int nFd;
  BOOL bSendChunkHeader;
};

#endif // _BUFFER_H_
