/***************************************************************************
                          imagefile.h  -  description
                             -------------------
    begin                : Mon Aug 14 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef IMAGEFILE_H
#define IMAGEFILE_H

#include "global.h"
#include "image_disk.h"
#include "image_net.h"
#include "buffer.h"

class CSavingWindow;
class CRestoringWindow;
class CVolumeHeader;

// ================================================
class CImage
{
 private:
  CImageBase *m_cid;
  CSavingWindow *m_guiSave;
  CRestoringWindow *m_guiRestore;
  QWORD m_qwCRC;
  QWORD m_qwIdentificator;
  DWORD m_dwVolumeNumber;
  COptions m_options;
  BOOL m_bSendChunkHeader;
  FILE *m_fImageFile;
  int m_nFdImage;

  // auto-mount options
  BOOL m_bIsMounted;
  char m_szMountDevice[MAXPATHLEN];
  char m_szMountPoint[MAXPATHLEN];
  char m_szMountFS[MAXPATHLEN];

 public:
  CImage(COptions * options, BOOL bSendChunkHeader); 
  ~CImage();
  
  QWORD get_qwIdentificator() {return m_qwIdentificator;}
  QWORD get_qwCurrentVolumeSize();
  DWORD get_dwVolumeNumber();
  BOOL  get_bIsOpened();
  char * get_szImageFilename();
  char * get_szOriginalFilename();
  char * get_szPath();
  QWORD get_qwCRC() { return m_qwCRC; }
  void get_qwFreeSpace(QWORD *qwFreeSpace) {m_cid->getFreeSpace(qwFreeSpace);}
  CImageBase *getImageDisk() {return m_cid;}
  
  void set_dwVolumeNumber(DWORD n);
  void set_bIsOpened(BOOL n);
  void set_szPath(char * str);
  void set_szImageFilename(char * str);
  void set_szOriginalFilename(char * str);
  void set_szImageFilename(char * str, int n);
  void set_szOriginalFilename(char * str, int n);
  void set_qwCRC(QWORD n) {m_qwCRC = n;}
  void set_qwIdentificator(QWORD n);
  void set_options(COptions * options);

 public: // read/write
  void write(WORD wKind, void *buf, DWORD dwLength, BOOL bUpdateCRC); // return 0 if error, or written bytes
  DWORD read(char *buf, DWORD dwLength, BOOL bUpdateCRC);

  void openReadingFd();
  void openWritingFd();
  int openWritingFdDisk();

  void writeTail();
  void readTail();
  
  void writeSplit();
  int getCompressionLevelForImage(char *szFilename);
  
  // open/close  
  void openWriting();
  void openReading(CVolumeHeader *vh = NULL);
  void closeReading(BOOL bForceExit = false);
  void closeWriting();

  // magic strings
  void readAndCheckMagic(char *szMagicString);
  void writeMagic(WORD wKind, char *szMagicString);
  
  // CRCs
  void writeCRC(WORD wKind, char *cData, DWORD dwDataLen);
  void readAndCheckCRC(char *cData, DWORD dwDataLen);

  // GUI
  void setGuiSave(CSavingWindow *gui) {m_guiSave = gui;}
  void setGuiRestore(CRestoringWindow *gui) {m_guiRestore = gui;}
  
 private:
  void splitChangeImageFileWrite();
  void splitChangeImageFileRead();

  int mountImageLocation();
  int umountImageLocation(); 

  void openWritingVolhead();
  void openReadingVolhead(CVolumeHeader *vh);
};

#endif // IMAGEFILE_H
