/***************************************************************************
                          common.cpp  -  procedures shared between client&server
                             -------------------
    begin                : Sun Jul 16 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
  #include <config.h>
#endif

#ifdef HAVE_MNTENT_H
  #include <mntent.h>
#else
  #warning "mntent.h not found"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <pthread.h>
#include <dirent.h>
#include <sys/types.h>

#ifdef HAVE_SYS_PARAM_H
  #include <sys/param.h>
#endif
#ifdef HAVE_SYS_MOUNT_H
  #include <sys/mount.h>
#endif
#ifdef HAVE_SYS_STATFS_H
  #include <sys/statfs.h>
#endif

#ifdef LEVELSUPPORT_GZIP
#include <zlib.h>
#endif // LEVELSUPPORT_GZIP // gzip compression
#ifdef LEVELSUPPORT_BZIP2
#include <bzlib.h>
#endif // LEVELSUPPORT_BZIP2 // bzip2 compression
#include <string.h>

#include "common.h"
#include "global.h"
#include "buffer.h"

DWORD g_dwDebugLevel;
int g_nDebugThreadMain;
int g_nDebugThreadBufferR;
int g_nDebugThreadBufferW;
pthread_mutex_t g_mutexDebug; // lock the buffer while working inside it
pthread_mutexattr_t g_mutexDebugAttr;
DWORD g_pidType=0;

// =======================================================
#define addText(format, args...) \
{ \
  nLenRest = nMaxLen - strlen(szDest); \
  snprintf(szTemp, sizeof(szTemp), format, ##args); \
  strncat(szDest, szTemp, nLenRest); \
}

void formatCompilOptions(char *szDest, int nMaxLen)
{
  int nLenRest;
  char szTemp[8192];

  memset(szDest, 0, nMaxLen);

  addText(i18n("==== Partition Image: compilation options used ====\n"));
  addText(i18n("\n* File Systems:\n"));

#ifdef JFS_SUPPORT
  addText(i18n("  - JFS File system support (IBM Aix)\n"));
#endif

#ifdef XFS_SUPPORT
  addText(i18n("  - XFS File system support (SGI Irix)\n"));
#endif

#ifdef LIBRARY_EXT2FS
  addText(i18n("  - Ext2fs support: based on libext2fs library\n"));
#else
  addText(i18n("  - Ext2fs support: internal code\n"));
#endif

  addText(i18n("\n* Debug options:\n"));

#ifdef PARTIMAGE_LOCAL_DEBUG
  addText(i18n("  - PARTIMAGE_LOCAL_DEBUG=%s\n"), PARTIMAGE_LOCAL_DEBUG);
#endif
#ifdef PARTIMAGE_LOG
  addText(i18n("  - PARTIMAGE_LOG=%s\n"), PARTIMAGE_LOG);
#endif
#ifdef PARTIMAGED_LOG
  addText(i18n("  - PARTIMAGED_LOG=%s\n"), PARTIMAGED_LOG);
#endif

  addText(i18n("  - DEFAULT_DEBUG_LEVEL=%d\n"), DEFAULT_DEBUG_LEVEL);
#ifdef APPEND_PID
  addText(i18n("  - APPEND_PID=%s\n"), (APPEND_PID ? i18n("true") :  i18n("false")));
#endif


  addText(i18n("\n* Supported compression levels:\n"));

#ifdef LEVELSUPPORT_GZIP
  addText(i18n("  - LEVELSUPPORT_GZIP=true\n"));
#endif
#ifdef LEVELSUPPORT_BZIP2
  addText(i18n("  - LEVELSUPPORT_BZIP2=true\n"));
#endif
#ifdef LEVELSUPPORT_LZO
  addText(i18n("  - LEVELSUPPORT_LZO=true\n"));
#endif


  addText(i18n("\n* Other options:\n"));
  addText(i18n("  - PARTIMAGED_USERS=%s\n"), PARTIMAGED_USERS);
  addText(i18n("  - CURRENT_IMAGE_FORMAT=%s\n\n"), CURRENT_IMAGE_FORMAT);
#ifdef MUST_LOGIN
  addText(i18n("  - USERS MUST LOGIN\n"));
#else
  addText(i18n("  - USERS DON'T LOGIN\n"));
#endif
#ifdef HAVE_SSL
  addText(i18n("  - SSL ENABLED\n"));
#else
  addText(i18n("  - SSL DISABLED\n"));
#endif
#ifdef MUST_CHEUID
  addText(i18n("  - CHEUID ENABLED\n"));
#else
  addText(i18n("  - CHEUID DISABLED\n"));
#endif
}

// =======================================================
#ifdef LEVELSUPPORT_GZIP
char *formatGzipError(int nError, char *szTemp, int nMaxLen)
{
  switch (nError)
    {
    case Z_OK:
      snprintf(szTemp, nMaxLen, "GZIP nError = %d: Z_OK\n", nError);
      break;
	  
    case Z_STREAM_END:
      snprintf(szTemp, nMaxLen, "GZIP = %d: Z_STREAM_END\n", nError);
      break;

    case Z_ERRNO:
      snprintf(szTemp, nMaxLen, "GZIP: nError = %d: Z_ERRNO and errno=%d=[%s]\n", nError, errno, strerror(errno));
      break;	  
	  
    case Z_NEED_DICT:
      snprintf(szTemp, nMaxLen, "GZIP: nError = %d: Z_NEED_DICT\n", nError);
      break;

    case Z_STREAM_ERROR:
      snprintf(szTemp, nMaxLen, "GZIP: nError = %d: Z_STREAM_ERROR\n", nError);
      break;

    case Z_DATA_ERROR:
      snprintf(szTemp, nMaxLen, "GZIP: nError = %d: Z_DATA_ERROR\n", nError);
      break;

    case Z_MEM_ERROR:
      snprintf(szTemp, nMaxLen, "GZIP: nError = %d: Z_MEM_ERROR\n", nError);
      break;

    case Z_BUF_ERROR:
      snprintf(szTemp, nMaxLen, "GZIP: nError = %d: Z_BUF_ERROR\n", nError);
      break;
	  
    default:
      snprintf(szTemp, nMaxLen, "GZIP: nError = %d\n", nError);
      break;
    };

  return szTemp;
}
#endif // LEVELSUPPORT_GZIP

// =======================================================
#ifdef LEVELSUPPORT_BZIP2
char *formatBzip2Error(int nError, char *szTemp, int nMaxLen)
{
  switch (nError)
    {
    case BZ_OK:
      snprintf(szTemp, nMaxLen, "BZIP2 nError = %d: Z_OK\n", nError);
      break;
	  
    case BZ_SEQUENCE_ERROR:
      snprintf(szTemp, nMaxLen, "BZIP2 = %d: BZ_SEQUENCE_ERROR\n", nError);
      break;

    case BZ_PARAM_ERROR:
      snprintf(szTemp, nMaxLen, "BZIP2: nError = %d: BZ_PARAM_ERROR\n", nError);
      break;

    case BZ_MEM_ERROR:
      snprintf(szTemp, nMaxLen, "BZIP2: nError = %d: BZ_MEM_ERROR\n", nError);
      break;

    case BZ_DATA_ERROR:
      snprintf(szTemp, nMaxLen, "BZIP2: nError = %d: BZ_DATA_ERROR\n", nError);
      break;

    case BZ_DATA_ERROR_MAGIC:
      snprintf(szTemp, nMaxLen, "BZIP2: nError = %d: BZ_DATA_ERROR_MAGIC\n", nError);
      break;

    case BZ_IO_ERROR:
      snprintf(szTemp, nMaxLen, "BZIP2: nError = %d: BZ_IO_ERROR\n", nError);
      break;

    case BZ_UNEXPECTED_EOF:
      snprintf(szTemp, nMaxLen, "BZIP2: nError = %d: BZ_UNEXPECTED_EOF\n", nError);
      break;

    case BZ_OUTBUFF_FULL:
      snprintf(szTemp, nMaxLen, "BZIP2: nError = %d: BZ_OUTBUFF_FULL\n", nError);
      break;

    case BZ_CONFIG_ERROR:
      snprintf(szTemp, nMaxLen, "BZIP2: nError = %d: BZ_CONFIG_ERROR\n", nError);
      break;

    default:
      snprintf(szTemp, nMaxLen, "BZIP2: nError = %d\n", nError);
      break;
    };

  return szTemp;
}
#endif // LEVELSUPPORT_BZIP2

// =======================================================
void pi_printf2(const char *szFile, const char *szFunction, int nLine, const char *fmt, ...)
{
   //va_list args;

   if (g_pidType == 1) // server
      textcolor2(BRIGHT, RED);      
   else //client
      textcolor2(BRIGHT, WHITE);      
      
#if 0 // Q_WS_X11 // don't printf under FrameBuffer Q_WS_QWS 
   va_start(args, fmt);
   vprintf(fmt, args);
   va_end(args);
#endif

   textcolor2(RESET, WHITE);      
}

// =======================================================
void showDebugMsg(FILE *fDebug, DWORD dwDebugLevel, const char *szFile,
		  const char *szFunction, int nLine, const char *fmt, ...)
{
  pthread_mutex_lock(&g_mutexDebug);

  va_list args;
  DWORD i;

  if (dwDebugLevel <= g_dwDebugLevel)
  {
    // add tabs to distinguish log levels
    for (i = 1; i < dwDebugLevel; i++)
      fprintf(fDebug, "\t");
    
    va_start(args, fmt);
    
    if (getpid() == g_nDebugThreadMain) 
      fprintf(fDebug, "[Main] %s->%s#%d: ", szFile, szFunction, nLine);
    else if (getpid() == g_nDebugThreadBufferR) 
      fprintf(fDebug, "[BufferR] %s->%s#%d: ", szFile, szFunction, nLine);
    else if (getpid() == g_nDebugThreadBufferW) 
      fprintf(fDebug, "[BufferW] %s->%s#%d: ", szFile, szFunction, nLine);
    else
      fprintf(fDebug, "[TH=%d] %s->%s#%d: ", getpid(), szFile, szFunction, nLine);
    
    vfprintf(fDebug, fmt, args);

    // do not slow down the progran if standard debug level
    // must be enabled: if no flush, the log won't be up-to-date if there
    // is a segfault
    //if (g_dwDebugLevel != 1)
      fflush(fDebug);
    
    va_end(args);
  }
  pthread_mutex_unlock(&g_mutexDebug);
}

// =======================================================
char *formatTime2(DWORD dwSec, char *szBuf, int nMaxLen)
{
  int nHour, nMin, nSec;

  nSec = dwSec % 60;
  dwSec -= nSec; dwSec /= 60; // now, dwSec = dwMinutes
  nMin = dwSec % 60;
  dwSec -= nMin; dwSec /= 60; // now, dwSec = dwHours
  nHour = dwSec;
  //nHour = (dwSec - (dwSec % 3600)) / 60;

  *szBuf = 0;
  if (nHour)
    snprintf(szBuf, nMaxLen, "%dh:", nHour);
  if (nMin || nHour)
    snprintf(szBuf+strlen(szBuf), nMaxLen-strlen(szBuf), "%dm:", nMin);
  snprintf(szBuf+strlen(szBuf), nMaxLen-strlen(szBuf), "%dsec", nSec);
  
  return szBuf;
}

// =======================================================
char *formatKind2(WORD wKind, char *szText, int nMaxLen)
{	
  switch(wKind)
    {
    case KIND_HEADER:
      snprintf(szText, nMaxLen, "KIND_HEADER");
      break;

    case KIND_BLOCKS:
      snprintf(szText, nMaxLen, "KIND_BLOCKS");
      break;

    case KIND_EOF:
      snprintf(szText, nMaxLen, "KIND_EOF");
      break;

    case KIND_VOLHEAD:
      snprintf(szText, nMaxLen, "KIND_VOLHEAD");
      break;

    case KIND_MAINTAIL:
      snprintf(szText, nMaxLen, "KIND_MAINTAIL");
      break;

    case KIND_MAINHEAD:
      snprintf(szText, nMaxLen, "KIND_MAINHEAD");
      break;

    case KIND_LOCALHEAD:
      snprintf(szText, nMaxLen, "KIND_LOCALHEAD");
      break;

    default:
      snprintf(szText, nMaxLen, "KIND_unknown");
      break;

    }

  return szText;
}

// =======================================================
char *formatSize2(QWORD qwSize, char *szText, int nMaxLen)
{	
  double dSize;
  
  QWORD llKiloB = ((QWORD) 1024);
  QWORD llMegaB = ((QWORD) 1024)*((QWORD) 1024);
  QWORD llGigaB = ((QWORD) 1024)*((QWORD) 1024)*((QWORD) 1024);
  QWORD llTeraB = ((QWORD) 1024)*((QWORD) 1024)*((QWORD) 1024)*((QWORD) 1024);
  
  if (qwSize < llKiloB) // In Bytes
    {
      snprintf(szText, nMaxLen, i18n("%lld bytes"), qwSize);
    }
  else if (qwSize < llMegaB) // In KiloBytes
    {
      dSize = ((double) qwSize) / ((double) 1024.0);
      snprintf(szText, nMaxLen, i18n("%.2f KB"), (float) dSize);
    }
  else if (qwSize < llGigaB) // In MegaBytes
    {
      dSize = ((double) qwSize) / ((double) (1024.0 * 1024.0));
      snprintf(szText, nMaxLen, i18n("%.2f MB"), (float) dSize);
    }
  else if (qwSize < llTeraB)// In GigaBytes
    {
      dSize = ((double) qwSize) / ((double) (1024.0 * 1024.0 * 1024.0));
      snprintf(szText, nMaxLen, i18n("%.2f GB"), (float) dSize);
    }
  else // In TeraBytes
    {
      dSize = ((double) qwSize)/((double) (1024.0 * 1024.0 * 1024.0 * 1024.0));
      snprintf(szText, nMaxLen, i18n("%.2f TB"), (float) dSize);
    }
  
  return szText;
}

// =======================================================
char *formatCompress2(DWORD dwCompression, char *szText, int nMaxLen)
{	
  switch (dwCompression)
    {
    case COMPRESS_NONE:
      snprintf(szText, nMaxLen, i18n("None"));
      break;
    case COMPRESS_GZIP:
      snprintf(szText, nMaxLen, "gzip");
      break;
    case COMPRESS_BZIP2:
      snprintf(szText, nMaxLen, "bzip2");
      break;
    case COMPRESS_LZO:
      snprintf(szText, nMaxLen, "lzo");
      break;
    default:
      memset(szText, 0, nMaxLen);
    }
  
  return szText;
}

// =======================================================
// return true if this is a stable version (x.y.z with y even)
//        false           an unstable                  y odd     
BOOL isVersionStable(const char *szVersion)
{
  int i;
  char szMinor[512];

  memset(szMinor, 0, sizeof(szMinor));
  for (; *szVersion && *szVersion != '.'; szVersion++);
  szVersion++; // skip the '.'
  for (i=0; szVersion[i] && szVersion[i] != '.'; i++)
    szMinor[i] = szVersion[i]; // copy the minor

  return ((atoi(szMinor) % 2) == 0);
}

// =======================================================
// avoid RACE CONDITIONS security problems
FILE *openFileDescriptorSecure(char *szFilename, char *szMode, int nFlags, mode_t mode)
{
  int nFd;
  FILE *fStream;
#ifdef HAVE_FSTAT64
  struct stat64 fStat;
#else
  struct stat fStat;
#endif

  // init
  errno = 0;
#ifdef OS_LINUX
  nFlags |= O_LARGEFILE;
#endif

  // open file descriptor
  nFd = open(szFilename, nFlags, mode);
  if (nFd == -1)
    return NULL;

#ifdef HAVE_FSTAT64
  if (fstat64(nFd, &fStat) == -1)
    return NULL;
#else
  if (fstat(nFd, &fStat) == -1)
    return NULL;
#endif

  if (!S_ISREG(fStat.st_mode)) // not a regular file
    return NULL;

  fStream = fdopen(nFd, szMode);
  if (fStream == NULL)
    return NULL;

  return fStream;
}

// =======================================================
int isMounted(const char *szDevice, char *szMountPoint)
{
  FILE *f;
  struct mntent *mnt;
  BOOL bMounted;
  char szBuf1 [MAXPATHLEN+1];
  char szBuf2 [MAXPATHLEN+1];
  char *ptrDevice;

  // init
  bMounted = false;
  memset(szBuf1, 0, MAXPATHLEN);
  memset(szBuf2, 0, MAXPATHLEN);

  if ((f = setmntent (MOUNTED, "r")) == 0)
    RETURN_int(-1);

  // relative path -> full path [result in ptrDevice]
  ptrDevice = (char*)szDevice;
  if (realpath(ptrDevice, szBuf1) != NULL)
    ptrDevice = szBuf1;

  // symbolic link -> regular device file [result in ptrDevice]
  if (readlink (ptrDevice, szBuf2, MAXPATHLEN-1) != -1)
    ptrDevice = szBuf2;

  // loop: compare device with each mounted device
  while (((mnt = getmntent (f)) != 0) && (!bMounted))
    {
      if (strcmp(ptrDevice, mnt->mnt_fsname) == 0)
	{
	  bMounted = true; 
	  strcpy(szMountPoint, mnt->mnt_dir); 
	} 
    }

  endmntent (f);
  
  return bMounted;
}

// =======================================================
int bigFseek(FILE *fStream, QWORD qwOffset) // SEEK_SET forced
{
  const long lGigaByte = (QWORD) (1024L * 1024L * 1024L);
  int nRes;
  
  // go to the beginning
  nRes = fseek(fStream, 0L, SEEK_SET);
  
  while (qwOffset > 0)
    {
      //showDebug(1, "big_fseek: qwOffset=%llu\n", qwOffset);
      
      if (qwOffset >= (QWORD)lGigaByte)
	{
	  nRes = fseek(fStream, (long)lGigaByte, SEEK_CUR);
	  //showDebug(1, "big_fseek: big seek\n");
	  if (nRes == -1)
	    RETURN_int(-1);
	  qwOffset -= ((QWORD)lGigaByte);	
	}
      else
	{
	  //showDebug(1, "big_fseek: small seek: %llu\n", qwOffset);
	  nRes = fseek(fStream, (long)qwOffset, SEEK_CUR);
	  return nRes;
	}
    }
  
  return 0;
}

// =======================================================
QWORD getPartitionSize(const char *szDevice)
{
  QWORD qwSize;
  long lSize;
  int fdDevice;
  
  qwSize = 0;
  
  fdDevice = open(szDevice, O_RDONLY);
  if (fdDevice < 0)
    return 0LL;
 
  if (ioctl(fdDevice, BLKGETSIZE, &lSize) >= 0)
    qwSize = ((QWORD) 512) * ((QWORD) lSize);
  
  close(fdDevice);
  return qwSize;
}

//------------------------------------------------
BOOL doesFileExists2(char *szPath)
{
#ifdef HAVE_FSTAT64
  struct stat64 fStat;
#else
  struct stat fStat;
#endif
  int nFd;
  int nRes;

  errno = 0;
#ifdef HAVE_OPEN64
  nFd = open64(szPath, O_RDONLY | O_NOFOLLOW | O_LARGEFILE);
#else
  nFd = open(szPath, O_RDONLY | O_NOFOLLOW | O_LARGEFILE);
#endif
  //fStream = fopen(szPath, "rb");
  //if (!fStream) // error
  if (nFd == -1)
    {
      RETURN_int (errno != ENOENT);
    }
  else // success
    {
#ifdef HAVE_FSTAT64
      nRes = fstat64(nFd, &fStat);
#else
      nRes = fstat(nFd, &fStat);
#endif

      showDebug(3, "fstat() returned %d\n", nRes);
      if (nRes == -1)
	return true; // for large files
      //fclose(fStream);
      ::close(nFd);
      showDebug(3, "fStat.st_mode=%ld\n",(long)fStat.st_mode);
      RETURN_int(S_ISREG(fStat.st_mode));
    }
}

//------------------------------------------------
void textcolor(int attr, int fg, int bg)
{
  char command[130]; /* Command is the control command to the terminal */
  snprintf(command, sizeof(command), "%c[%d;%d;%dm", 0x1B, attr, fg + 30, bg + 40);
  printf("%s", command);
}

//------------------------------------------------
void textcolor2(int attr, int fg)
{
  char command[130]; /* Command is the control command to the terminal */
  snprintf(command, sizeof(command), "%c[%d;%d;%dm", 0x1B, attr, fg + 30, BLACK + 40);
  printf("%s", command);
}

//------------------------------------------------
int convertDevfsNameToShortName(const char *szDevfs, char *szShort, int nMaxShort)
{
   DIR *fdDir;
   struct dirent *dire;
   struct stat st;
   int nRes;
   char szTemp[MAXPATHLEN];
   char szTemp1[MAXPATHLEN];
   char szTemp2[MAXPATHLEN];
   DWORD i, dwSlash;

   // in case of error, use this file:
   snprintf(szShort, nMaxShort, "%s", szDevfs);

   for (i=0, dwSlash=0; szDevfs[i]; i++)
      if (szDevfs[i] == '/')
	 dwSlash++;
   
   //printf ("Slash(%s)=%d\n", szDevfs, (int)dwSlash);

   // exit if not a devfs name
   if (dwSlash <= 2)
      return 0;

   fdDir = opendir("/dev/");
   if (!fdDir)
   {
      fprintf (stderr, "Error 001\n");
      return -1;
   }
   
   while ((dire = readdir(fdDir)) != NULL)
   {
      // stats
      snprintf(szTemp1, sizeof(szTemp1), "/dev/%s", dire->d_name);
      nRes = lstat(szTemp1, &st);
	 
      // if this is a sym link
      if ((nRes == 0) && S_ISLNK(st.st_mode))
      {
	 memset(szTemp, 0, sizeof(szTemp));
	 nRes = readlink(szTemp1, szTemp, sizeof(szTemp));
	 
	 memset(szTemp2, 0, sizeof(szTemp2));
	 snprintf(szTemp2, sizeof(szTemp2), "/dev/%s", szTemp);
	 
	 if (strcmp(szDevfs, szTemp2) == 0)
	 {
	    memset(szShort, 0, nMaxShort);
	    snprintf(szShort, nMaxShort, "%s", szTemp1);
	    closedir(fdDir);
	    return 0;
	 }
      }
   }
   
   closedir(fdDir);

   return 0;
}

