
/***************************************************************************
                          common.h  -  description
                             -------------------
    begin                : Sun Jul 16 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef COMMON_H
#define COMMON_H

#include <stdarg.h>
#include <pthread.h>

#include <sys/param.h>

#include "global.h"
#include "format.h"

// debug
extern DWORD g_dwDebugLevel;
extern int g_nDebugThreadMain;
extern int g_nDebugThreadBufferW;
extern int g_nDebugThreadBufferR;
extern pthread_mutex_t g_mutexDebug; // lock the buffer while working inside it
extern pthread_mutexattr_t g_mutexDebugAttr;
extern DWORD g_pidType;

int isMounted(const char *szDevice, char *szMountPoint);
BOOL isVersionStable(const char *szVersion);
BOOL doesFileExists2(char *szPath);

void formatCompilOptions(char *szDest, int nMaxLen);
FILE *openFileDescriptorSecure(char *szFilename, char *szMode, int nFlags, mode_t mode);

#ifdef LEVELSUPPORT_GZIP
char *formatGzipError(int nError, char *szTemp, int nMaxLen);
#endif // LEVELSUPPORT_GZIP

#ifdef LEVELSUPPORT_BZIP2
char *formatBzip2Error(int nError, char *szTemp, int nMaxLen);
#endif // LEVELSUPPORT_BZIP2

char *formatCompress2(DWORD dwCompression, char *szText, int nMaxLen);
#define formatCompress(dwCompression, szBuf) formatCompress2(dwCompression, szBuf, sizeof(szBuf)) 

int bigFseek(FILE *fStream, QWORD qwOffset); // SEEK_SET forced

QWORD getPartitionSize(const char *szDevice);

//void checkForEscape();

/** Show formated advanced debug informations in the log file (/var/log/partimage-debug.log in most cases)
 *  The debug level allow to tell if the message must be written in a normal use, or only in a debug session
 *  This function must not be used directly: use showDebug() instead.
 *  nDebug Level:
 *  + 0 -> no message in the log file
 *  + 1 -> important messages in the log file (default) good for users
 *  + 2 -> many messages in the log file: good for developers
 *  + 3 -> all messages in the log file: good for debugging
*/
void showDebugMsg(FILE *fDebug, DWORD dwDebugLevel, const char *szFile, const char *szFunction, int nLine, const char *fmt, ...);
void pi_printf2(const char *szFile, const char *szFunction, int nLine, const char *fmt, ...);

//BOOL doesFileExists(char *szPath);

/** Returns the formatted size in the appropriated unit. Examples:
 *  193 --> "193 bytes"
 *  9185 --> "8.97 KB"
 *  58955455 --> "56.22 MB" 
 *  @param qwSize numerical size
 *  @param szText buffer where to write the result string
 *  @return pointer to the result string (allows to use this function in an expression)
 */
char *formatKind2(WORD wKind, char *szText, int nMaxLen);
#define formatKind(wKind, szBuf) formatKind2(wKind, szBuf, sizeof(szBuf)) 

char *formatSize2(QWORD qwSize, char *szText, int nMaxLen);
#define formatSize(qwSize, szBuf) formatSize2(qwSize, szBuf, sizeof(szBuf)) 

/** Returns the formatted time in a string. This function receives a number of seconds, and
 *  format this in hour, minutes, seconds into a char buffer.
 *  Examples:
 *  70 --> "1min:10sec"
 *  3700 --> "1h:1min:30sec"
 *  @param qwSec number of seconds
 *  @param szBuf buffer where to write the result string
 *  @return pointer to the result string (allows to use this function in an expression)
 */
char *formatTime2(DWORD dwSec, char *szBuf, int nMaxLen);
#define formatTime(dwSec, szBuf) formatTime2(dwSec, szBuf, sizeof(szBuf)) 

int convertDevfsNameToShortName(const char *szDevfs, char *szShort, int nMaxShort);

// ===================================================================
struct COptions
{
  QWORD qwSplitSize; // 0L = auto, !0 = bytes size
  BOOL bOverwrite;
  BOOL bSaveFreeBlocks;
  BOOL bSplitWait; // Wait after each volume change
  BOOL bEraseWithNull;
  BOOL bCheckBeforeSaving;
  BOOL bFailIfMounted; // Fail if file system is mounted
  BOOL bAskDesc;
  BOOL bBatchMode;
  BOOL bSync;
  BOOL bUseSSL; 
  BOOL bBackupMBR;
  BOOL bSimulateMode;
  WORD wFinish;
  DWORD dwServerPort; // ip port
  DWORD dwDebugLevel;
  DWORD dwCompression;
  WORD wEncryptAlgo;   
  WORD wOperation;
  BOOL bShowPartDetails;
  char szServerName[MAX_HOSTNAMESIZE]; // partimaged server's name
  char szAutoMount[MAXPATHLEN];
  BYTE cPassword[MAX_PASSLEN];
};

#define RESET     0
#define BRIGHT    1
#define DIM       2
#define UNDERLINE 3
#define BLINK     4
#define REVERSE   7
#define HIDDEN    8 
#define BLACK     0
#define RED       1
#define GREEN     2
#define YELLOW    3
#define BLUE      4
#define MAGENTA   5
#define CYAN      6
#define WHITE     7 

void textcolor(int attr, int fg, int bg);
void textcolor2(int attr, int fg);

// values for COptions.wOperation
#define OPERATION_SAVE      1
#define OPERATION_RESTORE   2
#define OPERATION_RESTMBR   3
#define OPERATION_IMGINFO   4
#define OPERATION_EXIT      5
#define OPERATION_IMGCONV   6
#define OPERATION_DUPLICATE 7

// values for COptions.wCrypting
#define ENCRYPT_NONE 0
#define ENCRYPT_AES  1

// values for COptions.wCompression
/*#define COMPRESS_NONE  0*/
/*#define COMPRESS_LZO   1*/
/*#define COMPRESS_GZIP  2*/
/*#define COMPRESS_BZIP2 3*/

// values for COptions.wFinish
/*#define FINISH_WAIT   0*/
/*#define FINISH_QUIT   1*/
/*#define FINISH_REBOOT 2*/
/*#define FINISH_HALT   3*/

#endif //COMMON_H
