// swapping macros
#include "pitypes.h"

#define swab16(x) ((u16)( (((u16)(x) & (u16)0x00ffU) << 8) | (((u16)(x) & (u16)0xff00U) >> 8) ))

#define swab32(x) ((u32)( (((u32)(x) & (u32)0x000000ffUL) << 24) | \
                          (((u32)(x) & (u32)0x0000ff00UL) <<  8) | \
                          (((u32)(x) & (u32)0x00ff0000UL) >>  8) | \
                          (((u32)(x) & (u32)0xff000000UL) >> 24) ))
#define swab64(x) ((u64)(  (u64)(((u64)(x) & (u64)0x00000000000000ffULL) << 56) | \
                           (u64)(((u64)(x) & (u64)0x000000000000ff00ULL) << 40) | \
                           (u64)(((u64)(x) & (u64)0x0000000000ff0000ULL) << 24) | \
                           (u64)(((u64)(x) & (u64)0x00000000ff000000ULL) <<  8) | \
                           (u64)(((u64)(x) & (u64)0x000000ff00000000ULL) >>  8) | \
                           (u64)(((u64)(x) & (u64)0x0000ff0000000000ULL) >> 24) | \
                           (u64)(((u64)(x) & (u64)0x00ff000000000000ULL) >> 40) | \
                           (u64)(((u64)(x) & (u64)0xff00000000000000ULL) >> 56) )) 

// ------------- CPU TO XXX ---------------
#ifdef ENDIAN_BIG // BIG ENDIAN
#  define CpuToLe16(a) (swab16(a))
#  define CpuToBe16(a) (a)
#  define CpuToLe32(a) (swab32(a)) 
#  define CpuToBe32(a) (a)
#  define CpuToLe64(a) (swab64(a))
#  define CpuToBe64(a) (a)
#else // LITTLE_ENDIAN
#  define CpuToLe16(a) (a)
#  define CpuToBe16(a) (swab16(a))
#  define CpuToLe32(a) (a)
#  define CpuToBe32(a) (swab32(a))
#  define CpuToLe64(a) (a)
#  define CpuToBe64(a) (swab64(a))
#endif // ENDIAN

#define CpuToLe(a) \
    ((sizeof(a) == 8) ? CpuToLe64(a) : \
    ((sizeof(a) == 4) ? CpuToLe32(a) : \
    ((sizeof(a) == 2) ? CpuToLe16(a) : \
    (a))))

#define CpuToBe(a) \
    ((sizeof(a) == 8) ? CpuToBe64(a) : \
    ((sizeof(a) == 4) ? CpuToBe32(a) : \
    ((sizeof(a) == 2) ? CpuToBe16(a) : \
    (a))))

// ------------- XXX TO CPU --------------
#define Le16ToCpu(a) CpuToLe16(a)
#define Le32ToCpu(a) CpuToLe32(a)
#define Le64ToCpu(a) CpuToLe64(a)

#define Be16ToCpu(a) CpuToBe16(a)
#define Be32ToCpu(a) CpuToBe32(a)
#define Be64ToCpu(a) CpuToBe64(a)

#define LeToCpu(a) \
    ((sizeof(a) == 8) ? Le64ToCpu(a) : \
    ((sizeof(a) == 4) ? Le32ToCpu(a) : \
    ((sizeof(a) == 2) ? Le16ToCpu(a) : \
    (a))))

#define BeToCpu(a) \
    ((sizeof(a) == 8) ? Be64ToCpu(a) : \
    ((sizeof(a) == 4) ? Be32ToCpu(a) : \
    ((sizeof(a) == 2) ? Be16ToCpu(a) : \
    (a))))

// ================================================
// ==================== OLD DEFS ==================
// ================================================

// ------------- CPU TO XXX --------------
/*
#if PI_ENDIAN == ENDIAN_BIG // BIG ENDIAN
#  define CpuToLe16(a) (swab16(a))
#  define CpuToBe16(a) (a)
#  define CpuToLe32(a) (swab32(a)) 
#  define CpuToBe32(a) (a)
#  define CpuToLe64(a) (swab64(a))
#  define CpuToBe64(a) (a)
#  warning "BIG ENDIAN"
#else // LITTLE_ENDIAN
#  define CpuToLe16(a) (a)
#  define CpuToBe16(a) (swab16(a))
#  define CpuToLe32(a) (a)
#  define CpuToBe32(a) (swab32(a))
#  define CpuToLe64(a) (a)
#  define CpuToBe64(a) (swab64(a))
#  warning "LITTLE ENDIAN"
#endif // ENDIAN
*/
/*
#define CpuToLe16(a) ((g_wEndian == ENDIAN_LITTLE) ? a : swab16(a))
#define CpuToBe16(a) ((g_wEndian == ENDIAN_BIG) ? a : swab16(a))
#define CpuToLe32(a) ((g_wEndian == ENDIAN_LITTLE) ? a : swab32(a))
#define CpuToBe32(a) ((g_wEndian == ENDIAN_BIG) ? a : swab32(a))
#define CpuToLe64(a) ((g_wEndian == ENDIAN_LITTLE) ? a : swab64(a))
#define CpuToBe64(a) ((g_wEndian == ENDIAN_BIG) ? a : swab64(a))
*/

/*inline void setEndianess(BOOL bShowPrintf)
{
  DWORD dwNb = 0x11223344;
  BYTE *cBuffer;
  cBuffer = (BYTE*)&dwNb;

  g_wEndian = ENDIAN_UNKNOWN;
  
  if ((cBuffer[0] == 0x44) && (cBuffer[1] == 0x33) && (cBuffer[2] == 0x22) && (cBuffer[3] == 0x11)) // little endian
    {
      g_wEndian = ENDIAN_LITTLE;
      if (bShowPrintf) printf ("Processor = ENDIAN_LITTLE\n");
    }
  
  if ((cBuffer[0] == 0x11) && (cBuffer[1] == 0x22) && (cBuffer[2] == 0x33) && (cBuffer[3] == 0x44)) // big endian
    {
      g_wEndian = ENDIAN_BIG;
      if (bShowPrintf) printf ("Processor = ENDIAN_BIG\n");
    }

  if ((cBuffer[0] == 0x22) && (cBuffer[1] == 0x11) && (cBuffer[2] == 0x44) && (cBuffer[3] == 0x33)) // pdp endian
    {
      g_wEndian = ENDIAN_PDP;
      if (bShowPrintf) printf ("Processor = ENDIAN_PDP\n");
    }
}*/



