/***************************************************************************
                          format.h  -  all structures for image file format
                             -------------------
    begin                : Sun Jul 16 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef FORMAT_H
#define FORMAT_H

#include <sys/param.h>  // MAXPATHLEN
#include "global.h"
#include "common.h"

//SYS_NMLN
#define MAX_UNAMEINFOLEN 65

// on Alpha, sizeof(struct tm)=56
// on i386, sizeof(struct tm)=44
// after each struct tm member of structure, don't forget to add thing like
// this: BYTE cReservedTM[MAX_SIZE_STRUCT_TM-sizeof(struct tm)];
#define MAX_SIZE_STRUCT_TM 56

// number of modules which can compute crc
#define NB_CRC 5

#define IV_SIZE 32

// ===================================================================
    struct CVolumeHeaderVal
    {
      QWORD qwIdentificator; // to avoid using volumes from differents image files on restore
      DWORD dwVolumeNumber;
      DWORD dwCompress; // COMPRESS_XXXXXX
      char szMagicString[32];
      char szVersion[64]; // version of the image file
      //unsigned char cReserved[400]; // Adjust to fit with total header size
    } PACKED; 

struct CVolumeHeader // size must be 512 (adjust the reserved data)
{
  union
  {
    CVolumeHeaderVal val;
    char cReserved[512];
  };
};

// ===================================================================
    struct CMainTailVal
    {
      QWORD qwCRC;
      DWORD dwVolumeNumber;
      //DWORD dwReserved;
      //unsigned char cReserved[16368]; // Adjust to fit with total header size
    } PACKED; 

struct CMainTail // size must be 16384 (adjust the reserved data)
{
  union
  {
    CMainTailVal val;
    char cReserved[HEAD_SIZE];
  };
};

// ================================================
    struct CMainHeaderVal
    {
      QWORD qwPartSize; // size of the partition in bytes
      BOOL bSaveFreeBlocks;
      DWORD dwMainFlags;
      // MBR backup
      DWORD dwMbrCount; // how many MBR are saved in the image file
      DWORD dwMbrSize; // size of a MBR record (allow to change the size in the next versions)
      DWORD dwCompression;
      // future encryption support
      // reserved for future use (save DiskLabel, Extended partitions, ...)
      DWORD dwReservedFuture000;
      DWORD dwReservedFuture001;
      DWORD dwReservedFuture002;
      DWORD dwReservedFuture003;
      DWORD dwReservedFuture004;
      DWORD dwReservedFuture005;
      DWORD dwReservedFuture006;
      DWORD dwReservedFuture007;
      DWORD dwReservedFuture008;
      DWORD dwReservedFuture009;
      //DWORD dwReserved;          // reserved for align

      WORD wEncryptAlgo; // algo used to encrypt data

      char szFileSystem[512]; // ext2fs, ntfs, reiserfs, ...
      char szPartDescription[MAX_DESCRIPTION]; // user description of the partition
      char szOriginalDevice[MAX_DEVICENAMELEN]; // original partition name
      char szFirstImageFilepath[MAXPATHLEN]; // for splitted image files
      char szVersion[64]; // version of the image file
  
      // system and hardware infos
      char szUnameSysname[MAX_UNAMEINFOLEN];
      char szUnameNodename[MAX_UNAMEINFOLEN];
      char szUnameRelease[MAX_UNAMEINFOLEN];
      char szUnameVersion[MAX_UNAMEINFOLEN];
      char szUnameMachine[MAX_UNAMEINFOLEN];
      char cIV[IV_SIZE];          // IV array for crypting module
      char szHostname[MAX_HOSTNAMESIZE];

      struct tm dateCreate; // date of image creation
      //BYTE cReservedTM[MAX_SIZE_STRUCT_TM-sizeof(struct tm)];

      //BYTE cReserved[6456]; // Adjust to fit with total header size
    } PACKED; 

struct CMainHeader // size must be 16384 (adjust the reserved data)
{
  union
  {
    CMainHeaderVal val;
    char cReserved[HEAD_SIZE];
  };
};

// ================================================
/*struct CMainHeader // size must be 16384 (adjust the reserved data)
{
  union
  {
    struct val
    {
      QWORD qwPartSize; // size of the partition in bytes
      BOOL bSaveFreeBlocks;
      DWORD dwMainFlags;
      // MBR backup
      DWORD dwMbrCount; // how many MBR are saved in the image file
      DWORD dwMbrSize; // size of a MBR record (allow to change the size in the next versions)
      DWORD dwCompression;
      // future encryption support
      // reserved for future use (save DiskLabel, Extended partitions, ...)
      DWORD dwReservedFuture000;
      DWORD dwReservedFuture001;
      DWORD dwReservedFuture002;
      DWORD dwReservedFuture003;
      DWORD dwReservedFuture004;
      DWORD dwReservedFuture005;
      DWORD dwReservedFuture006;
      DWORD dwReservedFuture007;
      DWORD dwReservedFuture008;
      DWORD dwReservedFuture009;
      //DWORD dwReserved;          // reserved for align

      WORD wEncryptAlgo; // algo used to encrypt data

      char szFileSystem[512]; // ext2fs, ntfs, reiserfs, ...
      char szPartDescription[MAX_DESCRIPTION]; // user description of the partition
      char szOriginalDevice[MAX_DEVICENAMELEN]; // original partition name
      char szFirstImageFilepath[MAXPATHLEN]; // for splitted image files
      char szVersion[64]; // version of the image file
  
      // system and hardware infos
      char szUnameSysname[MAX_UNAMEINFOLEN];
      char szUnameNodename[MAX_UNAMEINFOLEN];
      char szUnameRelease[MAX_UNAMEINFOLEN];
      char szUnameVersion[MAX_UNAMEINFOLEN];
      char szUnameMachine[MAX_UNAMEINFOLEN];
      char cIV[IV_SIZE];          // IV array for crypting module
      char szHostname[MAX_HOSTNAMESIZE];

      struct tm dateCreate; // date of image creation
      //BYTE cReservedTM[MAX_SIZE_STRUCT_TM-sizeof(struct tm)];

      //BYTE cReserved[6456]; // Adjust to fit with total header size
    } val PACKED; 
    char cReserved[HEAD_SIZE];
  };
};
*/
// ================================================
struct CLocalHeaderVal
    {
      QWORD qwBlockSize;
      QWORD qwUsedBlocks;
      QWORD qwBlocksCount;
      QWORD qwBitmapSize; // bytes in the bitmap
      QWORD qwBadBlocksCount;
      char szLabel[64];
      //BYTE cReserved[16280]; // Adjust to fit with total header size
    } PACKED; 

struct CLocalHeader // size must be 16384 (adjust the reserved data)
{
  union
  {
    CLocalHeaderVal val;
    char cReserved[HEAD_SIZE];
  };
};

// ================================================
/*struct CChunkHeader
{
  QWORD qwPos; // number of the first block of the chunk
  DWORD dwCRC[NB_CRC]; // CRC of the CHECK_FREQUENCY blocks
  DWORD dwCompress;
  DWORD dwChunkCompressedSize;
  WORD  wKind;
  char cMagic[3]; // must be 'C','H','K'
};*/

#endif //FORMAT_H
