/***************************************************************************
       fs_base.cpp  -  Base class of all File System with common code
                             -------------------
    begin                : Thu Jul 13 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <sys/param.h>

#include "fs_base.h"
#include "global.h"
#include "imagefile.h"
#include "common.h"
#include "cbitmap.h"
#include "exceptions.h"
#include "misc.h"
#include "interface_newt.h"

// =======================================================
CFSBase::CFSBase(char *szDevice, FILE *fDeviceFile, QWORD qwPartSize)
{
  BEGIN;

  m_fDeviceFile = fDeviceFile;
  strncpy(m_szDevice, szDevice, MAX_DEVICENAMELEN);
  memset(&m_header, 0, sizeof(CLocalHeader));
  m_qwPartSize = qwPartSize;
  showDebug(1, "m_qwPartSize=%llu = %llu * 512\n",m_qwPartSize,m_qwPartSize/512LL);
  m_bCheckSuperBlockUsedSize = false;
  m_bCheckSuperBlockFreeSize = false;
  m_bCanBeMore = false;

  RETURN;
}

// =======================================================
CFSBase::~CFSBase()
{
}

// =======================================================
// may throw: errno, readbitmap, reading, writing
int CFSBase::saveUsedBlocksToImageFile(CPiTimer *timer, COptions *options)
{
/*  BEGIN;

  char cBuffer[COPYBUF_SIZE+1];
  BOOL bUsed;
  DWORD dwCount;
  int nRes;
  QWORD i;
  DWORD dwStats;
  DWORD dwTemp;

  DWORD dwSavedBlocks = 0L;
  DWORD dwBlocksInBuffer;
  DWORD dwDataSize;
  QWORD qwBlocksToCopyCount;

  // init ESC key
  //g_bSigInt = false;

  // go to the beginning of the disk
  nRes = fseek(m_fDeviceFile, 0L, SEEK_SET);
  if (nRes == -1)
    return -1; //THROW(ERR_ERRNO, errno);
  
  // init the bitmap optimizations
  m_bitmap.setMaxFreeBlocksCopiedPerOper(MAX_FREE_BLOCKS_SKIPPED); // could be another value
  dwBlocksInBuffer = (COPYBUF_SIZE-(COPYBUF_SIZE % SGET(m_header,qwBlockSize))) / (SGET(m_header,qwBlockSize)); 
  m_bitmap.setMaxUsedBlocksCopiedPerOper(dwBlocksInBuffer);
  showDebug(3, "CPY: setMax=%lu\n", dwBlocksInBuffer);

  if (options->bSaveFreeBlocks)
    qwBlocksToCopyCount = SGET(m_header, qwBlocksCount);
  else
    qwBlocksToCopyCount = SGET(m_header, qwUsedBlocks);

  dwStats = 0;
  i = 0;
  while (i < SGET(m_header, qwBlocksCount)) // while there are blocks to copy
    {

      nRes = m_bitmap.readBitmap(i, &bUsed, &dwCount, options->bSaveFreeBlocks);
      if (nRes == -1)
        return -1; //THROW(ERR_READ_BITMAP, i);

      // stop at the end of the partition
      dwCount = min(dwCount, SGET(m_header, qwBlocksCount)-i);

      showDebug(3, "CPY: dwCount=%lu, bUsed=%d\n", dwCount, bUsed);

      if (bUsed) // if used blocks are marked in the bitmap
	{
	  dwDataSize = SGET(m_header, qwBlockSize*dwCount);
	  
	  while (dwDataSize > 0)
	    {
	      dwTemp = min(dwDataSize, COPYBUF_SIZE);
	      
	      errno = 0;
	      nRes = fread(cBuffer, dwTemp, 1, m_fDeviceFile);
	      if (nRes != 1)
		return -1; //THROW(ERR_READING, (DWORD)i, errno);

	      try { m_image->write(KIND_BLOCKS, cBuffer, dwTemp, true); }
	      catch (CExceptions * excep)
		{
		  showDebug(1, "saveUsedBlocksToImageFile: excep %d\n", excep->GetExcept());
		  throw excep;
		}
	      dwDataSize -= dwTemp;
	    }


 	  // show stats
	  //if (dwStats == 8)
	  {
	    m_guiSave->showStats(timer, SGET(m_header, qwBlockSize), dwSavedBlocks, qwBlocksToCopyCount);
	    dwStats = 0;
	  }


	  dwStats++;

	  //checkForEscape();

	  // ---- end -----

	  dwSavedBlocks += dwCount;

	}
      else // if free blocks were marked
	{
	  nRes = fseek(m_fDeviceFile, SGET(m_header, qwBlockSize)*dwCount, SEEK_CUR);
	  if (nRes == -1)
            {
              return -1; //THROW(ERR_ERRNO, errno);
            }
	}

      i += dwCount;

      // draw progress
      m_guiSave->showStats(timer, SGET(m_header, qwBlockSize), dwSavedBlocks, SGET(m_header, qwUsedBlocks));
    }

  m_guiSave->showStats(timer, SGET(m_header, qwBlockSize), dwSavedBlocks, qwBlocksToCopyCount);
  g_qwCopiedBytesCount = SGET(m_header, qwBlockSize) * dwSavedBlocks; // for final message box
  
  RETURN_int(0);
*/
   return 0;
}

// =======================================================
// may raise errno, check_num, check_crc,zer, writing, wrong_magic
// NEW optimized version
int CFSBase::restoreUsedBlocksToDisk(CPiTimer *timer, COptions *options)
{
/*
  BEGIN;

   char cZeroBuffer[ZERO_BLOCK_SIZE];
   char cBuffer[COPYBUF_SIZE+1];
   BOOL bUsed;
   DWORD dwCount;
   int nRes;
   QWORD i;
   DWORD dwBlocksInBuffer;
   DWORD dwTemp;
   DWORD dwTemp2;
   DWORD dwDataSize;
   DWORD dwSavedBlocks = 0L;
   //DWORD dwCRC;
   DWORD dwStats;

   memset(cZeroBuffer, 0, ZERO_BLOCK_SIZE);
   //dwCRC = increaseCrc32(0, NULL, 0);  

   // init ESC key
   //g_bSigInt = false;

   // go to the beginning of the disk
   nRes = fseek(m_fDeviceFile, 0L, SEEK_SET);
   if (nRes == -1)
      return -1; //THROW(ERR_ERRNO, errno);

   // init the bitmap optimizations
   m_bitmap.setMaxFreeBlocksCopiedPerOper(MAX_FREE_BLOCKS_SKIPPED); // could be another value
   dwBlocksInBuffer = (COPYBUF_SIZE-(COPYBUF_SIZE % SGET(m_header, qwBlockSize))) / (SGET(m_header, qwBlockSize)); 
   m_bitmap.setMaxUsedBlocksCopiedPerOper(dwBlocksInBuffer);
   showDebug(3, "CPY: setMax=%lu\n", dwBlocksInBuffer);
  
   dwStats = 0;
   i = 0;
   while (i < SGET(m_header, qwBlocksCount))
   {
      showDebug(1, "i = %lld, m_header.qwBlocksCount = %lld\n", i, SGET(m_header, qwBlocksCount));

      m_bitmap.readBitmap(i, &bUsed, &dwCount, options->bSaveFreeBlocks);
      dwCount = min(dwCount, SGET(m_header, qwBlocksCount)-i); // stop at the end of the partition

      showDebug(3, "CPY: dwCount=%lu, bUsed=%d\n", dwCount, bUsed);

      if (bUsed) // if used blocks are marked in the bitmap
      {
	 dwDataSize = SGET(m_header, qwBlockSize*dwCount);
	 dwSavedBlocks += dwCount;

	 while (dwDataSize > 0)
	 {
	    dwTemp2 = min(dwDataSize, COPYBUF_SIZE);

	    try { m_image->read(cBuffer, dwTemp2, true); }
	    catch (CExceptions * excep)
	    {
	       showDebug(1, "restoreUsedBlocksToDisk: except %d\n",
			 excep -> GetExcept());
	       throw excep;
	    }	

	    errno = 0;
	    if (!options->bSimulateMode)
	    {
	       nRes = fwrite(cBuffer, dwTemp2, 1, m_fDeviceFile);
	       if (nRes != 1)
	       {
		  return -1; //THROW(ERR_WRITING, (DWORD) 0, errno);
	       }				
	    }
	    else
	       showDebug(10, "fwrite simulated\n");

	    dwDataSize -= dwTemp2;
	 }

	 // show stats
	 //if (dwStats == 8)
	 {
	    // update stats
	    m_guiRestore->showStats(timer, SGET(m_header, qwBlockSize), dwSavedBlocks,
				    SGET(m_header, qwUsedBlocks), SGET(m_header, qwBlocksCount), 
				    options->bEraseWithNull | options->bSaveFreeBlocks);
	    dwStats = 0;
	 }
	 dwStats++;

	 //checkForEscape();

	 showDebug(9, "end while dwDataSize\n");
	  
	 // ---- end -----
	  
      }
      else // if free blocks were marked in the bitmap
      {
	 if (options->bEraseWithNull == true) // if free and we must write 0 bytes on unused clusters
	 {
	    dwTemp = SGET(m_header, qwBlockSize) * dwCount; // data to copy
	      
	    while (dwTemp > 0)
	    {
	       errno = 0;
	       if (!options->bSimulateMode)
	       {
		  nRes = fwrite(cZeroBuffer, min(ZERO_BLOCK_SIZE, dwTemp),
				1, m_fDeviceFile);
		  if (nRes != 1)
		     return -1; //THROW(ERR_ZEROING, errno);
	       }
	       else
		  showDebug(10, "fwrite simulated\n");
		  
	       dwTemp -= min(ZERO_BLOCK_SIZE, dwTemp);
	    }
	    dwSavedBlocks += dwCount;
	 }
	 else // don't write zero bytes
	 {
	    nRes = fseek(m_fDeviceFile, SGET(m_header, qwBlockSize)*dwCount, SEEK_CUR);
	    if (nRes == -1)
	       return -1; //THROW(ERR_ERRNO, errno);
	 }			
      }
      i += dwCount;
   }
  
   // update stats
   m_guiRestore->showStats(timer, SGET(m_header, qwBlockSize), dwSavedBlocks,
			   SGET(m_header, qwUsedBlocks), SGET(m_header, qwBlocksCount), options->bEraseWithNull | options->bSaveFreeBlocks);
   g_qwCopiedBytesCount = SGET(m_header, qwBlockSize) * dwSavedBlocks; // for final message box

   RETURN_int(0);
*/
   return 0;
}

// =======================================================
int CFSBase::saveToImage(CPiTimer *timer, CImage *image, COptions *options, CSavingWindow *gui)
{
/*
   BEGIN;

   m_image = image;
   m_guiSave = gui;
  
   // ---- 0. Initializations
  
   // read informations
   showDebug(3, "TRACE: 001 (init)\n");  
   //g_interface -> StatusLine(i18n("reading partition properties"));
   try { readSuperBlock(); }
   catch ( CExceptions * excep )
   {
      showDebug(1, "saveToImage: excep %d\n", excep->GetExcept());

      if (excep->GetExcept() == ERR_BLOCKSIZE)
	 throw excep;
      else
	 return -1; //THROW(ERR_READ_SUPERBLOCK);
   }	
  
   // check the file system
   showDebug(3, "TRACE: 002 (fsck)\n");  
   if (options->bCheckBeforeSaving == true)
   {
      g_interface -> StatusLine(i18n("checking the file system with fsck"));

      try { fsck(); }
      catch ( CExceptions * excep )
      { return -1;  }
   }

   // read bitmap into memory (must be done before writing header)
   showDebug(3, "TRACE: 003 (readBitmap)\n");  
   showDebug(1, "begin of readBitmap()\n");
   readBitmap();
   showDebug(1, "end of readBitmap()\n");
  
   // check the bitmap has been read without error
   checkBitmapInfos();

   // print informations	
   if (options->bBatchMode == false)
      printfInformations();
  
   // write header
   showDebug(3, "TRACE: 004 (writeHeader)\n");  
   //g_interface -> StatusLine(i18n("writing header"));
   try { m_image->writeMagic(KIND_LOCALHEAD, MAGIC_BEGIN_LOCALHEADER); }
   catch ( CExceptions * excep )
   { throw excep; }
   errno = 0;
   try { m_image->write(KIND_LOCALHEAD, &m_header, sizeof(CLocalHeader), true); }
   catch ( CExceptions * excep )
   {
      showDebug(1, "saveToImage: excep %d\n", excep->GetExcept());
      throw excep;
   }

   // write CRC of CLocaleader
   m_image->writeCRC(KIND_LOCALHEAD, (char*)&m_header, sizeof(CLocalHeader));
  
   // copy bitmap to image file
   showDebug(3, "TRACE: 005 (writeBitmap)\n");  
   try { m_image->writeMagic(KIND_HEADER, MAGIC_BEGIN_BITMAP); }
   catch ( CExceptions * excep )
   { throw excep; }

   try { m_image->write(KIND_HEADER, m_bitmap.data(), SGET(m_header, qwBitmapSize), true); }
   catch ( CExceptions * excep )
   {
      showDebug(1, "saveToImage: excep %d\n", excep->GetExcept());
      throw excep;
   }

   // write informations specific for the FS
   showDebug(3, "TRACE: 006 (writeInfo)\n");  
   try { m_image->writeMagic(KIND_HEADER, MAGIC_BEGIN_INFO); }
   catch ( CExceptions * excep )
   { throw excep; }

   try { m_image->write(KIND_HEADER, (char*)getInfos(), INFOS_STRUCT_SIZE, true); }
   catch ( CExceptions * excep )
   {
      showDebug(1, "saveToImage: excep %d\n", excep->GetExcept());
      throw excep;
   }
  
   // write CRC of info
   m_image->writeCRC(KIND_HEADER, (char*)getInfos(), INFOS_STRUCT_SIZE);

   // write magic string
   try { m_image->writeMagic(KIND_HEADER, MAGIC_BEGIN_DATABLOCKS); }
   catch ( CExceptions * excep )
   {
      showDebug(1, "saveToImage: excep %d\n", excep->GetExcept());
      throw excep;
   }
  
   // time of beginning
   timer->timeStart = time(0);
    
   // ---- write all used blocks
   showDebug(3, "TRACE: 007 (copyUsedBlocks)\n");  

   showDebug(3, "begin of saveUsedBlocksToImageFile()\n");  
   try { saveUsedBlocksToImageFile(timer, options); }
   catch ( CExceptions * excep)
   {
      showDebug(1, "saveToImage: excep %d\n", excep->GetExcept());
      throw excep;
   }

   RETURN_int(0);
*/
return 0;
}

// =======================================================
int CFSBase::restoreFromImage(CPiTimer *timer, CImage *image, COptions *options, CRestoringWindow *gui)
{
/*
  BEGIN;

  // init
  m_image = image;
  m_guiRestore = gui;
  
  g_interface -> StatusLine(i18n("reading partition informations"));
  
  // 1. ---- Read header
  showDebug(3, "TRACE: 001 (readHeader)\n");  
  try { m_image->readAndCheckMagic(MAGIC_BEGIN_LOCALHEADER); } 
  catch ( CExceptions * excep )
    {
      showDebug(1, "restoreFromImage: %d\n", excep->GetExcept()); 
      throw excep;
    }

  try {m_image->read((char*)&m_header, sizeof(CLocalHeader), true);}
  catch(CExceptions * excep)
    {
      showDebug(1, "restoreFromImage: %d\n", excep->GetExcept()); 
      throw excep;
    }

  // read CRC of info
  m_image->readAndCheckCRC((char*)&m_header, sizeof(CLocalHeader));

  
  // check the current format is supported
//  if (strncmp(headReiser.szAppFormat, CURRENT_REISER_FORMAT, 
//    strlen(CURRENT_REISER_FORMAT)) < 0)
//    {
//    g_interface->ErrorNewerRelease();
//    RETURN_int(-1);
//    }	
    
  // 2. ---- Print informations and ask confirmation to restore
//  showDebug(1, "TRACE: 002 (printfInfos)\n");  
//    if (options->bBatchMode == false)
//    {
//    printfInformations();
//      
//    // ask confirmation
//    nRes = g_interface -> askRestore(m_szDevice,
//    image->get_szImageFilename());
//    if (nRes == 2) // 2 --> cancel
//    RETURN_int(-1);
//    }
  
  // time of beginning
  timer->timeStart = time(0);
  
  // ---- 5. read bitmap
  showDebug(3, "TRACE: 003 (readBitmap)\n");  
  try { m_image->readAndCheckMagic(MAGIC_BEGIN_BITMAP); } 
  catch ( CExceptions * excep )
    {
      showDebug(1, "restoreFromImage: %d\n", excep->GetExcept()); 
      throw excep;
    }
  
  // init bitmap size
  m_bitmap.init(SGET(m_header, qwBitmapSize));
  
  // copy bitmap blocks to bitmap in memory
  try{m_image->read((char*)m_bitmap.data(), (long)(SGET(m_header, qwBitmapSize)), true);}
  catch (CExceptions * excep)
    {
      showDebug(1, "restoreFromImage: %d\n", excep->GetExcept()); 
      throw excep;
    }

  // write informations specific for the FS
  showDebug(3, "TRACE: 004 (readInfo)\n");  
  try {m_image->readAndCheckMagic(MAGIC_BEGIN_INFO); }
  catch ( CExceptions * excep )
    {
      showDebug(1, "restoreFromImage: %d\n", excep->GetExcept()); 
      throw excep;
    }

  try { m_image->read((char*)getInfos(), INFOS_STRUCT_SIZE,true); }
  catch ( CExceptions * excep )
    {
      showDebug(1, "restoreFromImage: %d\n", excep->GetExcept()); 
      throw excep;
    }

  // read CRC of info
  m_image->readAndCheckCRC((char*)getInfos(), INFOS_STRUCT_SIZE);

  // 2. ---- Print informations and ask confirmation to restore
  showDebug(3, "TRACE: 002 (printfInfos)\n");  
  if (options->bBatchMode == false)
    {
      printfInformations();
      
      // ask confirmation
      if (!options->bSimulateMode)
	{
	  nRes = g_interface -> askRestore(m_szDevice, image->get_szImageFilename());
	  if (nRes == MSGBOX_NO) // 2 --> cancel
	    return -1; //THROW(ERR_CANCELED);
	}
    }

  // read and check magic string
  showDebug(3, "TRACE: 005 (copyUsedBlocks)\n");  
  try { m_image->readAndCheckMagic(MAGIC_BEGIN_DATABLOCKS); }
  catch ( CExceptions * excep )
    {
      showDebug(1, "restoreFromImage: %d\n", excep->GetExcept()); 
      throw excep;
    }
  
  // ---- 7. Restore all used blocks
  errno = 0;
  g_interface -> StatusLine(i18n("copying used data blocks"));
  showDebug(1, "begin of restoreUsedBlocksToDisk()\n");  
  try { restoreUsedBlocksToDisk(timer, options); }
  catch ( CExceptions * excep )
    {
      showDebug(1, "restoreFromImage: %d\n", excep->GetExcept()); 
      throw excep;
    }
  showDebug(9, "end of restoreUsedBlocksToDisk()\n");  

  RETURN_int(0);
*/
  return 0;
}

// =======================================================
int CFSBase::getStdInfos(char *szDest, int nMaxLen, bool bShowBlocksInfo)
{
  BEGIN;

  char szTemp1[512];
  char szTemp2[512];
  char szTemp3[512];
  char szTemp4[512];
  char szBlocksInfo[8192];

  if (bShowBlocksInfo)
    {
      SNPRINTF(szBlocksInfo, i18n("Block size....................%llu bytes\n"
				  "Total blocks count............%llu\n"
				  "Used blocks count.............%llu\n"
				  "Free blocks count.............%llu\n"),
	       SGET(m_header, qwBlockSize), 
	       SGET(m_header, qwBlocksCount),
	       SGET(m_header, qwUsedBlocks),
	       SGET(m_header, qwBlocksCount)-SGET(m_header,qwUsedBlocks));     
    }
  else
    {
      *szBlocksInfo = 0; // empty
    }
 
  snprintf(szDest, nMaxLen, i18n("%s" // Blocks infos
				 "Space usage:..................%llu %%\n"
				 "Total space:..................%s\n"
				 "Used space....................%s\n"
				 "Free space....................%s\n"
				 "Bitmap size...................%s\n"
				 "Label.........................%s\n"),
	   szBlocksInfo, 
	   (100 * SGET(m_header, qwUsedBlocks)) / SGET(m_header, qwBlocksCount), 
	   formatSize(SGET(m_header, qwBlocksCount) * SGET(m_header, qwBlockSize), szTemp4),
	   formatSize(SGET(m_header, qwUsedBlocks) * SGET(m_header, qwBlockSize), szTemp1),
	   // used space
	   formatSize((SGET(m_header, qwBlocksCount)-SGET(m_header,qwUsedBlocks)) *
		      SGET(m_header, qwBlockSize), szTemp2), // free space
	   formatSize(SGET(m_header, qwBitmapSize), szTemp3),
	   m_header.val.szLabel);

  showDebug(1, "\n====================\n%s====================\n\n", szDest);
 
  RETURN_int(0);
}			

// =======================================================
int CFSBase::readData(void *cBuffer, QWORD qwOffset, DWORD dwLength)
{
  BEGIN;

  int nRes;

  errno = 0;
  nRes = bigFseek(m_fDeviceFile, qwOffset);
  if (nRes != 0)
    {
      showDebug(1, "ERROR: bigFseek(qwOffset=%llu) failed in readData(qwOffset, dwLength=%lu) errno=%d=%s\n", qwOffset, dwLength, errno, strerror(errno));
      RETURN_int(-1);
    }

  errno = 0;
  nRes = fread((char *)cBuffer, dwLength, 1, m_fDeviceFile);
  if (nRes != 1)
    {
      showDebug(1, "ERROR: fread(dwLength=%llu) failed in readData(qwOffset=%llu, dwLength) errno=%d=%s\n", dwLength, qwOffset, errno, strerror(errno));
      RETURN_int(-1);
    }
  
  RETURN_int(0); // success
}

// =======================================================
int CFSBase::calculateSpaceFromBitmap()
{ 
  BEGIN;

  QWORD qwUsedBlocksCount = 0;
  QWORD qwFreeBlocksCount = 0;
  QWORD i;

  for (i=0; i < SGET(m_header,qwBlocksCount); i++)
    {
      if (m_bitmap.isBitSet(i))
	qwUsedBlocksCount++;
      else
	qwFreeBlocksCount++;
    }

  SPUT(m_header, qwUsedBlocks, qwUsedBlocksCount);
  //m_header.qwUsedBlocks = qwUsedBlocksCount;

  showDebug(1, "BITMAP: qwUsedBlocks = %llu\n", qwUsedBlocksCount);
  showDebug(1, "BITMAP: qwFreeBlocks = %llu\n", qwFreeBlocksCount);

  RETURN_int(0);
}

// =======================================================
int CFSBase::checkBitmapInfos()
{
   QWORD qwBitmapUsedSize;
   QWORD qwBitmapFreeSize;
   qwBitmapUsedSize = SGET(m_header, qwUsedBlocks) * SGET(m_header, qwBlockSize);
   qwBitmapFreeSize = (SGET(m_header, qwBlocksCount) - SGET(m_header, qwUsedBlocks)) * SGET(m_header, qwBlockSize);

   showDebug(1, "CHECK BITMAP: qwSuperBlockUsedSize.....%llu = %llu sectors\n", m_qwSuperBlockUsedSize, m_qwSuperBlockUsedSize/512LL);
   showDebug(1, "CHECK BITMAP: qwBitmapUsedSize.........%llu = %llu sectors\n", qwBitmapUsedSize, qwBitmapUsedSize/512LL);

   showDebug(1, "CHECK BITMAP: qwSuperBlockFreeSize.....%llu = %llu sectors\n", m_qwSuperBlockFreeSize, m_qwSuperBlockFreeSize/512LL);     
   showDebug(1, "CHECK BITMAP: qwBitmapFreeSize.........%llu = %llu sectors\n", qwBitmapFreeSize, qwBitmapFreeSize/512LL);
  
   if ((m_bCheckSuperBlockUsedSize)) // && (qwBitmapUsedSize != m_qwSuperBlockUsedSize))
   {
      if ((!m_bCanBeMore && (qwBitmapUsedSize != m_qwSuperBlockUsedSize)) || (m_bCanBeMore && (qwBitmapUsedSize < m_qwSuperBlockUsedSize)))
      {
	 showDebug(1, "ERROR BITMAP Used and m_bCanBeMore=%d\n", m_bCanBeMore);
	 return -1; //THROW(ERR_READ_BITMAP, (DWORD)0);
      }
   }

   showDebug(1, "CHECK BITMAP: qwSuperBlockFreeSize.....%llu = %llu sectors\n", m_qwSuperBlockFreeSize, m_qwSuperBlockFreeSize/512LL);
   showDebug(1, "CHECK BITMAP: qwBitmapFreeSize.........%llu = %llu sectors\n", qwBitmapFreeSize, qwBitmapFreeSize/512LL);

   if ((m_bCheckSuperBlockFreeSize)) // && (qwBitmapFreeSize != m_qwSuperBlockFreeSize))
   {
      if ((!m_bCanBeMore && (qwBitmapFreeSize != m_qwSuperBlockFreeSize)) || (m_bCanBeMore && (qwBitmapFreeSize > m_qwSuperBlockFreeSize)))
      {
	 showDebug(1, "ERROR BITMAP Free and m_bCanBeMore=%d\n", m_bCanBeMore);
	 return -1; //THROW(ERR_READ_BITMAP, (DWORD)0);
      }
   }

   RETURN_int(0);
}

// =======================================================
void CFSBase::setSuperBlockInfos(BOOL bCheckSuperBlockUsedSize, BOOL bCheckSuperBlockFreeSize, QWORD qwSuperBlockUsedSize, QWORD qwSuperBlockFreeSize, BOOL bCanBeMore/*=false*/) 
{
   m_bCanBeMore = bCanBeMore;
   m_qwSuperBlockUsedSize = qwSuperBlockUsedSize; m_qwSuperBlockFreeSize = qwSuperBlockFreeSize;
   m_bCheckSuperBlockUsedSize = bCheckSuperBlockUsedSize;
   m_bCheckSuperBlockFreeSize = bCheckSuperBlockFreeSize;
}

// =======================================================
int CFSBase::getGeneralInformations(CLocalHeader *head)
{
  m_image = NULL;
  m_guiSave = NULL;

  // ---- 0. Initializations
  // read informations
  try { readSuperBlock(); }
  catch ( CExceptions * excep )
    {
      /*if (excep->GetExcept() == ERR_BLOCKSIZE)
	throw excep;
      else
	return -1; //THROW(ERR_READ_SUPERBLOCK);*/
       return -1;
    }	

  // check the file system
  /*if (options->bCheckBeforeSaving == true)
    {
       //g_interface -> StatusLine(i18n("checking the file system with fsck"));

      try { fsck(); }
      catch ( CExceptions * excep )
        { return -1; //THROW(ERR_FSCHK); }
    }*/

  // read bitmap into memory (must be done before writing header)
  try {readBitmap(/*options*/);}
  catch (int *a)
  {
     printf ("STATS: errro 003\n");
  }

  // check the bitmap has been read without error
  checkBitmapInfos();

  memcpy(head, &m_header, sizeof(CLocalHeader));

  return (0);
}

