/***************************************************************************
                          fs_base.h  -  description
                             -------------------
    begin                : Sun Mar 04 2001
    copyright            : (C) 2001 by Franck Ladurelle
    email                : ladurelf@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef _FS_BASE_H_
#define _FS_BASE_H_

#include "global.h"
#include "cbitmap.h"
#include "format.h"
#include "modules.h"

struct CImage;
struct CVolumeHeader;
struct CMainTail;
struct CMainHeader;
struct COptions;
class CSavingWindow;
class CRestoringWindow;

#define ZERO_BLOCK_SIZE 512
#define INFOS_STRUCT_SIZE 16384

// ================================================
class CFSBase
{
 public: // defined in CFSBase
  CFSBase(char *szDevice, FILE *fDeviceFile, QWORD qwPartSize);
  virtual ~CFSBase();

  int saveUsedBlocksToImageFile(CPiTimer *timer, COptions *options);
  int restoreUsedBlocksToDisk(CPiTimer *timer, COptions *options);

  int saveToImage(CPiTimer *timer, CImage *image, COptions *options, CSavingWindow *gui);
  int restoreFromImage(CPiTimer *timer, CImage *image, COptions *options, CRestoringWindow *gui);

  int getStdInfos(char *szDest, int nMaxLen, bool bShowBlocksInfo);
  int readData(void *cBuffer, QWORD qwOffset, DWORD dwLength);
  int calculateSpaceFromBitmap();

  void setSuperBlockInfos(BOOL bCheckSuperBlockUsedSize, BOOL bCheckSuperBlockFreeSize, 
			  QWORD qwSuperBlockUsedSize, QWORD qwSuperBlockFreeSize, BOOL bCanBeMore=false);
  int checkBitmapInfos();
  int getGeneralInformations(CLocalHeader *head);

 private:
  QWORD m_qwSuperBlockUsedSize;  
  QWORD m_qwSuperBlockFreeSize; 
  BOOL m_bCheckSuperBlockUsedSize;
  BOOL m_bCheckSuperBlockFreeSize;
  BOOL m_bCanBeMore;
  CSavingWindow *m_guiSave;
  CRestoringWindow *m_guiRestore;

 public: // defined in CFSXXXXX
  virtual void printfInformations() = 0;
  virtual int readBitmap(/*COptions *options*/)=0;
  virtual int readSuperBlock()=0;
  virtual int fsck()=0;
  virtual void *getInfos()=0;
 
 public: //private:
  CImage *m_image;
  CBlocksUseBitmap m_bitmap;
  QWORD m_qwPartSize;

  // ex: "/dev/hdc5"
  char m_szDevice[MAX_DEVICENAMELEN+1];

  CLocalHeader m_header; // general infos (block count, block size, used space, ...)
  
  // File descriptors
  FILE *m_fDeviceFile;
};

#endif // _FS_BASE_H_
