/***************************************************************************
          fs_ext2.cpp  -  EXT2FS (Linux standard) File System Support
                             -------------------
    begin                : Mon May 22 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "fs_ext2.h"
#include "global.h"
#include "errors.h"
#include "exceptions.h"
#include "imagefile.h"

// =======================================================
CExt2Part::CExt2Part(char *szDevice, FILE *fDeviceFile, QWORD qwPartSize): CFSBase(szDevice, fDeviceFile, qwPartSize)
{
  BEGIN;
  RETURN;
}

// =======================================================
CExt2Part::~CExt2Part() 
{
  BEGIN;
  RETURN;
}

// =======================================================
int CExt2Part::fsck()
{
  BEGIN;

  char szCommand[1024];
  int nRes;

  SNPRINTF(szCommand, "/sbin/e2fsck -nf -C 0 %s > /dev/null 2>/dev/null", m_szDevice);
  //snprintf(szCommand, sizeof(szCommand), "/sbin/e2fsck -nf -C 0 %s > /dev/null 2>/dev/null", m_szDevice);
  //*(szCommand+1023) = '\0';
  
  showDebug(1, "FSCK: [%s]\n", szCommand);

  nRes = system(szCommand);
  if (nRes != 0)
    {	
//      nRes = g_interface -> askIgnoreFSError("e2fsck");
//      if (nRes == MSGBOX_CANCEL)
        return -1; //THROW(ERR_FSCHK);
    }

  RETURN_int(0);  
}

// =======================================================
void CExt2Part::printfInformations()
{
  char szFullText[8192];
  char szText[8192];
  
  *(szText) = '\0';
  getStdInfos(szText, sizeof(szText), false); 
  
  SNPRINTF(szFullText, i18n("%s" // standard infos
			    "Number of groups..............%lu\n"
			    "First block...................%lu\n"
			    "Ext3fs journalization.........%s\n"
			    "Sparse super block............%s\n"
			    "Large files support...........%s\n"
			    "File system revision..........%lu\n"),
	   szText, IGET(dwGroupsCount), IGET(dwFirstBlock), 
	   (IGET(dwFeatureCompat) & EXT3_FEATURE_COMPAT_HAS_JOURNAL) ? i18n("yes") : i18n("no"),
	   (IGET(dwFeatureRoCompat) & EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER) ? i18n("yes") : i18n("no"),
	   (IGET(dwFeatureRoCompat) & EXT2_FEATURE_RO_COMPAT_LARGE_FILE) ? i18n("yes") : i18n("no"),
	   IGET(dwRevLevel));
  
//  if (ISEXT3FS)
//    g_interface->msgBoxOk(i18n("Ext3fs informations"), szFullText);
//  else
//    g_interface->msgBoxOk(i18n("Ext2fs informations"), szFullText);
}			

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%% begin of CODE WITHOUT LIBEXT2FS %%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// =======================================================
int CExt2Part::readSuperBlock() 
{
  BEGIN;

  CExt2Super sb;
  int nRes;

  // init
  memset(&m_info, 0, sizeof(CInfoExt2Header));
  //int nExt2Magic = EXT2_SUPER_MAGIC;

  // 0. go to the beginning of the super block
  nRes = fseek(m_fDeviceFile, 1024, SEEK_SET);
  if (nRes == -1)
    goto error_readSuperBlock;
  
  // 1. read and print important informations
  nRes = fread(&sb, sizeof(sb), 1, m_fDeviceFile);
  if (nRes != 1)
    goto error_readSuperBlock;

  // check partition has an ext2 file system
  //if (memcmp(&sb.s_magic, &nExt2Magic, 2))
  if (sb.s_magic != Le16ToCpu(EXT2_SUPER_MAGIC))
    goto error_readSuperBlock;

  // read and print important informations	
  ISET(dwBlockSize, EXT2_MIN_BLOCK_SIZE << Le32ToCpu(sb.s_log_block_size)); //sb.s_blocksize; //EXT2_BLOCK_SIZE(m_fs->super);
  showDebug(1, "blksize=%lu\n", IGET(dwBlockSize));
  ISET(dwFirstBlock, Le32ToCpu(sb.s_first_data_block));
  showDebug(1, "first=%lu\n", IGET(dwFirstBlock));
  ISET(dwTotalBlocksCount, Le32ToCpu(sb.s_blocks_count));
  showDebug(1, "total blocks=%lu\n", IGET(dwTotalBlocksCount));
  ISET(dwBlocksPerGroup, Le32ToCpu(sb.s_blocks_per_group));

  ISET(dwGroupsCount, (IGET(dwTotalBlocksCount) - IGET(dwFirstBlock) + IGET(dwBlocksPerGroup) - 1) / Le32ToCpu(sb.s_blocks_per_group));
  showDebug(1, "groups=%lu\n", IGET(dwGroupsCount));
  ISET(dwLogicalBlocksPerExt2Block, IGET(dwBlockSize) / LOGICAL_EXT2_BLKSIZE);
  showDebug(1, "m_info.dwLogicalBlocksPerExt2Block=%lu\n", IGET(dwLogicalBlocksPerExt2Block));
  showDebug(1, "logperblok=%lu\n", IGET(dwLogicalBlocksPerExt2Block));
  
  ISET(dwDescPerBlock, IGET(dwBlockSize) / sizeof(CExt2GroupDesc));
  ISET(dwDescBlocks, (IGET(dwGroupsCount) + IGET(dwDescPerBlock) - 1) / IGET(dwDescPerBlock));
  //debugWin("m_info.dwDescBlocks=%lu",(DWORD)m_info.dwDescBlocks);

  // virtual blocks used in the abstract CFSBase
  SPUT(m_header,qwBlocksCount, IGET(dwTotalBlocksCount) * IGET(dwLogicalBlocksPerExt2Block));
  SPUT(m_header,qwBlockSize, LOGICAL_EXT2_BLKSIZE);
  SPUT(m_header,qwBitmapSize, ((SGET(m_header,qwBlocksCount+7)) / 8)+16);
  SPUT(m_header,qwUsedBlocks, (IGET(dwTotalBlocksCount) - Le32ToCpu(sb.s_free_blocks_count)) * IGET(dwLogicalBlocksPerExt2Block)); 

  //debugWin("freeblks=%lu", sb.s_free_blocks_count);
  strncpy(m_header.val.szLabel, sb.s_volume_name, 64);

  // features
  ISET(dwFeatureCompat, Le32ToCpu(sb.s_feature_compat));
  ISET(dwFeatureIncompat, Le32ToCpu(sb.s_feature_incompat));
  ISET(dwFeatureRoCompat, Le32ToCpu(sb.s_feature_ro_compat));

  // misc infos
  ISET(dwRevLevel, Le32ToCpu(sb.s_rev_level));	// Revision level
  memcpy(m_info.val.cUuid, sb.s_uuid, 16); // 128-bit uuid for volume

  //success_readSuperBlock:
  setSuperBlockInfos(true, true, SGET(m_header,qwUsedBlocks)*SGET(m_header,qwBlockSize), ((QWORD)Le32ToCpu(sb.s_free_blocks_count)) * ((QWORD)IGET(dwBlockSize)));
  showDebug(1, "end success\n");  
  RETURN_int(0);  

 error_readSuperBlock:
//  g_interface -> ErrorReadingSuperblock(errno);
  return -1; //THROW(ERR_WRONG_FS);
}

// =======================================================
int CExt2Part::readBitmap(/*COptions *options*/) // FULLY WORKING
{
  BEGIN;
 
  DWORD i, j;
  CExt2GroupDesc *desc;
  int nRes;
  DWORD dwBlocksInThisGroup;
  DWORD dwBootBlocks;
  char *cTempBitmap;
  DWORD dwBit, dwByte;
  DWORD dwExt2DataBlock;
  char *cPtr;

  // debug
  DWORD dwUsed;
  DWORD dwFree;

  dwBootBlocks = IGET(dwFirstBlock) / IGET(dwLogicalBlocksPerExt2Block);
  //debugWin("dwBootBlocks=%lu and m_info.dwLogicalBlocksPerExt2Block=%lu",dwBootBlocks,m_info.dwLogicalBlocksPerExt2Block);

  cTempBitmap = new char[((IGET(dwTotalBlocksCount)+7)/8)+4096];
  if (!cTempBitmap)
    goto error_readBitmap;

  // init bitmap size
  nRes = m_bitmap.init(SGET(m_header,qwBitmapSize));
  if (nRes == -1)
    goto error_readBitmap;

  // load group descriptors
  desc = new CExt2GroupDesc[IGET(dwGroupsCount)+ IGET(dwDescPerBlock)];
  if (!desc)
    goto error_readBitmap;

  // for each descriptor BLOCK (not group descriptor!)
  for (cPtr=(char*)desc, i=0; i < IGET(dwDescBlocks); i++,cPtr+=IGET(dwBlockSize))
    {
      nRes = readData(cPtr, ((QWORD)IGET(dwBlockSize)) * ((QWORD)(IGET(dwFirstBlock)+1+i)), IGET(dwBlockSize));
      if (nRes == -1)
	goto error_readBitmap;
    }

  dwUsed=0;
  dwFree=0;
  
  for (i = 0; i < IGET(dwGroupsCount); i++) 
    {
      if (IGET(dwFirstBlock)+((i+1)* IGET(dwBlocksPerGroup)) > IGET(dwTotalBlocksCount))
	dwBlocksInThisGroup = (IGET(dwTotalBlocksCount)-IGET(dwFirstBlock)) - (i*IGET(dwBlocksPerGroup));
      else
	dwBlocksInThisGroup = IGET(dwBlocksPerGroup);

      if (Le32ToCpu(desc[i].bg_block_bitmap))
	{
	  // -- read the bitmap block
	  nRes = readData(cTempBitmap+(i*IGET(dwBlockSize)), ((QWORD)IGET(dwBlockSize)) * ((QWORD)Le32ToCpu(desc[i].bg_block_bitmap)), IGET(dwBlockSize));
	  if (nRes == -1)
	    goto error_readBitmap;
	}
      else
	{
	  memset(cTempBitmap+(i*IGET(dwBlockSize)), 0, IGET(dwBlockSize));
	}
    }

  // convert bitmap to little endian
  DWORD *dwPtr;
  DWORD dwLen;
  dwLen = sizeof(cTempBitmap) / sizeof(DWORD);
  dwPtr = (DWORD *)cTempBitmap;
  for (i=0; i < dwLen; i++, dwPtr++)
    *dwPtr = CpuToLe32(*dwPtr);

  // bitmap is full of 0 at init, then we just have to
  // write 1 for used blocks

  // the boot block of 1024 bytes = used
  for (i=0; i < dwBootBlocks; i++)
    m_bitmap.setBit(i, true);

  dwUsed=0; dwFree=0;
  for (i=dwBootBlocks, dwExt2DataBlock=0; dwExt2DataBlock < ( IGET(dwTotalBlocksCount)- IGET(dwFirstBlock)); dwExt2DataBlock++)
    {	
      dwBit = dwExt2DataBlock % 8;
      dwByte = (dwExt2DataBlock - dwBit) / 8;
       
      if ((cTempBitmap[dwByte] & (1 << dwBit)) != 0)
	{
	  for (j=0; j < IGET(dwLogicalBlocksPerExt2Block); j++, i++)
	    m_bitmap.setBit(i, true);
	  dwUsed++;
	}
      else
	{
	  for (j=0; j < IGET(dwLogicalBlocksPerExt2Block); j++, i++)
	    m_bitmap.setBit(i, false);
	  dwFree++; 
	}
    }

  //debugWin("used=%lu\nfree=%lu\ntotal=%lu",dwUsed,dwFree,dwUsed+dwFree);
  calculateSpaceFromBitmap();

  //success_readBitmap:
  delete []cTempBitmap;
  showDebug(1, "end success\n");  
  RETURN_int(0); // success

 error_readBitmap:
  delete []cTempBitmap;
  showDebug(1, "end error\n");  
//  g_interface->msgBoxError(i18n("There was an error while reading the bitmap"));
  return -1; //THROW(ERR_READ_BITMAP);
}

