/***************************************************************************
                          cfatpart.h  -  description
                             -------------------
    begin                : Thu Jul 13 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef FS_FAT_H
#define FS_FAT_H

#include "global.h"
#include "fs_base.h"

#define MAX_BOOT_SIZE		      4096
#define fat32to28(A) (A & 0x0FFFFFFF)

#define LOGICAL_FAT_BLKSIZE 512 // the size of the abstract block used in CFSBase


#define ATTR_RO      1  // read-only
#define ATTR_HIDDEN  2  // hidden
#define ATTR_SYS     4  // system
#define ATTR_VOLUME  8  // volume label
#define ATTR_DIR     16 // directory
#define ATTR_ARCH    32 // archived

// ================================================
struct CDirEntry
{
  s8	szName[11];	// name and extension
  u8	cAttr;		// attribute bits
  u8    lcase;		// Case for base and extension
  u8	ctime_ms;	// Creation time, milliseconds
  u16	ctime;		// Creation time
  u16	cdate;		// Creation date
  u16	adate;		// Last access date
  u16   starthi;	// High 16 bits of cluster in FAT32 
  u16	time,date,start;// time, date and first cluster
  u32	size;		// file size (in bytes)
} PACKED;

// ================================================
struct CFatFsInfoSector
{
  DWORD       dwMagic;          // Magic for info sector ('RRaA') 
  BYTE        junk[0x1dc];
  DWORD       reserved1;      // Nothing as far as I can tell 
  DWORD       signature;      // 0x61417272 ('rrAa') 
  DWORD       dwFreeClusters;  // Free cluster count.  -1 if unknown 
  DWORD       dwNextCluster;   // Most recently allocated cluster. 
  DWORD       reserved2[3];
  WORD       reserved3;
  WORD       boot_sign;
};

// ================================================
    struct CInfoFatHeadeVal
    {
      DWORD dwTotalSectorsCount;
      DWORD dwClustersCount;
      DWORD dwRootDirSectors;
      DWORD dwRootEntriesCount;
      DWORD dwSectorsPerFAT;
      DWORD dwDataSectors;
      DWORD dwFileSystem;
      DWORD dwUsedClusters;
      DWORD dwDamagedClusters;
      DWORD dwFreeClusters;
      DWORD dwBytesPerFatEntry;
      DWORD dwReserved;		// reserved to align data
      WORD wBytesPerSector;
      WORD wReservedSectorsCount;
      WORD wRootEntriesCount;
      BYTE cSectorsPerCluster;
      BYTE cNumberOfFATs;
      WORD wFsInfoSector;
      //BYTE cReserved[16326]; // Adjust to fit with total header size
    } PACKED; 

struct CInfoFatHeader // size must be 16384 (adjust the reserved data)
{
  union
  {
    CInfoFatHeadeVal val;
    char cReserved[INFOS_STRUCT_SIZE];
  };
};

// ================================================
class CFatPart: public CFSBase
{
 public:
  CFatPart(char *szDevice, FILE *fDeviceFile, QWORD qwPartSize);
  ~CFatPart();
  
  static  int detect(BYTE *cBootSect, char *szFileSystem); 
  virtual void printfInformations();
  virtual int readBitmap(/*COptions *options*/);
  virtual int readSuperBlock();
  virtual int fsck();
  virtual void* getInfos() {return (void*)&m_info;}

  int compareFatCopies();
  int readFSInfo(WORD wFsInfoSector);

 private:
  CInfoFatHeader m_info;
};

#endif // FS_FAT_H
