/***************************************************************************
              fs_hfs.cpp  - HFS/HFS+: MacOs, Hierarchical File System
                             -------------------
    begin                : Thu Jul 13 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "fs_hfs.h"
#include "global.h"
#include "errors.h"
#include "exceptions.h"
#include "misc.h"
#include "imagefile.h"
#include "minmax.h"

// =======================================================
CHfsPart::CHfsPart(char *szDevice, FILE *fDeviceFile, QWORD qwPartSize): CFSBase(szDevice, fDeviceFile, qwPartSize)
{
  BEGIN;
  RETURN;
}

// =======================================================
CHfsPart::~CHfsPart()
{
  BEGIN;
  RETURN;
}

// =======================================================
int CHfsPart::fsck()
{
  BEGIN;
  RETURN_int(0);
}

// =======================================================
int CHfsPart::readSuperBlock()
{
  BEGIN;

  QWORD qwBlocksInRegularAlloc;
  hfsSuperBlock sb, sb2;
  int nRes;

  // read superblock
  nRes = readData(&sb, 1024, sizeof(sb));
  if (nRes == -1)
    {
      //localDebug("ERROR 1\n");
      //      g_interface -> ErrorReadingSuperblock(errno);
      return -1; //THROW(ERR_READING, (DWORD)0, errno);
    }

  if (Be16ToCpu(GET_WORD(&sb.drSigWord)) != HFS_SUPER_MAGIC)
    {
      //localDebug("ERROR 2\n");
      //      g_interface -> ErrorReadingSuperblock(errno);
      return -1; //THROW(ERR_READING, (DWORD)0, errno);
    }

  // init
  memset(&m_info, 0, sizeof(CInfoHfsHeader));

  ISET(qwAllocCount, BeToCpu(GET_WORD(&sb.drNmAlBlks))); 
  ISET(dwAllocSize, BeToCpu(GET_DWORD(&sb.drAlBlkSize))); 
  ISET(dwBlocksPerAlloc, IGET(dwAllocSize) / 512); 
  ISET(qwBitmapSectLocation, (QWORD)BeToCpu(GET_WORD(&sb.drVBMSt)));
  ISET(qwFirstAllocBlock, (QWORD)BeToCpu(GET_WORD(&sb.drAlBlSt)));
  ISET(qwFreeAllocs, (QWORD)BeToCpu(GET_WORD(&sb.drFreeBks)));
  
  // if not a 512 multiple
  if ((IGET(dwAllocSize) % 512) != 0)
    return -1; //THROW(ERR_READING, (DWORD)0, errno);

  // check the last but one block is a superblock
  qwBlocksInRegularAlloc = IGET(dwBlocksPerAlloc) * IGET(qwAllocCount);
  nRes = readData(&sb2, (IGET(qwFirstAllocBlock)+qwBlocksInRegularAlloc)*512LL, sizeof(sb2));
  if (nRes == -1)
    {
      //localDebug("ERROR 3\n");
      //      g_interface -> ErrorReadingSuperblock(errno);
      return -1; //THROW(ERR_READING, (DWORD)0, errno);
    }

  if (Be16ToCpu(GET_WORD(&sb.drSigWord)) != HFS_SUPER_MAGIC)
    {
      //localDebug("ERROR 4\n");
      //      g_interface -> ErrorReadingSuperblock(errno);
      return -1; //THROW(ERR_READING, (DWORD)0, errno);
    }

  // show infos
  //localDebug("m_info.qwAllocCount=%llu\n", IGET(qwAllocCount));
  //localDebug("m_info.dwAllocSize=%lu=%lu sectors\n", IGET(dwAllocSize), IGET(dwBlocksPerAlloc));
  //localDebug("m_info.qwBitmapSectLocation=%llu\n", IGET(qwBitmapSectLocation));
  //localDebug("m_info.qwFirstAllocBlock=%llu\n", IGET(qwFirstAllocBlock));
  //localDebug("m_info.qwFreeAllocs=%llu\n", IGET(qwFreeAllocs));

  // label == pascal string
  memset(m_header.val.szLabel, 0, 64);
  memcpy(m_header.val.szLabel, ((char*)&sb.drVN)+1, min(((BYTE*)&sb.drVN)[0], 63));
  //localDebug("label=[%s]\n", m_header.val.szLabel);

  // calculate sectors count
  SPUT(m_header,qwBlocksCount, IGET(qwFirstAllocBlock) // first reserved (2 boot blocks, 1 superblock, ...)
       + (IGET(dwBlocksPerAlloc) * IGET(qwAllocCount)) // allocation blocks
       + 1 // alternate superblock
       + 1); // last volume of sector
  //localDebug("m_header.qwBlocksCount = %llu\n", SGET(m_header,qwBlocksCount));
   
  SPUT(m_header,qwBlockSize, 512);

  SPUT(m_header,qwBitmapSize, (SGET(m_header,qwBlocksCount)+7)/8);
  
  QWORD qwUsedBlocks = IGET(qwFirstAllocBlock) + 2 + ((IGET(qwAllocCount)-IGET(qwFreeAllocs)) * IGET(dwBlocksPerAlloc));
  QWORD qwFreeBlocks = IGET(qwFreeAllocs) * IGET(dwBlocksPerAlloc);
  setSuperBlockInfos(true, true, qwUsedBlocks*512, qwFreeBlocks*512);

  RETURN_int(0);  
}

// =======================================================
void CHfsPart::printfInformations()
{
  char szFullText[8192];
  char szText[8192];
  char szTemp1[1024];
  
  getStdInfos(szText, sizeof(szText), true);
  
  SNPRINTF(szFullText, i18n("%s" // standard infos
			    "Allocation Group count:.......%llu\n"
			    "Blocks per Allocation Group...%lu\n"
			    "Allocation Group size:........%s\n"
			    "First allocation block:.......%llu\n"),
	   szText, IGET(qwAllocCount), IGET(dwBlocksPerAlloc),
	   formatSize(IGET(dwAllocSize),szTemp1), IGET(qwFirstAllocBlock));
    
//  g_interface->msgBoxOk(i18n("HFS informations"), szFullText);
}			

// =======================================================
int CHfsPart::readBitmap(/*COptions *options*/)
{
  BEGIN;

  QWORD qwBlocksInRegularAlloc;
  QWORD i, j, k;
  BYTE *cTempData;
  DWORD *dwTempData;
  QWORD qwCurSector;
  QWORD qwLastAllocSector;
  DWORD dwBitmapData;
  int nRes;
  
  // init bitmap
  const DWORD dwHighBit= 1<<31;
  //localDebug("m_header.qwBitmapSize=%llu\n", SGET(m_header,qwBitmapSize));
  m_bitmap.init(SGET(m_header,qwBitmapSize)+1024);

  // special sectors (begin and end of volume) are used
  qwBlocksInRegularAlloc = IGET(dwBlocksPerAlloc) * IGET(qwAllocCount);
  qwLastAllocSector = IGET(qwFirstAllocBlock)+qwBlocksInRegularAlloc;

  for (i=0; i < IGET(qwFirstAllocBlock); i++)
    {
      //localDebug("Block [%llu] is special\n", i);
      m_bitmap.setBit(i, true);
    }
  //localDebug("Blocks [%llu] and [%llu] are special\n", qwLastAllocSector, qwLastAllocSector+1);
  m_bitmap.setBit(qwLastAllocSector, true); // alternate superblock
  m_bitmap.setBit(qwLastAllocSector+1LL, true); // last volume of sector

  // regular allocation blocks
  cTempData = (BYTE*)malloc(SGET(m_header,qwBitmapSize)+1024);
  if (!cTempData)
    return -1; //THROW(ERR_NOMEM);

  nRes = readData(cTempData, (QWORD)(3*512), SGET(m_header,qwBitmapSize)+512);
  if (nRes == -1)
    {
      //localDebug("ERROR 5\n");
//      g_interface -> ErrorReadingSuperblock(errno);
      return -1; //THROW(ERR_READING, (DWORD)0, errno);
    }

  dwTempData = (DWORD*)cTempData;
  qwCurSector = IGET(qwFirstAllocBlock);

  for (i=0; i < IGET(qwAllocCount); i+=32)
    {
      dwBitmapData = Be32ToCpu(*dwTempData);
      dwTempData++;
      
      for (j=0; j < 32; j++)
	{
	  if (dwBitmapData & dwHighBit) // block is free
	    {
	      for (k=0; k < IGET(dwBlocksPerAlloc); k++, qwCurSector++)
		{
		  if (qwCurSector < qwLastAllocSector)
		    {
		      m_bitmap.setBit(qwCurSector, true);
		      //localDebug("used -> %llu\n", qwCurSector);
		    }
		}
	    }
	  else // block is used
	    {
	      for (k=0; k < IGET(dwBlocksPerAlloc); k++, qwCurSector++)
		{
		  if (qwCurSector < qwLastAllocSector)
		    {
		      m_bitmap.setBit(qwCurSector, false);
		      //localDebug("free -> %llu\n", qwCurSector);

		    }		
		}
	    }	  
	  dwBitmapData<<=1;
	}
    }
  
  free(cTempData);

  calculateSpaceFromBitmap();

  RETURN_int(0);  
}			

