/***************************************************************************
                          chpfspart.h  -  description
                             -------------------
    begin                : Thu Jul 13 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef FS_HPFS_H
#define FS_HPFS_H

#include "global.h"
#include "fs_base.h"

/* Notation */

typedef u32 secno;			/* sector number, partition relative */

typedef secno dnode_secno;		/* sector number of a dnode */
typedef secno fnode_secno;		/* sector number of an fnode */
typedef secno anode_secno;		/* sector number of an anode */

/* sector 0 */

/* The boot block is very like a FAT boot block, except that the
   29h signature byte is 28h instead, and the ID string is "HPFS". */

#define BB_MAGIC 0xaa55

struct hpfs_boot_block
{
  u8 jmp[3];
  u8 oem_id[8];
  u8 bytes_per_sector[2];	/* 512 */
  u8 sectors_per_cluster;
  u8 n_reserved_sectors[2];
  u8 n_fats;
  u8 n_rootdir_entries[2];
  u8 n_sectors_s[2];
  u8 media_byte;
  u16 sectors_per_fat;
  u16 sectors_per_track;
  u16 heads_per_cyl;
  u32 n_hidden_sectors;
  u32 n_sectors_l;		/* size of partition */
  u8 drive_number;
  u8 mbz;
  u8 sig_28h;		/* 28h */
  u8 vol_serno[4];
  u8 vol_label[11];
  u8 sig_hpfs[8];		/* "HPFS    " */
  u8 pad[448];
  u16 magic;			/* aa55 */
};


/* sector 16 */

/* The super block has the pointer to the root directory. */

#define SB_MAGIC 0xf995e849

struct hpfs_super_block
{
  u32 magic;			/* f995 e849 */
  u32 magic1;			/* fa53 e9c5, more magic? */
  //u32 huh202;			/* ?? 202 = N. of B. in 1.00390625 S.*/
  char version;				/* version of a filesystem  usually 2 */
  char funcversion;			/* functional version - oldest version
  					   of filesystem that can understand
					   this disk */
  u16 zero;		/* 0 */
  fnode_secno root;			/* fnode of root directory */
  secno n_sectors;			/* size of filesystem */
  u32 n_badblocks;			/* number of bad blocks */
  secno bitmaps;			/* pointers to free space bit maps */
  u32 zero1;			/* 0 */
  secno badblocks;			/* bad block list */
  u32 zero3;			/* 0 */
  time_t last_chkdsk;			/* date last checked, 0 if never */
  /*u32 zero4;*/			/* 0 */
  time_t last_optimize;			/* date last optimized, 0 if never */
  secno n_dir_band;			/* number of sectors in dir band */
  secno dir_band_start;			/* first sector in dir band */
  secno dir_band_end;			/* last sector in dir band */
  secno dir_band_bitmap;		/* free space map, 1 dnode per bit */
  char volume_name[32];			/* not used */
  secno user_id_table;			/* 8 preallocated sectors - user id */
  u32 zero6[103];			/* 0 */
};

/* sector 17 */

/* The spare block has pointers to spare sectors.  */

#define SP_MAGIC 0xf9911849

struct hpfs_spare_block
{
  u32 magic;			/* f991 1849 */
  u32 magic1;			/* fa52 29c5, more magic? */

  u32 dirty: 1;			/* 0 clean, 1 "improperly stopped" */
  /*u32 flag1234: 4;*/		/* unknown flags */
  u32 sparedir_used: 1;		/* spare dirblks used */
  u32 hotfixes_used: 1;		/* hotfixes used */
  u32 bad_sector: 1;		/* bad sector, corrupted disk (???) */
  u32 bad_bitmap: 1;		/* bad bitmap */
  u32 fast: 1;			/* partition was fast formatted */
  u32 old_wrote: 1;		/* old version wrote to partion */
  u32 old_wrote_1: 1;		/* old version wrote to partion (?) */
  u32 install_dasd_limits: 1;	/* HPFS386 flags */
  u32 resynch_dasd_limits: 1;
  u32 dasd_limits_operational: 1;
  u32 multimedia_active: 1;
  u32 dce_acls_active: 1;
  u32 dasd_limits_dirty: 1;
  u32 flag67: 2;
  u8 mm_contlgulty;
  u8 unused;

  secno hotfix_map;			/* info about remapped bad sectors */
  u32 n_spares_used;		/* number of hotfixes */
  u32 n_spares;			/* number of spares in hotfix map */
  u32 n_dnode_spares_free;		/* spare dnodes unused */
  u32 n_dnode_spares;		/* length of spare_dnodes[] list,
					   follows in this block*/
  secno code_page_dir;			/* code page directory block */
  u32 n_code_pages;		/* number of code pages */
  /*u32 large_numbers[2];*/	/* ?? */
  u32 super_crc;			/* on HPFS386 and LAN Server this is
  					   checksum of superblock, on normal
					   OS/2 unused */
  u32 spare_crc;			/* on HPFS386 checksum of spareblock */
  u32 zero1[15];			/* unused */
  dnode_secno spare_dnodes[100];	/* emergency free dnode list */
  u32 zero2[1];			/* room for more? */
};
// ================================================
    struct CInfoHpfsHeaderVal
    {
      DWORD dwBitmapPointer;
      DWORD dwBitmapQuadBlocksCount;
      char cHpfsVersion;

      //BYTE cReserved1[7];
      //BYTE cReserved[16368]; // Adjust to fit with total header size
    } PACKED; 

struct CInfoHpfsHeader // size must be 16384 (adjust the reserved data)
{
  union
  {
    CInfoHpfsHeaderVal val;
    char cReserved[INFOS_STRUCT_SIZE];
  };
};

// ================================================
class CHpfsPart : public CFSBase
{
 public:
  CHpfsPart(char *szDevice, FILE *fDeviceFile, QWORD qwPartSize);
  ~CHpfsPart();
  
  static  int detect(BYTE *cBootSect, char *szFileSystem); 
  virtual void printfInformations();
  virtual int readBitmap(/*COptions *options*/);
  virtual int readSuperBlock();
  virtual int fsck();
  virtual void* getInfos() {return (void*)&m_info;}

 private:
  int getBitmapQuadBlocksCount(DWORD *dwBitmapQuadBlocksCount);
  
 private:
  CInfoHpfsHeader m_info;
};

#endif // FS_HPFS_H
