/***************************************************************************
                          cntfspart.h  -  description
                             -------------------
    begin                : Mon May 22 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef FS_NTFS_H
#define FS_NTFS_H

#include "global.h"
#include "fs_base.h"

#define MAX_NTFS_BLOCK_SIZE 8192

#define NTFS_GETU8(p)      (*(u8*)(p))
#define NTFS_GETU16(p)     ((u16)Le16ToCpu(*(u16*)(p)))
#define NTFS_GETU24(p)     ((u32)NTFS_GETU16(p) | ((u32)NTFS_GETU8(((char*)(p))+2)<<16))
#define NTFS_GETU32(p)     ((u32)Le32ToCpu(*(u32*)(p)))
#define NTFS_GETU40(p)     ((u64)NTFS_GETU32(p)|(((u64)NTFS_GETU8(((char*)(p))+4))<<32))
#define NTFS_GETU48(p)     ((u64)NTFS_GETU32(p)|(((u64)NTFS_GETU16(((char*)(p))+4))<<32))
#define NTFS_GETU56(p)     ((u64)NTFS_GETU32(p)|(((u64)NTFS_GETU24(((char*)(p))+4))<<32))
#define NTFS_GETU64(p)     ((u64)Le64ToCpu(*(u64*)(p)))

#define NTFS_GETS8(p)        ((*(s8*)(p)))
#define NTFS_GETS16(p)       ((s16)Le16ToCpu(*(s16*)(p)))
#define NTFS_GETS24(p)       (NTFS_GETU24(p) < 0x800000 ? (s32)NTFS_GETU24(p) : (s32)(NTFS_GETU24(p) - 0x1000000))
#define NTFS_GETS32(p)       ((s32)Le32ToCpu(*(s32*)(p)))
#define NTFS_GETS40(p)       (((s64)NTFS_GETU32(p)) | (((s64)NTFS_GETS8(((char*)(p))+4)) << 32))
#define NTFS_GETS48(p)       (((s64)NTFS_GETU32(p)) | (((s64)NTFS_GETS16(((char*)(p))+4)) << 32))
#define NTFS_GETS56(p)       (((s64)NTFS_GETU32(p)) | (((s64)NTFS_GETS24(((char*)(p))+4)) << 32))
#define NTFS_GETS64(p)	     ((s64)NTFS_GETU64(p))

// ================================================
class CNtfsRunList
{
 private:
  QWORD *m_qwOffset;
  QWORD m_qwClustersCount;
  
 public:
  CNtfsRunList();
  ~CNtfsRunList();
  
 public:
  int setCluster(QWORD qwNumber, QWORD qwOffset); // ex: cluser 0 position is 12589
  int init(QWORD qwClustersCount);
  void freeMemory();
  void show();
  QWORD offset(QWORD qwCluster);
  QWORD clustersCount();
};

// ================================================
struct CInfoNtfsHeaderVal
{
  QWORD qwTotalSectorsCount;
  QWORD qwLCNOfMftDataAttrib;
  DWORD dwFileRecordSize;
  DWORD dwClusterSize;
  
  WORD nBytesPerSector;
  WORD nNtfsVersion; 
  BYTE cSectorsPerCluster;
} PACKED; 

struct CInfoNtfsHeader // size must be 16384 (adjust the reserved data)
{
  union
  {
    CInfoNtfsHeaderVal val;
    char cReserved[INFOS_STRUCT_SIZE];
  };
};

// ================================================
class CNtfsPart : public CFSBase
{
 public:
  CNtfsPart(char *szDevice, FILE *fDeviceFilek, QWORD qwPartSize);
  ~CNtfsPart();

  static  int detect(BYTE *cBootSect, char *szFileSystem); 
  virtual void printfInformations();
  virtual int readBitmap(/*COptions *options*/);
  virtual int readSuperBlock();
  virtual int fsck();
  virtual void* getInfos() {return (void*)&m_info;}

  // file reading options
  int readDataAttribute(BYTE *cAttribData, CNtfsRunList *runlist, QWORD *qwBitmapSize, QWORD *qwAllocatedSize);
  int readFilenameAttribute(BYTE *cAttribData, char *szFilename);
  int readFileRecord(BYTE *cRecordData, CNtfsRunList *runlist, QWORD *qwBitmapSize, QWORD *qwAllocatedSize);
  int processRunList(BYTE *cAttribData, CNtfsRunList *runlist, QWORD *qwAllocatedSize, QWORD *qwRealSize);
  int searchForRecordFilename(BYTE *cFileRecord, char *szFilename, char *szUnicodeName, int nUnicodeNameLen);
  int checkFilenameForRecordIs(BYTE *cRecordData, char *szFilename, bool bShowDebug = true);
  int readVolumeLabel();

private:
  // informations
  CInfoNtfsHeader m_info;
};

#endif // FS_NTFS_H
