/***************************************************************************
                          creiserpart.h  -  description
                             -------------------
    begin                : Thu Jul 13 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef FS_REISER_H
#define FS_REISER_H

#include "global.h"
#include "fs_base.h"

#define REISERFS_SUPER_MAGIC_STRING "ReIsErFs"
#define REISER2FS_SUPER_MAGIC_STRING "ReIsEr2Fs"
#define REISERFS_DISK_OFFSET_IN_BYTES (64 * 1024)

#define MAX_REISER_BLOCK_SIZE						8192

#define VER_3_5 1
#define VER_3_6 2

// ================================================
struct reiserfs_super_block
{
  u32 s_block_count;			/* blocks count         */
  u32 s_free_blocks;                  /* free blocks count    */
  u32 s_root_block;           	/* root block number    */
  u32 s_journal_block;           	/* journal block number    */
  u32 s_journal_dev;           	/* journal device number  */
  u32 s_orig_journal_size; 		/* size of the journal on FS creation.  used to make sure they don't overflow it */
  u32 s_journal_trans_max ;           /* max number of blocks in a transaction.  */
  u32 s_journal_block_count ;         /* total size of the journal. can change over time  */
  u32 s_journal_max_batch ;           /* max number of blocks to batch into a trans */
  u32 s_journal_max_commit_age ;      /* in seconds, how old can an async commit be */
  u32 s_journal_max_trans_age ;       /* in seconds, how old can a transaction be */
  u16 s_blocksize;                   	/* block size           */
  u16 s_oid_maxsize;			/* max size of object id array, see get_objectid() commentary  */
  u16 s_oid_cursize;			/* current size of object id array */
  u16 s_state;                       	/* valid or error       */
  char s_magic[12];                     /* reiserfs magic string indicates that file system is reiserfs */
  u32 s_hash_function_code;		/* indicate, what hash fuction is being use to sort names in a directory*/
  u16 s_tree_height;                  /* height of disk tree */
  u16 s_bmap_nr;                      /* amount of bitmap blocks needed to address each block of file system */
  u16 s_reserved;
};

// ================================================
struct CInfoReiserHeaderVal
{
  DWORD dwVersion;
  DWORD dwBitmapBlocksCount;
} PACKED; 

struct CInfoReiserHeader // size must be 16384 (adjust the reserved data)
{
  union
  {
    CInfoReiserHeaderVal val;
    char cReserved[INFOS_STRUCT_SIZE];
  };
};

// ================================================
class CReiserPart: public CFSBase
{
public:
  CReiserPart(char *szDevice, FILE *fDeviceFile, QWORD qwPartSize);
  ~CReiserPart();
  
  static  int detect(BYTE *cBootSect, char *szFileSystem); 
  virtual void printfInformations();
  virtual int readBitmap(/*COptions *options*/);
  virtual int readSuperBlock();
  virtual int fsck();
  virtual void* getInfos() {return (void*)&m_info;}

 private:
  CInfoReiserHeader m_info;
};

#endif // FS_REISER_H
