/***************************************************************************
       fs_base.cpp  -  Base class of all File System with common code
                             -------------------
    begin                : Thu Jul 13 2000
    copyright            : (C) 2000 by Franois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <sys/param.h>

#include "fs_work.h"
#include "imagefile.h"
#include "common.h"
#include "cbitmap.h"
#include "exceptions.h"
#include "misc.h"
#include "interface_newt.h"

// =======================================================
/*CFSWork::CFSWork(char *szDevice, FILE *fDeviceFile, QWORD qwPartSize): CFSBase(szDevice, fDeviceFile, qwPartSize)
{
  BEGIN;

  m_fDeviceFile = fDeviceFile;
  strncpy(m_szDevice, szDevice, MAX_DEVICENAMELEN);
  memset(&m_header, 0, sizeof(CLocalHeader));
  m_qwPartSize = qwPartSize;
  showDebug(1, "m_qwPartSize=%llu = %llu * 512\n",m_qwPartSize,m_qwPartSize/512LL);
  m_bCheckSuperBlockUsedSize = false;
  m_bCheckSuperBlockFreeSize = false;
  m_bCanBeMore = false;

  RETURN;

}*/

// =======================================================
/*CFSWork::~CFSWork()
{
}*/

// =======================================================
// may throw: errno, readbitmap, reading, writing
int /*CFSWork::*/saveUsedBlocksToImageFile(CFSBase *fs, COptions *options, CPiTimer *timer)
{
  BEGIN;

  char cBuffer[COPYBUF_SIZE+1];
  BOOL bUsed;
  DWORD dwCount;
  int nRes;
  QWORD i;
  DWORD dwStats;
  DWORD dwTemp;

  DWORD dwSavedBlocks = 0L;
  DWORD dwBlocksInBuffer;
  DWORD dwDataSize;
  QWORD qwBlocksToCopyCount;

  // init ESC key
  //g_bSigInt = false;

  // go to the beginning of the disk
  nRes = fseek(fs->m_fDeviceFile, 0L, SEEK_SET);
  if (nRes == -1)
    return -1; //THROW(ERR_ERRNO, errno);
  
  // init the bitmap optimizations
  fs->m_bitmap.setMaxFreeBlocksCopiedPerOper(MAX_FREE_BLOCKS_SKIPPED); // could be another value
  dwBlocksInBuffer = (COPYBUF_SIZE-(COPYBUF_SIZE % SGET(fs->m_header,qwBlockSize))) / (SGET(fs->m_header,qwBlockSize)); 
  fs->m_bitmap.setMaxUsedBlocksCopiedPerOper(dwBlocksInBuffer);
  showDebug(3, "CPY: setMax=%lu\n", dwBlocksInBuffer);

  if (options->bSaveFreeBlocks)
    qwBlocksToCopyCount = SGET(fs->m_header, qwBlocksCount);
  else
    qwBlocksToCopyCount = SGET(fs->m_header, qwUsedBlocks);

  dwStats = 0;
  i = 0;
  while (i < SGET(fs->m_header, qwBlocksCount)) // while there are blocks to copy
    {

      nRes = fs->m_bitmap.readBitmap(i, &bUsed, &dwCount, options->bSaveFreeBlocks);
      if (nRes == -1)
        return -1; //THROW(ERR_READ_BITMAP, i);

      // stop at the end of the partition
      dwCount = min(dwCount, SGET(fs->m_header, qwBlocksCount)-i);

      showDebug(3, "CPY: dwCount=%lu, bUsed=%d\n", dwCount, bUsed);

      if (bUsed) // if used blocks are marked in the bitmap
	{
	  dwDataSize = SGET(fs->m_header, qwBlockSize*dwCount);
	  
	  while (dwDataSize > 0)
	    {
	      dwTemp = min(dwDataSize, COPYBUF_SIZE);
	      
	      errno = 0;
	      nRes = fread(cBuffer, dwTemp, 1, fs->m_fDeviceFile);
	      if (nRes != 1)
		return -1; //THROW(ERR_READING, (DWORD)i, errno);

	      try { fs->m_image->write(KIND_BLOCKS, cBuffer, dwTemp, true); }
	      catch (CExceptions * excep)
		{
		  showDebug(1, "saveUsedBlocksToImageFile: excep %d\n", excep->GetExcept());
		  throw excep;
		}
	      dwDataSize -= dwTemp;
	    }


 	  // show stats
	  //if (dwStats == 8)
	  {
	    fs->m_guiSave->showStats(timer, SGET(fs->m_header, qwBlockSize), dwSavedBlocks, qwBlocksToCopyCount);
	    dwStats = 0;
	  }


	  dwStats++;

	  //checkForEscape();

	  // ---- end -----

	  dwSavedBlocks += dwCount;

	}
      else // if free blocks were marked
	{
	  nRes = fseek(fs->m_fDeviceFile, SGET(fs->m_header, qwBlockSize)*dwCount, SEEK_CUR);
	  if (nRes == -1)
            {
              return -1; //THROW(ERR_ERRNO, errno);
            }
	}

      i += dwCount;

      // draw progress
      fs->m_guiSave->showStats(timer, SGET(fs->m_header, qwBlockSize), dwSavedBlocks, SGET(fs->m_header, qwUsedBlocks));
    }

  fs->m_guiSave->showStats(timer, SGET(fs->m_header, qwBlockSize), dwSavedBlocks, qwBlocksToCopyCount);
  g_qwCopiedBytesCount = SGET(fs->m_header, qwBlockSize) * dwSavedBlocks; // for final message box

  RETURN_int(0);
}

// =======================================================
// may raise errno, check_num, check_crc,zer, writing, wrong_magic
// NEW optimized version
int /*CFSWork::*/restoreUsedBlocksToDisk(CFSBase *fs, COptions *options, CPiTimer *timer)
{
   BEGIN;

   char cZeroBuffer[ZERO_BLOCK_SIZE];
   char cBuffer[COPYBUF_SIZE+1];
   BOOL bUsed;
   DWORD dwCount;
   int nRes;
   QWORD i;
   DWORD dwBlocksInBuffer;
   DWORD dwTemp;
   DWORD dwTemp2;
   DWORD dwDataSize;
   DWORD dwSavedBlocks = 0L;
   //DWORD dwCRC;
   DWORD dwStats;

   memset(cZeroBuffer, 0, ZERO_BLOCK_SIZE);
   //dwCRC = increaseCrc32(0, NULL, 0);  

   // init ESC key
   //g_bSigInt = false;

   // go to the beginning of the disk
   nRes = fseek(fs->m_fDeviceFile, 0L, SEEK_SET);
   if (nRes == -1)
      return -1; //THROW(ERR_ERRNO, errno);

   // init the bitmap optimizations
   fs->m_bitmap.setMaxFreeBlocksCopiedPerOper(MAX_FREE_BLOCKS_SKIPPED); // could be another value
   dwBlocksInBuffer = (COPYBUF_SIZE-(COPYBUF_SIZE % SGET(fs->m_header, qwBlockSize))) / (SGET(fs->m_header, qwBlockSize)); 
   fs->m_bitmap.setMaxUsedBlocksCopiedPerOper(dwBlocksInBuffer);
   showDebug(3, "CPY: setMax=%lu\n", dwBlocksInBuffer);
  
   dwStats = 0;
   i = 0;
   while (i < SGET(fs->m_header, qwBlocksCount))
   {
      showDebug(1, "i = %lld, fs->m_header.qwBlocksCount = %lld\n", i, SGET(fs->m_header, qwBlocksCount));

      fs->m_bitmap.readBitmap(i, &bUsed, &dwCount, options->bSaveFreeBlocks);
      dwCount = min(dwCount, SGET(fs->m_header, qwBlocksCount)-i); // stop at the end of the partition

      showDebug(3, "CPY: dwCount=%lu, bUsed=%d\n", dwCount, bUsed);

      if (bUsed) // if used blocks are marked in the bitmap
      {
	 dwDataSize = SGET(fs->m_header, qwBlockSize*dwCount);
	 dwSavedBlocks += dwCount;

	 while (dwDataSize > 0)
	 {
	    dwTemp2 = min(dwDataSize, COPYBUF_SIZE);

	    try { fs->m_image->read(cBuffer, dwTemp2 /*dwDataSize*/, true); }
	    catch (CExceptions * excep)
	    {
	       showDebug(1, "restoreUsedBlocksToDisk: except %d\n",
			 excep -> GetExcept());
	       throw excep;
	    }	

	    errno = 0;
	    if (!options->bSimulateMode)
	    {
	       nRes = fwrite(cBuffer, dwTemp2 /*dwDataSize*/, 1, fs->m_fDeviceFile);
	       if (nRes != 1)
	       {
		  return -1; //THROW(ERR_WRITING, (DWORD) 0, errno);
	       }				
	    }
	    else
	       showDebug(10, "fwrite simulated\n");

	    dwDataSize -= dwTemp2;
	 }

	 // show stats
	 //if (dwStats == 8)
	 {
	    // update stats
	    fs->m_guiRestore->showStats(timer, SGET(fs->m_header, qwBlockSize), dwSavedBlocks,
				    SGET(fs->m_header, qwUsedBlocks), SGET(fs->m_header, qwBlocksCount), 
				    options->bEraseWithNull | options->bSaveFreeBlocks);
	    dwStats = 0;
	 }
	 dwStats++;

	 //checkForEscape();

	 showDebug(9, "end while dwDataSize\n");
	  
	 // ---- end -----
	  
      }
      else // if free blocks were marked in the bitmap
      {
	 if (options->bEraseWithNull == true) // if free and we must write 0 bytes on unused clusters
	 {
	    dwTemp = SGET(fs->m_header, qwBlockSize) * dwCount; // data to copy
	      
	    while (dwTemp > 0)
	    {
	       errno = 0;
	       if (!options->bSimulateMode)
	       {
		  nRes = fwrite(cZeroBuffer, min(ZERO_BLOCK_SIZE, dwTemp),
				1, fs->m_fDeviceFile);
		  if (nRes != 1)
		     return -1; //THROW(ERR_ZEROING, errno);
	       }
	       else
		  showDebug(10, "fwrite simulated\n");
		  
	       dwTemp -= min(ZERO_BLOCK_SIZE, dwTemp);
	    }
	    dwSavedBlocks += dwCount;
	 }
	 else // don't write zero bytes
	 {
	    nRes = fseek(fs->m_fDeviceFile, SGET(fs->m_header, qwBlockSize)*dwCount, SEEK_CUR);
	    if (nRes == -1)
	       return -1; //THROW(ERR_ERRNO, errno);
	 }			
      }
      i += dwCount;
   }
  
   // update stats
   fs->m_guiRestore->showStats(timer, SGET(fs->m_header, qwBlockSize), dwSavedBlocks,
			   SGET(fs->m_header, qwUsedBlocks), SGET(fs->m_header, qwBlocksCount), options->bEraseWithNull | options->bSaveFreeBlocks);
   g_qwCopiedBytesCount = SGET(fs->m_header, qwBlockSize) * dwSavedBlocks; // for final message box
  
   RETURN_int(0);
}

// =======================================================
int /*CFSWork::*/saveToImage(CFSBase *fs, CImage *image, COptions *options, CSavingWindow *gui, CPiTimer *timer)
{
   BEGIN;
  
   fs->m_image = image;
   fs->m_guiSave = gui;
  
   // ---- 0. Initializations
  
   // read informations
   showDebug(3, "TRACE: 001 (init)\n");  
   //g_interface -> StatusLine(i18n("reading partition properties"));
   try { fs->readSuperBlock(); }
   catch ( CExceptions * excep )
   {
      showDebug(1, "saveToImage: excep %d\n", excep->GetExcept());

      if (excep->GetExcept() == ERR_BLOCKSIZE)
	 throw excep;
      else
	 return -1; //THROW(ERR_READ_SUPERBLOCK);
   }	
  
   // check the file system
   showDebug(3, "TRACE: 002 (fsck)\n");  
   if (options->bCheckBeforeSaving == true)
   {
      g_interface -> StatusLine(i18n("checking the file system with fsck"));

      try { fs->fsck(); }
      catch ( CExceptions * excep )
      { return -1; /*THROW(ERR_FSCHK);*/ }
   }

   // read bitmap into memory (must be done before writing header)
   showDebug(3, "TRACE: 003 (readBitmap)\n");  
   showDebug(1, "begin of readBitmap()\n");
   fs->readBitmap(/*options*/);
   showDebug(1, "end of readBitmap()\n");
  
   // check the bitmap has been read without error
   fs->checkBitmapInfos();

   // print informations	
   if (options->bBatchMode == false)
      fs->printfInformations();
  
   // write header
   showDebug(3, "TRACE: 004 (writeHeader)\n");  
   //g_interface -> StatusLine(i18n("writing header"));
   try { fs->m_image->writeMagic(KIND_LOCALHEAD, MAGIC_BEGIN_LOCALHEADER); }
   catch ( CExceptions * excep )
   { throw excep; }
   errno = 0;
   try { fs->m_image->write(KIND_LOCALHEAD, &fs->m_header, sizeof(CLocalHeader), true); }
   catch ( CExceptions * excep )
   {
      showDebug(1, "saveToImage: excep %d\n", excep->GetExcept());
      throw excep;
   }

   // write CRC of CLocaleader
   fs->m_image->writeCRC(KIND_LOCALHEAD, (char*)&fs->m_header, sizeof(CLocalHeader));
  
   // copy bitmap to image file
   showDebug(3, "TRACE: 005 (writeBitmap)\n");  
   try { fs->m_image->writeMagic(KIND_HEADER, MAGIC_BEGIN_BITMAP); }
   catch ( CExceptions * excep )
   { throw excep; }

   try { fs->m_image->write(KIND_HEADER, fs->m_bitmap.data(), SGET(fs->m_header, qwBitmapSize), true); }
   catch ( CExceptions * excep )
   {
      showDebug(1, "saveToImage: excep %d\n", excep->GetExcept());
      throw excep;
   }

   // write informations specific for the FS
   showDebug(3, "TRACE: 006 (writeInfo)\n");  
   try { fs->m_image->writeMagic(KIND_HEADER, MAGIC_BEGIN_INFO); }
   catch ( CExceptions * excep )
   { throw excep; }

   try { fs->m_image->write(KIND_HEADER, (char*)fs->getInfos(), INFOS_STRUCT_SIZE, true); }
   catch ( CExceptions * excep )
   {
      showDebug(1, "saveToImage: excep %d\n", excep->GetExcept());
      throw excep;
   }
  
   // write CRC of info
   fs->m_image->writeCRC(KIND_HEADER, (char*)fs->getInfos(), INFOS_STRUCT_SIZE);

   // write magic string
   try { fs->m_image->writeMagic(KIND_HEADER, MAGIC_BEGIN_DATABLOCKS); }
   catch ( CExceptions * excep )
   {
      showDebug(1, "saveToImage: excep %d\n", excep->GetExcept());
      throw excep;
   }
  
   // time of beginning
   timer->timeStart = time(0);
    
   // ---- write all used blocks
   showDebug(3, "TRACE: 007 (copyUsedBlocks)\n");  
   g_interface -> StatusLine(i18n("copying used data blocks"));

   showDebug(3, "begin of saveUsedBlocksToImageFile()\n");  
   try { saveUsedBlocksToImageFile(fs, options, timer); }
   catch ( CExceptions * excep)
   {
      showDebug(1, "saveToImage: excep %d\n", excep->GetExcept());
      throw excep;
   }
  
   RETURN_int(0);
}

// =======================================================
int /*CFSWork::*/restoreFromImage(CFSBase *fs, CImage *image, COptions *options, CRestoringWindow *gui, CPiTimer *timer)
{
  int nRes;

  BEGIN;
  
  // init
  fs->m_image = image;
  fs->m_guiRestore = gui;
  
  g_interface -> StatusLine(i18n("reading partition informations"));
  
  // 1. ---- Read header
  showDebug(3, "TRACE: 001 (readHeader)\n");  
  try { fs->m_image->readAndCheckMagic(MAGIC_BEGIN_LOCALHEADER); } 
  catch ( CExceptions * excep )
    {
      showDebug(1, "restoreFromImage: %d\n", excep->GetExcept()); 
      throw excep;
    }

  try {fs->m_image->read((char*)&fs->m_header, sizeof(CLocalHeader), true);}
  catch(CExceptions * excep)
    {
      showDebug(1, "restoreFromImage: %d\n", excep->GetExcept()); 
      throw excep;
    }

  // read CRC of info
  fs->m_image->readAndCheckCRC((char*)&fs->m_header, sizeof(CLocalHeader));

  
  // check the current format is supported
  /*if (strncmp(headReiser.szAppFormat, CURRENT_REISER_FORMAT, 
    strlen(CURRENT_REISER_FORMAT)) < 0)
    {
    g_interface->ErrorNewerRelease();
    RETURN_int(-1);
    }*/	
    
  // 2. ---- Print informations and ask confirmation to restore
  /*showDebug(1, "TRACE: 002 (printfInfos)\n");  
    if (options->bBatchMode == false)
    {
    printfInformations();
      
    // ask confirmation
    nRes = g_interface -> askRestore(fs->m_szDevice,
    image->get_szImageFilename());
    if (nRes == 2) // 2 --> cancel
    RETURN_int(-1);
    }*/
  
  // time of beginning
  timer->timeStart = time(0);
  
  // ---- 5. read bitmap
  showDebug(3, "TRACE: 003 (readBitmap)\n");  
  try { fs->m_image->readAndCheckMagic(MAGIC_BEGIN_BITMAP); } 
  catch ( CExceptions * excep )
    {
      showDebug(1, "restoreFromImage: %d\n", excep->GetExcept()); 
      throw excep;
    }
  
  // init bitmap size
  fs->m_bitmap.init(SGET(fs->m_header, qwBitmapSize));
  
  // copy bitmap blocks to bitmap in memory
  try{fs->m_image->read((char*)fs->m_bitmap.data(), (long)(SGET(fs->m_header, qwBitmapSize)), true);}
  catch (CExceptions * excep)
    {
      showDebug(1, "restoreFromImage: %d\n", excep->GetExcept()); 
      throw excep;
    }

  // write informations specific for the FS
  showDebug(3, "TRACE: 004 (readInfo)\n");  
  try {fs->m_image->readAndCheckMagic(MAGIC_BEGIN_INFO); }
  catch ( CExceptions * excep )
    {
      showDebug(1, "restoreFromImage: %d\n", excep->GetExcept()); 
      throw excep;
    }

  try { fs->m_image->read((char*)fs->getInfos(), INFOS_STRUCT_SIZE,true); }
  catch ( CExceptions * excep )
    {
      showDebug(1, "restoreFromImage: %d\n", excep->GetExcept()); 
      throw excep;
    }

  // read CRC of info
  fs->m_image->readAndCheckCRC((char*)fs->getInfos(), INFOS_STRUCT_SIZE);

  // 2. ---- Print informations and ask confirmation to restore
  showDebug(3, "TRACE: 002 (printfInfos)\n");  
  if (options->bBatchMode == false)
    {
      fs->printfInformations();
      
      // ask confirmation
      if (!options->bSimulateMode)
	{
	  nRes = g_interface -> askRestore(fs->m_szDevice, image->get_szImageFilename());
	  if (nRes == MSGBOX_NO) // 2 --> cancel
	    return -1; //THROW(ERR_CANCELED);
	}
    }

  // read and check magic string
  showDebug(3, "TRACE: 005 (copyUsedBlocks)\n");  
  try { fs->m_image->readAndCheckMagic(MAGIC_BEGIN_DATABLOCKS); }
  catch ( CExceptions * excep )
    {
      showDebug(1, "restoreFromImage: %d\n", excep->GetExcept()); 
      throw excep;
    }
  
  // ---- 7. Restore all used blocks
  errno = 0;
  g_interface -> StatusLine(i18n("copying used data blocks"));
  showDebug(1, "begin of restoreUsedBlocksToDisk()\n");  
  try { restoreUsedBlocksToDisk(fs, options, timer); }
  catch ( CExceptions * excep )
    {
      showDebug(1, "restoreFromImage: %d\n", excep->GetExcept()); 
      throw excep;
    }
  showDebug(9, "end of restoreUsedBlocksToDisk()\n");  
  
  RETURN_int(0);
}

