/***************************************************************************
                            net.cpp  -  description
                             -------------------
    begin                : Wed Sep  4 19:21:54 UTC 2002
    copyright            : (C) 2002 by Francois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <string.h>
#include <netdb.h>
#include <sys/time.h>
#include <errno.h>

#include "global.h"
#include "newnet.h"
#include "common.h"
#include "result.h"

#define BACKLOG_SIZE 5 

DWORD dwTotalSent = 0;

//------------------------------------------------
CNewNet::CNewNet(char *szCnxName)
{
   m_qwBytesIn = 0;
   m_qwBytesOut = 0;
   snprintf(m_szCnxName, sizeof(m_szCnxName), szCnxName);
   pi_printf ("NET[%s]: init\n", m_szCnxName);
}

//------------------------------------------------
CNewNet::~CNewNet()
{
   pi_printf ("NET[%s]: close\n", m_szCnxName);
   close(m_sockConnect);
   close(m_sock);
}

//------------------------------------------------
int CNewNet::messageToRead()
{
   int nRes;
   char c;

   nRes = ::recv(m_sockConnect, &c, 1, MSG_PEEK);

   return nRes;
}

//------------------------------------------------
int CNewNet::init(u16 port)
{
   int yes=1;

   m_myPid = getpid();
   m_sock = socket(AF_INET, SOCK_STREAM, 0);
   if (m_sock == -1)
   {
      perror("CNewNet::init socket");
      return -1;
   }

   setsockopt(m_sock, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof (int));

   struct sockaddr_in m_localSa;
   memset(&m_localSa, 0, sizeof (m_localSa));
   m_localSa.sin_family		= AF_INET;
   m_localSa.sin_addr.s_addr	= htonl(INADDR_ANY);
   m_localSa.sin_port		= htons(port);
   if (bind(m_sock, (struct sockaddr *) &m_localSa, sizeof(struct sockaddr_in)))
   {
      perror("CNewNet::init bind");
      return -1;
   }

   return (m_sock);
}
 
//------------------------------------------------
int CNewNet::recv(void *buf, u32 len)
{
   int nRes;

   nRes = ::recv(m_sockConnect, buf, (size_t)len, MSG_WAITALL); // == (ssize_t)len
   m_qwBytesIn += nRes;

   //pi_printf ("NET[%s]: recv(%d): m_qwBytesIn=%llu and m_qwBytesOut=%llu\n", m_szCnxName, nRes, m_qwBytesIn, m_qwBytesOut);
   return (nRes);
}

//------------------------------------------------
int CNewNet::send(void *buf, u32 len)
{
   int nRes;
   
   if (len > 0)
   {
      nRes = ::send(m_sockConnect, buf, (size_t)len, 0);
      m_qwBytesOut += nRes;
      //pi_printf ("NET[%s]: send(%d): m_qwBytesIn=%llu and m_qwBytesOut=%llu\n", m_szCnxName, nRes, m_qwBytesIn, m_qwBytesOut);
      return nRes;
   }   
   else
   {
      return 0;
   }
}

//------------------------------------------------
u8  CNewNet::recv_u8()
{
   u8 cRes;
   recv(&cRes, sizeof(cRes));
   return cRes;
}

//------------------------------------------------
u16 CNewNet::recv_u16()
{
   u16 wRes;
   recv(&wRes, sizeof(wRes));
   return wRes;
}

//------------------------------------------------
u32 CNewNet::recv_u32()
{
   u32 dwRes;
   int nRes;
   nRes = recv(&dwRes, sizeof(dwRes));
   errno = 0;
   if (nRes != sizeof(dwRes))
   {
      /*pi_printf ("NET[%s] nRes=%d and errno = %d=%s\n", m_szCnxName, nRes, errno, strerror(errno));
      fprintf(stderr, "NET[%s]", m_szCnxName); 
      perror("recv_u32 error");*/
      return 0xEEEEEEEE;
   }
   return dwRes;
}

//------------------------------------------------
u64 CNewNet::recv_u64()
{
   u64 qwRes;
   recv(&qwRes, sizeof(qwRes));
   return qwRes;
}

//------------------------------------------------
s32  CNewNet::send_u8(u8 a)
{
   int nRes;
   nRes = send(&a, sizeof(a));
   return (nRes);
}

//------------------------------------------------
s32  CNewNet::send_u16(u16 a)
{
   return (send(&a, sizeof(a)));
}

//------------------------------------------------
s32  CNewNet::send_u32(u32 a)
{
   return (send(&a, sizeof(a)));
}

//------------------------------------------------
s32  CNewNet::send_u64(u64 a)
{
   return (send(&a, sizeof(a)));
}

//------------------------------------------------
int CNewNet::recv_string(void *buf)
{
   u32 len;
   *((char*)buf) = 0; // end of string if empty

   len = recv_u32();
   return (recv(buf, len));
}

//------------------------------------------------
int CNewNet::recv_string(void *buf, u32 *len)
{
   *len = 0;
   *len = recv_u32();

   return (recv(buf, *len));
}

//------------------------------------------------
int CNewNet::send_string(void *buf)
{
   u32 len;

   if (!buf)
   {
      send_u32(0);
      return 0;
   }
   else
   {
      len = strlen((char*)buf) + 1;
      send_u32(len);
      return (send(buf, len));
   }
}

//------------------------------------------------
int CNewNet::send_string(void *buf, u32 len)
{
   send_u32(len);
   return (send(buf, len));
}

//------------------------------------------------
int CNewNet::connect(const char *szAddr, u16 nPort, char *szBanner, char *szVersion)
{
   sockaddr_in		addr;
   struct hostent         *h;
   char szTemp[1024];
   u32 dwRes;

   init(0/*nPort*/);

   // convert name -> IP
   errno = 0;
   h = gethostbyname(szAddr);
   if (!h)
   {
      errno = HOST_NOT_FOUND;
      return -1;
   }
  
   // connect
   addr.sin_family	= AF_INET;
   addr.sin_port	= htons(nPort);
   memcpy(&addr.sin_addr.s_addr, h->h_addr_list[0],h->h_length);

   errno = 0;
   if (::connect(m_sock, (struct sockaddr *)&addr, sizeof (addr)))
      perror("CNewNet::connect ::connect");

   m_sockConnect = m_sock;
   
   // check the banner
   snprintf(szTemp, sizeof(szTemp), "%s/%s", szBanner, szVersion);
   send_string(szTemp);
   dwRes = recv_u32();

   if (dwRes != 0)
      return -1;

   return 0;
}

//------------------------------------------------
int CNewNet::accept(int nPort, char *szBanner, char *szVersion)
{
   socklen_t		len;
   struct sockaddr	temp;
   char szTemp[1024];
   char szTemp2[1024];
   u32 dwLen;

   init(nPort);
   
   if (listen(m_sock, BACKLOG_SIZE))
   {
      perror("CNewNet::init listen");
      return -1;
   }

   len = sizeof (temp);
   
   m_sockConnect = ::accept(m_sock, &temp, &len);
   if (m_sockConnect == -1)
      perror("CNewNet::accept ::accept");

   // check the banner
   snprintf(szTemp, sizeof(szTemp), "%s/%s", szBanner, szVersion);
   dwLen = recv_string(szTemp);
   
   if (strncmp(szTemp, szTemp2, sizeof(szTemp)) != 0)
      send_u32(1); // error
   else
      send_u32(0); // ok
      
   return m_sockConnect;
}

//------------------------------------------------
int CNewNet::send_result(CResult *res)
{
   //pi_printf ("-- BEGIN CNewNet::send_result\n");
   u32 dwRes;
   dwRes = res->getResult();

   send_u32(dwRes);
   pi_printf ("   error=%lu=%.8x\n", dwRes, dwRes);

   if (dwRes == RESULT_ERROR) // error
   {
      pi_printf("send_result: error [%s]\n", res->getError() );
      send_string(res->getError());
   }
   else
      pi_printf("send_result: succss\n");
      

   //pi_printf ("-- END CNewNet::send_result\n");
   return 0;
}

//------------------------------------------------
int CNewNet::recv_result(CResult *res)
{
   u32 dwRes;
   char szTemp[2048];

   dwRes = recv_u32();
   res->setResult(dwRes);

   if (dwRes == RESULT_ERROR)
   {
      recv_string(szTemp);
      res->setError(szTemp);
   }
   
   return 0;
}

//------------------------------------------------
s8  CNewNet::recv_s8()
{
   s8 cRes;
   recv(&cRes, sizeof(cRes));
   return cRes;
}

//------------------------------------------------
s16 CNewNet::recv_s16()
{
   s16 wRes;
   recv(&wRes, sizeof(wRes));
   return wRes;
}

//------------------------------------------------
s32 CNewNet::recv_s32()
{
   s32 dwRes;
   int nRes;
   nRes = recv(&dwRes, sizeof(dwRes));
   errno = 0;
   if (nRes != sizeof(dwRes))
   {
      /*pi_printf ("NET[%s] nRes=%d and errno = %d=%s\n", m_szCnxName, nRes, errno, strerror(errno));
      fprintf(stderr, "NET[%s]", m_szCnxName); 
      perror("recv_s32 error");*/
      return 0xEEEEEEEE;
   }
   return dwRes;
}

//------------------------------------------------
s64 CNewNet::recv_s64()
{
   s64 qwRes;
   recv(&qwRes, sizeof(qwRes));
   return qwRes;
}

//------------------------------------------------
s32 CNewNet::send_s8(s8 a)
{
   int nRes;
   nRes = send(&a, sizeof(a));
   return (nRes);
}

//------------------------------------------------
s32 CNewNet::send_s16(s16 a)
{
   return (send(&a, sizeof(a)));
}

//------------------------------------------------
s32 CNewNet::send_s32(s32 a)
{
   return (send(&a, sizeof(a)));
}

//------------------------------------------------
s32 CNewNet::send_s64(s64 a)
{
   return (send(&a, sizeof(a)));
}

