/***************************************************************************
                         operations.cpp  -  description
                             -------------------
    begin                : Wed Sep  4 19:21:54 UTC 2002
    copyright            : (C) 2002 by Francois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <parted/parted.h>
#include <parted/disk.h>

#include "operations.h"
#include "newnet.h"
#include "workserver.h"
#include "common.h"
#include "result.h"

#define DEV_CANCHANGE // allows to make changes

//------------------------------------------------
void operationPiSave(CNewNet *net)
{
   COptions options;
   char szDevice[4096];
   char szImageFile[4096];

   net->recv(&options, sizeof(COptions));
   net->recv_string(szDevice);
   net->recv_string(szImageFile);

   pi_printf ("--> SAVEPART: device=[%s], image=[%s]\n", szDevice, szImageFile);

   return ;
}

//------------------------------------------------
void operationPiRestore(CNewNet *net)
{
   return ;
}

//------------------------------------------------
void operationPiCopy(CNewNet *net)
{
   return ;
}

//------------------------------------------------
void operationPartedNewLabel(CNewNet *net)
{
   CResult res;

   char szDevice[1024];

   net->recv_string(szDevice);

   pi_printf ("--> MKLABEL(dev=%s)\n", szDevice);

#ifdef DEV_CANCHANGE
   res = partedPartNewLabel(szDevice); 
#endif // DEV_CANCHANGE
   
   net->send_result(&res);
}

//------------------------------------------------
int locateProgram(char *szResultPath, int nPathLen, char *szBin)
{
   DWORD i;
   char *szLocations[] = {"/sbin", "/usr/sbin", "/usr/local/sbin", "/bin", "/usr/bin", "/usr/local/bin", ""};
   char szTemp[2048];

   for (i=0; *(szLocations[i]); i++)
   {
      snprintf(szTemp, sizeof(szTemp), "%s/%s", szLocations[i], szBin);
      //pi_printf ("--> pgm=%s/%s\n", szLocations[i], szBin);
      if (doesFileExists2(szTemp))
      {
	 snprintf(szResultPath, nPathLen, "%s/%s", szLocations[i], szBin);
	 return 0; // okay
      }
   }

   return -1; // error
}

//------------------------------------------------
int execProgram(char *szProgram, char *argv[], char *szOutput, int nMaxOutput)
{
   int nFd[2];
   int nRes;
   pid_t pid;

   // init
   memset(szOutput, 0, nMaxOutput);
   if (pipe(nFd) < 0)
      return -1;

   pid = fork();
   switch (pid)
   {
      case -1: // error
	 perror ("fork() in execProgram");
	 return -1;

      case 0: // child
 	 dup2(nFd[1], STDOUT_FILENO); // stdout --> parent process
 	 dup2(nFd[1], STDERR_FILENO); // stdout --> parent process
	 nRes = execve(szProgram, argv, NULL);
	 exit(0);
	 break;

      default: // parent
	 read(nFd[0], szOutput, nMaxOutput);
	 //pi_printf ("\n\n\n[[[[[[[%s]]]]]]]\n\n\n", szString);
	 waitpid(pid, NULL, 0);
	 sleep(1);
	 break;
   };

   return 0;
}

//------------------------------------------------
CResult partitionFormat(char *szDevice, char *szFs, char *szLabel)
{
   char szFullPath[4096];
   char szFullLabel[MAX_LABEL];
   char szOutput[8192];
   char *argv[5];
   DWORD i;
   int nFs;
   int nRes;

   nFs = -1;
   for (i=0; g_fs[i].dwNumber != FS_UNKNOWN; i++)
   {
      if ((strcmp(g_fs[i].szDesc, szFs)) == 0)
	 nFs = (int)i;
   }

   if (nFs == -1) // fs not found
      return CResult("File System unknown");
   
   nRes = locateProgram(szFullPath, sizeof(szFullPath), g_fs[nFs].szCmdMkfsName);
   if (nRes == -1)
      return CResult("Can't find program %s which is required <br>to format the partition.<br>"
		     "Please, install %s. <br>You can search on http://www.freshmeat.net.", 
		     g_fs[nFs].szCmdMkfsName, g_fs[nFs].szFsTools);

   if (szLabel && *szLabel) // not empty
      snprintf (szFullLabel, sizeof(szFullLabel), "%s", szLabel);
   else
      snprintf (szFullLabel, sizeof(szFullLabel), "none");

   argv[0] = g_fs[nFs].szCmdMkfsOpts;
   argv[1] = g_fs[nFs].szCmdMkfsLabelOpts;
   argv[2] = szFullLabel;
   argv[3] = g_fs[nFs].szCmdMkfsOpts;
   argv[4] = szDevice;
   argv[5] = NULL;

   // pi_printf debug
   pi_printf ("FORMATTING: path=[%s]\n", szFullPath);
   for (int j=0; j < 4; j++)
      pi_printf ("argv[%d]=%s\n", j, argv[j]);

   execProgram(szFullPath, argv, szOutput, sizeof(szOutput));

   return CResult();
}

//------------------------------------------------
void operationFormat(CNewNet *net)
{
   char szDevice[1024];
   char szFs[512];
   char szLabel[MAX_LABEL+1];
   CResult res;
   
   pi_printf("server: operationFormat(): will recv args\n");
   net->recv_string(szDevice);
   net->recv_string(szFs);
   net->recv_string(szLabel);
   
#ifdef DEV_CANCHANGE
   res = partitionFormat(szDevice, szFs, szLabel); 
#endif // DEV_CANCHANGE

   pi_printf("server: operationFormat(): will send result\n");
   net->send_result(&res);
   pi_printf("server: operationFormat(): had sent result\n");
}

//------------------------------------------------
CResult partedPartCreate(char *szDevice, DWORD dwType, DWORD dwStartSect, DWORD dwEndSect, char *szNewName, int nMaxNewNameLen)
{
   PedPartitionType part_type;
   PedDisk *disk;
   PedPartition *part;
   PedDevice* dev;
   PedSector sector_start;
   PedSector sector_end;
   PedFileSystem*			fs;
   PedConstraint*			constraint;
   const PedFileSystemType* def_type = ped_file_system_type_get ("ext2");

   //pi_printf ("partedPartCreate(): enter (szDevice=[%s], dwType=%lu)\n", szDevice, dwType);

   // init
   sector_start = (PedSector)dwStartSect;
   sector_end = (PedSector)dwEndSect;

   // type
   switch (dwType)
   {
      case PARTTYPE_PRIMARY: 
	 part_type = (PedPartitionType)0; // PED_PARTITION_PRIMARY
	 break;

      case PARTTYPE_LOGICAL: 
	 part_type = PED_PARTITION_LOGICAL;
	 break;

      case PARTTYPE_EXTENDED:
	 part_type = PED_PARTITION_EXTENDED;
	 break;	 

      default: // primary
	 part_type = (PedPartitionType)0; // PED_PARTITION_PRIMARY
	 break;
   }

   dev = ped_device_get (szDevice);
   if (!dev)
   {
      pi_printf ("Error 001\n");
      return CResult("partedPartCreate: error 001");
   }

   disk = ped_disk_new (dev);
   if (!disk)
      goto error;

   /*if (!command_line_get_new_part_type (_("Partition type?"), disk,
     &part_type))
     goto error_destroy_disk;

     if (part_type == PED_PARTITION_EXTENDED) {
     ped_exception_throw (PED_EXCEPTION_ERROR, PED_EXCEPTION_CANCEL,
     _("Extended partitions can't have file systems.  "
     "Did you want mkpart?"));
     goto error_destroy_disk;
     }*/

   /*fs_type = command_line_get_fs_type (_("File system type?"), def_type);
     if (!fs_type)
     goto error_destroy_disk;*/

   constraint = ped_file_system_get_create_constraint (def_type, dev);

   /*if (!command_line_get_sector (_("Start?"), dev, 0, &start))
     goto error_destroy_constraint;
     if (!command_line_get_sector (_("End?"), dev, 0, &end))
     goto error_destroy_constraint;*/
   part = ped_partition_new (disk, part_type, def_type, sector_start, sector_end);
   if (!part)
      goto error_destroy_constraint;
   /*if (!_grow_over_small_freespace (&part->geom, disk))
     goto error_destroy_part;*/
   if (!ped_disk_add_partition (disk, part, constraint))
      goto error_destroy_part;

   /*if (!_solution_check_distant (start, end,
     part->geom.start, part->geom.end,
     _("You requested to create a partition at %.3f-%.3fMb. The "
     "closest Parted can manage is %.3f-%.3fMb.")))
     goto error_remove_part;*/

   pi_printf ("You requested to create a partition at %.3f-%.3fMb. The closest Parted can manage is %.3f-%.3fMb.\n", 
	   (float)sector_start, (float)sector_end, (float)part->geom.start, (float)part->geom.end);

   /*PedTimer timer = ped_timer_new (_timer_handler, &timer_context);
     if (!timer)
     goto error_done_commands;*/

   if (part_type != PED_PARTITION_EXTENDED)
   {
      fs = ped_file_system_create (&part->geom, def_type, NULL /*timer*/);
      if (!fs) 
	 goto error_remove_part;
      ped_file_system_close (fs);
   }

   ped_constraint_destroy (constraint);

   if (!ped_partition_set_system (part, def_type))
      goto error_destroy_disk;

   // get new name
   snprintf(szNewName, nMaxNewNameLen, "%s%d", szDevice, part->num);

   if (!ped_disk_commit (disk))
      goto error_destroy_disk;
   ped_disk_destroy (disk);

   return CResult(); // success

  error_remove_part:
   ped_disk_remove_partition (disk, part);
  error_destroy_part:
   ped_partition_destroy (part);
  error_destroy_constraint:
   ped_constraint_destroy (constraint);
  error_destroy_disk:
   ped_disk_destroy (disk);
  error:
   return CResult("partedPartCreate(): error"); // error
}

//------------------------------------------------
void operationPartedCreate(CNewNet *net)
{
   char szDevice[1024];
   DWORD dwType;
   DWORD dwSectorStart;
   DWORD dwSectorEnd;
   char szFs[512];
   char szLabel[MAX_LABEL];
   char szPart[2048];
   CResult res;

   net->recv_string(szDevice);

   dwType = net->recv_u32();
   dwSectorStart = net->recv_u32();
   dwSectorEnd = net->recv_u32();
   net->recv_string(szFs);
   net->recv_string(szLabel);

   //pi_printf ("--> MKPART(dev=%s, start=%lu, end=%lu, type=%lu, fs=%s, label=%s)\n", szDevice, dwSectorStart, dwSectorEnd, dwType, szFs, szLabel);

#ifdef DEV_CANCHANGE
   res = partedPartCreate(szDevice, dwType, dwSectorStart, dwSectorEnd, szPart, sizeof(szPart)); 
   pi_printf("operationPartedCreate:...partedPartCreate(new=%s) Error %d\n", szPart, res.getResult());
   
   if ((res.getResult() == RESULT_SUCCESS) && (dwType != PARTTYPE_EXTENDED))
   {
      res = partitionFormat(szPart, szFs, szLabel); 
      pi_printf("operationPartedCreate:...partitionFormat() Error %d\n", res.getResult());
   }
#endif // DEV_CANCHANGE

   net->send_result(&res);
}

//------------------------------------------------
void operationPartedDelete(CNewNet *net)
{
   char szDevice[1024];
   DWORD dwMinor;
   CResult res;

   net->recv_string(szDevice);
   dwMinor = net->recv_u32();

#ifdef DEV_CANCHANGE
   res = partedPartDelete(szDevice, dwMinor, true);
#endif // DEV_CANCHANGE

   net->send_result(&res);
}

//------------------------------------------------
CResult partedPartDelete(char *szDevice, DWORD dwMinor, BOOL bProcess)
{
   PedDisk *disk;
   PedPartition *part;
   PedDevice* dev;
   
   dev = ped_device_get (szDevice);
   if (!dev)
   {
      pi_printf ("Error 001\n");
      return CResult("partedPartDelete: error 001");
   }

   disk = ped_disk_new (dev);
   if (!disk)
   {
      pi_printf ("Error 002\n");
      return CResult("partedPartDelete: error 002");
   }

   part = ped_disk_get_partition (disk, dwMinor);
   if (!part)
   {
      pi_printf ("Error 003\n");
      return CResult("partedPartDelete: error 003");
   }

   /*if (!_partition_warn_busy (part))
   {
      pi_printf ("warning: partition busy\n");
   }*/

   if (bProcess)
   {
      ped_disk_delete_partition (disk, part);
      ped_disk_commit (disk);
      ped_disk_destroy (disk);
   }
   
   return CResult();
}

//------------------------------------------------
CResult partedPartNewLabel(char *szDevice)
{
   PedDisk*		disk;
   //const PedDiskType*	type;
   PedDevice* dev;

   dev = ped_device_get (szDevice);
   if (!dev)
   {
      pi_printf ("Error 001\n");
      return CResult("partedPartNewLabel: error 001");
   }

   const PedDiskType*	def_type = ped_disk_probe (dev);

   //ped_exception_fetch_all ();
   /*disk = ped_disk_new (dev);
   if (!disk) ped_exception_catch ();*/
   //ped_exception_leave_all ();

   /*if (disk) 
   {
      if (!_disk_warn_busy (disk)) 
      {
	 ped_disk_destroy (disk);
	 goto error;
      }
      ped_disk_destroy (disk);
   }*/

   /*type = def_type; //command_line_get_disk_type (_("New disk label type?"), def_type);
      if (!type)
	 goto error;*/

   disk = ped_disk_new_fresh (dev, def_type);
   if (!disk)
      goto error;

   if (!ped_disk_commit (disk))
      goto error_destroy_disk;
   ped_disk_destroy (disk);
   return CResult(); // ok

  error_destroy_disk:
   ped_disk_destroy (disk);
  error:
   return CResult("partedPartNewLabel: error");
}

//------------------------------------------------
/*
static int do_mkpart (PedDevice** dev)
{
	PedDisk*			disk;
	PedPartition*			part;
	PedPartitionType		part_type;
	const PedFileSystemType*	fs_type;
	const PedFileSystemType* def_type = ped_file_system_type_get ("ext2");
	PedSector			start, end;
	PedConstraint*			constraint;
	char*				peek_word;

	disk = ped_disk_new (*dev);
	if (!disk)
		goto error;
	constraint = ped_constraint_any (*dev);
	if (!constraint)
		goto error_destroy_disk;

	if (!command_line_get_new_part_type (_("Partition type?"), disk,
					     &part_type))
		goto error_destroy_disk;

	peek_word = command_line_peek_word ();
	if (part_type == PED_PARTITION_EXTENDED
	    || (peek_word && isdigit (peek_word[0]))) {
		fs_type = NULL;
	} else {
		fs_type = command_line_get_fs_type (_("File system type?"),
						    def_type);
		if (!fs_type)
			goto error_destroy_constraint;
	}
	if (peek_word)
		ped_free (peek_word);

	if (!command_line_get_sector (_("Start?"), *dev, 0, &start))
		goto error_destroy_constraint;
	if (!command_line_get_sector (_("End?"), *dev, 0, &end))
		goto error_destroy_constraint;
	part = ped_partition_new (disk, part_type, fs_type, start, end);
	if (!part)
		goto error_destroy_constraint;
	if (!_grow_over_small_freespace (&part->geom, disk))
		goto error_destroy_part;
	if (!ped_disk_add_partition (disk, part, constraint))
		goto error_destroy_part;
	ped_constraint_destroy (constraint);

	if (!_solution_check_distant (start, end,
				      part->geom.start, part->geom.end,
		_("You requested to create a partition at %.3f-%.3fMb. The "
		  "closest Parted can manage is %.3f-%.3fMb.")))
		goto error_remove_part;

	if (!ped_partition_set_system (part, fs_type))
		goto error_destroy_disk;
	if (!ped_disk_commit (disk))
		goto error_destroy_disk;
	ped_disk_destroy (disk);
	return 1;

error_remove_part:
	ped_disk_remove_partition (disk, part);
error_destroy_part:
	ped_partition_destroy (part);
error_destroy_constraint:
	ped_constraint_destroy (constraint);
error_destroy_disk:
	ped_disk_destroy (disk);
error:
	return 0;
}
*/
