/***************************************************************************
                         workserver.cpp  -  description
                             -------------------
    begin                : Wed Sep  4 19:21:54 UTC 2002
    copyright            : (C) 2002 by Francois Dupoux
    email                : fdupoux@partimage.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <parted/parted.h>
#include <parted/disk.h>
#include <sys/mount.h>

#include "operations.h"
#include "workserver.h"
#include "common.h"
#include "newnet.h"

#include "fs/fs_reiser.h"
#include "fs/fs_ufs.h"
#include "fs/fs_ext2.h"
#include "fs/fs_hpfs.h"
#include "fs/fs_ntfs.h"
#include "fs/fs_fat.h"
#include "fs/fs_jfs.h"
#include "fs/fs_xfs.h"
#include "fs/fs_hfs.h"
#include "fs/fs_unknown.h"

//------------------------------------------------
void serverapp(int nPort)
{
   CNewNet *net;
   u32 dwCmd = 0;

   g_pidType = 1;

   pi_printf ("+++ server: Waiting connexion\n");

   net = new CNewNet("ServerCnx");
   net->accept(nPort, PIGUI_BANNER_TEXT, PIGUI_BANNER_VERSION);

   pi_printf ("+++ server: connected\n");

   do
   {
      pi_printf ("----------- WAIT FOR NEW Command\n");
      dwCmd = net->recv_u32();

      switch(dwCmd)
      {
 	 case PICMD_GET_HDLIST: // [u8 bShowDetails]
	    pi_printf ("+++++++++++ PICMD_GET_HDLIST\n");
	    getHdList(net);
	    break;
	    	    
 	 case PICMD_GET_FSLIST: // []
	    pi_printf ("+++++++++++ PICMD_GET_FSLIST\n");
	    getFsList(net);
	    break;

	 case PICMD_OPERATION_GET_PARTDETAILS: // [string part]
	    getPartDetails(net);
	    break;

 	 case PICMD_GET_ISMOUNTED: // [string part]
	    pi_printf ("+++++++++++ PICMD_GET_ISMOUNTED\n");
	    isMounted(net);
	    break;

	 case PICMD_UNMOUNTED: // [string part]
	    pi_printf ("+++++++++++ PICMD_UNMOUNTED\n");
	    unmountPart(net);
	    break;	    

	 case PICMD_OPERATION_FORMAT: // [string part, string fs, string label]
	    pi_printf ("+++++++++++ PICMD_OPERATION_FORMAT\n");
	    operationFormat(net);
	    //net->send_u32(1);
	    break;
	    	    
	 case PICMD_GET_HDLIST_FORONEDISK: // [u8 bShowDetails, string device]
	    pi_printf ("++++++++++ PICMD_GET_HDLIST_FORONEDISK\n");
	    getHdListForOneDisk(net);
	    break;
	    
	 case PICMD_OPERATION_PI_SAVE: // [options, string device, string file]
	    operationPiSave(net);
	    //net->send_u32(1);
	    break;

	 case PICMD_OPERATION_PI_RESTORE: // [string device, string file]
	    operationPiRestore(net);
	    //net->send_u32(1);
	    break;

	 case PICMD_OPERATION_PI_COPY:
	    operationPiCopy(net);
	    //net->send_u32(1);
	    break;

	 case PICMD_OPERATION_PARTED_NEWLABEL: // [string device]
	    pi_printf ("+++++++++++ PICMD_OPERATION_PARTED_NEWLABEL\n");
	    operationPartedNewLabel(net);
	    //net->send_u32(1);
	    break;

	 case PICMD_OPERATION_PARTED_CREATE: // [string device, u32 part_type, u32 dwSectorStart, u32 dwSectorEnd, string fs, string label]
	    pi_printf ("+++++++++++ PICMD_OPERATION_PARTED_CREATE\n");
	    operationPartedCreate(net);
	    //net->send_u32(1);
	    break;

	 case PICMD_OPERATION_PARTED_DELETE: // [string device, u32 minor]
	    pi_printf ("+++++++++++ PICMD_OPERATION_PARTED_DELETE\n");
	    operationPartedDelete(net);
	    //net->send_u32(1);
	    break;

	 case PICMD_EXIT:
	    net -> send_u32(PICMD_EXIT);
	    break;
      }
   } while (dwCmd != PICMD_EXIT);

   pi_printf ("+++ server: exit\n");
}

//------------------------------------------------
CFileSys g_fs[] = { {"reiserfs",    true,  FS_REISER_3_6,  "ReiserFs",   0x00, 0x00, 0xFF, "reiserfsprogs", "mkreiserfs",   "-f",   "-l"},
                    {"xfs",         true,  FS_XFS,         "XFS (SGI)",  0x00, 0xFF, 0xFF, "xfsprogs",      "mkfs.xfs",     "-f",   "-L"},
                    {"jfs",         true,  FS_JFS,         "JFS (IBM)",  0xFF, 0xFF, 0x00, "jfsutils",      "mkfs.jfs",     "-q",   "-L"},
                    {"ext2",        false, FS_EXT2,        "Ext2fs",     0xCC, 0x00, 0x00, "e2fsprogs",     "mkfs.ext2",    "",     "-L"},
                    {"ext3",        true,  FS_EXT3,        "Ext3fs",     0xFF, 0x00, 0x00, "e2fsprogs",     "mkfs.ext2",    "-j",   "-L"},
                    {"fat16",       false, FS_FAT16,       "FAT16",      0x00, 0xFF, 0x00, "dosfstools",    "mkdosfs",      "-F16", "-n"},
                    {"fat32",       false, FS_FAT32,       "FAT32",      0x00, 0xCC, 0x00, "dosfstools",    "mkdosfs",      "-F32", "-n"},
                    {"ntfs",        true,  FS_NTFS,        "NTFS",       0xFF, 0x00, 0xFF, "linux-ntfs",    "mkntfs",       "",     "-L"},
                    {"ffs",         true,  FS_FFS,         "BSD FFS",    0xFF, 0x88, 0x88, "",              "",             "",     ""},
		    {"unknown",     false, FS_UNKNOWN,     "Unknown FS", 0xFF, 0xFF, 0xFF, "",              "",             "",     ""}
};

//------------------------------------------------
void unmountPart(CNewNet *net)
{
   char szPart[2048];
   char szMountPoint[4096];
   u8 bMounted;
   int nRes;

   memset(szMountPoint, 0, sizeof(szMountPoint));
   net->recv_string(szPart);

   bMounted = !!isMounted(szPart, szMountPoint);

   if (!bMounted)
   {
      net->send_u8(0); // okay
      return;
   }

   // umount partition
   errno = 0;
   nRes = umount(szMountPoint);
   
   pi_printf("umount(%s)=%d. Error=%s\n", szMountPoint, nRes, strerror(errno));

   net->send_u8(!!nRes); // send result
}

//------------------------------------------------
void isMounted(CNewNet *net)
{
   char szPart[2048];
   char szMountPoint[4096];
   u8 bMounted;

   memset(szMountPoint, 0, sizeof(szMountPoint));
   net->recv_string(szPart);

   bMounted = !!isMounted(szPart, szMountPoint);

   net->send_u8(bMounted); // yes or no
   if (bMounted)
      net->send_string(szMountPoint);
}

//------------------------------------------------
void getFsList(CNewNet *net)
{
   DWORD i;

   for (i=0; g_fs[i].dwNumber != FS_UNKNOWN; i++)
   {
      net->send_u32(1); // not end
      net->send(&g_fs[i], sizeof(CFileSys));
   }

   net->send_u32(0); // end
}

//------------------------------------------------
int fillPartInfo(CPartInfo *pi, PedPartition *part, char *szDevice, QWORD qwDiskSize, BOOL bShowDetails)
{
   char szMountPoint[4096];
   DWORD i;
   int nRes;

   // get data
   snprintf(pi->szMagic, sizeof(pi->szMagic), "%s", MAGIC_INFOGUI);
   pi->dwMinor = (DWORD)part->num;
   pi->qwSize = (QWORD)512*(QWORD)part->geom.length;
   snprintf (pi->szPartName, sizeof(pi->szPartName), "%s%d", szDevice, part->num);
   pi->dwType = part->type;
   pi->qwGeomStart = (QWORD) part->geom.start;
   pi->qwGeomEnd  = (QWORD) part->geom.end;

   PedPartitionFlag flag;
   pi->dwFlags=0;
   for (flag=ped_partition_flag_next((PedPartitionFlag)0); flag; flag=ped_partition_flag_next(flag))
   {
      if (ped_partition_get_flag(part, flag))
	 pi->dwFlags |= (1 << flag);
   }

   pi->dwP1000Disk = (DWORD)((1000LL * pi->qwSize) / qwDiskSize);
   if (isMounted(pi->szPartName, szMountPoint))
      snprintf(pi->szMountPoint, sizeof(pi->szMountPoint), "%s", szMountPoint);

   // fill data
   if (part -> fs_type)
      sprintf (pi->szFilesysName, "%s", part->fs_type->name);

   // unknown FS
   if (!part->fs_type)
   {
      sprintf (pi->szFilesysName, "unknown");
      pi->dwFileSystemType = FS_UNKNOWN;
      pi->cColorRed   = 0x00;
      pi->cColorGreen = 0x00;
      pi->cColorBlue  = 0x00;
   }

   for (i=0; *g_fs[i].szName; i++)
   {
      if ((part->fs_type) && (strcmp(part->fs_type->name, g_fs[i].szName) == 0))
      {
	 pi->dwFileSystemType = g_fs[i].dwNumber;
	 pi->cColorRed = g_fs[i].cColorRed;
	 pi->cColorGreen = g_fs[i].cColorGreen;
	 pi->cColorBlue = g_fs[i].cColorBlue;
      }
   }

   if (bShowDetails)
   {
      nRes = getDeviceInfos(&pi->headInfo, pi->szPartName, pi->szFilesysName, pi->qwSize);
      //pi_printf ("getDeviceInfos(%s) = %d\n", pi->szPartName, nRes);

      if (nRes == 0 && (SGET(pi->headInfo, qwBlocksCount) != 0))
	 pi->dwPctUsage = (DWORD)((SGET(pi->headInfo, qwUsedBlocks) * 100LL) / SGET(pi->headInfo, qwBlocksCount));
      else
	 pi->dwPctUsage = 0;
   }
   else
   {
      pi->dwPctUsage = 100;
      SPUT(pi->headInfo, qwBlockSize, 512);
      SPUT(pi->headInfo, qwBlocksCount, (pi->qwSize / SGET(pi->headInfo, qwBlockSize)));
      SPUT(pi->headInfo, qwUsedBlocks, (pi->qwSize / SGET(pi->headInfo, qwBlockSize)));
   }

   return 0;
}

//------------------------------------------------
BOOL isSpaceInsideExtended(CDiskInfo di, DWORD dwSpaceSectorStart, DWORD dwSpaceSectorEnd)
{
   DWORD dwExtSectorStart;
   DWORD dwExtSectorEnd;
   DWORD dwMiddle;
   
   dwExtSectorStart = di.piExtended.qwGeomStart;
   dwExtSectorEnd = di.piExtended.qwGeomEnd;

   /*
     pi_printf ("\n---------------------\n");
     pi_printf ("*******> dwSpaceSectorStart=%lu\n", dwSpaceSectorStart);
     pi_printf ("*******> dwSpaceSectorEnd=%lu\n",   dwSpaceSectorEnd);
     pi_printf ("*******> dwExtSectorStart=%lu\n", dwExtSectorStart);
     pi_printf ("*******> dwExtSectorEnd=%lu\n", dwExtSectorEnd);
     pi_printf ("*******> (dwSpaceSectorStart >= dwExtSectorStart)=%d\n", (dwSpaceSectorStart >= dwExtSectorStart));
     pi_printf ("*******> (dwSpaceSectorEnd <= dwExtSectorEnd)=%d\n", (dwSpaceSectorEnd <= dwExtSectorEnd));
     pi_printf ("---------------------\n\n");
   */

   if (di.bExtendedExists == false)
   {
      pi_printf ("di.bExtendedExists == false\n");      
      return false;
   }

   dwMiddle = dwSpaceSectorStart + ((dwSpaceSectorEnd - dwSpaceSectorStart) / 2);

   return ((dwMiddle >= dwExtSectorStart) && (dwMiddle <= dwExtSectorEnd));

   //return ((dwSpaceSectorStart >= dwExtSectorStart) && (dwSpaceSectorEnd <= dwExtSectorEnd));
}


//------------------------------------------------
int getHdListDoDisk(CNewNet *net, void *vdevice, BOOL bShowDetails)
{
   PedDisk *disk;
   PedPartition *part;
   CPartInfo *pinfo;
   CDiskInfo dinfo;
   char szTemp[1024];
   DWORD dwInfoCount;
   PedDevice *device;
   int nRes;

   device = (PedDevice *)vdevice;

   disk = ped_disk_new(device);
   if (!disk)
   {
      pi_printf("Error in ped_disk_new\n");
      return -1;
   }

   // device
   memset(&dinfo, 0, sizeof(dinfo));
   dinfo.qwDiskSize = (disk -> dev -> length)*512;
   nRes = convertDevfsNameToShortName(device->path, dinfo.szDevice, sizeof(dinfo.szDevice));
   snprintf(dinfo.szDiskName, sizeof(dinfo.szDiskName), "%s (%s)", dinfo.szDevice, formatSize(dinfo.qwDiskSize, szTemp));
   //printf ("convert (%s)->(%s)\n", device->path, dinfo.szDevice);

   //snprintf(dinfo.szDevice, sizeof(dinfo.szDevice), "%s", device->path); 
   //snprintf(dinfo.szDiskName, sizeof(dinfo.szDiskName), "%s (%s)", device->path, formatSize(dinfo.qwDiskSize, szTemp));

   // partition list
   part = NULL;
   dwInfoCount = 0;
      
   // alloc memory
   pinfo = new CPartInfo[20];
   memset(pinfo, 0, sizeof(CPartInfo)*20);
   dinfo.bExtendedExists = false;
   dinfo.dwPrimaryCount = 0;

   // ---- 1. get informations : fill pinfo[MAX_PARTS+2];
   while ((part = ped_disk_next_partition(disk, part)) != 0)
   {
      if (part->num != -1)
      {
	 if (((int)part->type == PED_PARTITION_EXTENDED) || ((int)part->type == 0 /* primary */))
	 {
	    dinfo.dwPrimaryCount++;
	 }

	 if (part->type != PED_PARTITION_EXTENDED)
	 {
	    nRes = convertDevfsNameToShortName(device->path, szTemp, sizeof(szTemp));
	    fillPartInfo(&pinfo[dwInfoCount], part, szTemp /*device->path*/, dinfo.qwDiskSize, bShowDetails);
	    dwInfoCount++;
	 }
	 else // PED_PARTITION_EXTENDED
	 {
	    //pi_printf ("EXTENDED: start=%lu, end=%lu\n", (DWORD)part->geom.start, (DWORD)part->geom.end);
	    dinfo.bExtendedExists = true;
	    dinfo.piExtended.dwType = (DWORD)part->type;
	    dinfo.piExtended.qwGeomStart = (QWORD)part->geom.start;
	    dinfo.piExtended.qwGeomEnd = (QWORD)part->geom.end;
	 }
      }
   }


   // ---- 2. sort data : order fill pinfo[MAX_PARTS+2];
   sortTable(pinfo, dwInfoCount);

   // ---- 3. add holes at the end of the table
   dwInfoCount = addFreeSpace(dinfo, pinfo, dwInfoCount, (disk -> dev -> length));

   // ---- 4. reorder the table
   sortTable(pinfo, dwInfoCount);

   // send data
   net->send(&dinfo, sizeof(dinfo));
   net->send_u32(dwInfoCount);

   net->send(pinfo, (sizeof(CPartInfo)*dwInfoCount));
   delete pinfo;

   return 0;
}

//------------------------------------------------
void getHdList(CNewNet *net)
{
   PedDevice *device;

   //CPartInfo pinfo[MAX_PARTS+2];
   u8 bShowDetails;

   // get options
   bShowDetails = net->recv_u8();

   device = NULL;
   ped_device_probe_all();
   
   while ((device = ped_device_get_next(device)) != NULL)
   {
      net->send_u32(1); // not end
      getHdListDoDisk(net, device, bShowDetails);
   };

   net->send_u32(0); // end of devices
}

//------------------------------------------------
void getHdListForOneDisk(CNewNet *net)
{
   char szDevice[1024];
   PedDevice *device;
   u8 bShowDetails;

   // get options
   bShowDetails = net->recv_u8();

   net->recv_string(szDevice);

   device = ped_device_get (szDevice);
   if (!device)
   {
      pi_printf ("Error 001\n");
      return ;
   }

   getHdListDoDisk(net, device, bShowDetails);
}

//------------------------------------------------
void sortTable(CPartInfo *pinfo, DWORD dwInfoCount)
{
   DWORD i,j;
   QWORD qwBest;
   DWORD dwBestPos;
   CPartInfo *tab2;

   // alloc mem
   tab2 = new CPartInfo[dwInfoCount+1];

   // order
   for (i=0; i < dwInfoCount; i++)
   {
      // search the first remaining partition in the list
      qwBest = (QWORD)-1;
      dwBestPos = 0;
      for (j=0; j < dwInfoCount; j++)
      {
	 if ((pinfo[j].qwGeomStart < qwBest) && (pinfo[j].qwGeomStart))
	 {
	    qwBest = pinfo[j].qwGeomStart;
	    dwBestPos = j;
	 }
      }
      
      // copy best and erase
      //pi_printf ("SORT: best=%lu\n", dwBestPos);
      memcpy(&tab2[i], &pinfo[dwBestPos], sizeof(CPartInfo));
      memset(&pinfo[dwBestPos], 0, sizeof(CPartInfo));
   }

   // end: copy and count
   memcpy(pinfo, tab2, dwInfoCount*sizeof(CPartInfo));
   for (i=0; i < dwInfoCount; i++)
      pinfo[i].dwPosition = i;

   //free mem
   delete []tab2;
}

//------------------------------------------------
DWORD addFreeSpace(CDiskInfo di, CPartInfo *pinfo, DWORD dwInfoCount, QWORD qwDiskSize)
{
   DWORD i;
   CPartInfo *tab2;
   //CPartInfo tab2[MAX_PARTS];
   CPartInfo piHoles[MAX_NEWPARTS];
   DWORD dwHoles;

   // alloc mem
   tab2 = new CPartInfo[dwInfoCount+3];
   memset(tab2, 0, (dwInfoCount+3)*sizeof(CPartInfo));
   memset(piHoles, 0, sizeof(piHoles)); 
   memcpy(&tab2[1], pinfo, dwInfoCount*sizeof(CPartInfo));
   tab2[0].qwGeomEnd = 0LL; // (Sentinel for start)
   tab2[dwInfoCount+1].qwGeomStart = qwDiskSize-1; // (Sentinel for end)

   // search for holes
   dwHoles = 0;
   for (i=0; i <= dwInfoCount; i++)
   {
      //pi_printf ("-> Checking between %lu and %lu [%.10llu -> %.10llu] diff=[%llu]\n", 
      //      i, i+1, tab2[i].qwGeomEnd, tab2[i+1].qwGeomStart, tab2[i+1].qwGeomStart - tab2[i].qwGeomEnd);
      if (tab2[i+1].qwGeomStart - tab2[i].qwGeomEnd > MIN_SECTORS_IN_HOLE)
      {
	 //pi_printf ("   -> Hole between %d and %d and size ", (int)i, (int)i+1);
	 piHoles[dwHoles].qwGeomStart = tab2[i].qwGeomEnd+1;
	 piHoles[dwHoles].qwGeomEnd = tab2[i+1].qwGeomStart-1;
	 piHoles[dwHoles].qwSize = ((QWORD)(piHoles[dwHoles].qwGeomEnd - piHoles[dwHoles].qwGeomStart))*512LL;
	 piHoles[dwHoles].dwP1000Disk = (DWORD)((1000LL * piHoles[dwHoles].qwSize) / (qwDiskSize*512LL));
	 piHoles[dwHoles].dwType = PED_PARTITION_FREESPACE;
	 piHoles[dwHoles].bInsideExtended = isSpaceInsideExtended(di, piHoles[dwHoles].qwGeomStart, piHoles[dwHoles].qwGeomEnd);
	 dwHoles++;
      }
   }
   
   // free and copy mem
   if (dwHoles) // else pinfo was not changed
   {
      memcpy(&pinfo[dwInfoCount], piHoles, dwHoles*sizeof(CPartInfo));
      dwInfoCount += dwHoles;
   }

   delete []tab2;
   return dwInfoCount;
}

//------------------------------------------------
int getDeviceInfos(CLocalHeader *info, char *szDevice, char *szFileSystem, QWORD qwPartSize)
{
   CFSBase *FS = NULL;
   FILE *fDeviceFile;
   int nRes;

   // ----------- open partition file
   errno = 0;
   fDeviceFile = fopen(szDevice, "rb");
   if (fDeviceFile == NULL)
      return (-1);

   memset(info, 0, sizeof(CLocalHeader));

   if ((strcmp(szFileSystem, "ext2") == 0) || (strcmp(szFileSystem, "ext3") == 0)) // if EXT2 / EXT3
      FS = new CExt2Part(szDevice, fDeviceFile, qwPartSize);
   else if (strncmp(szFileSystem, "reiserfs", 8) == 0) // if REISER
      FS = new CReiserPart(szDevice, fDeviceFile, qwPartSize);
   else if (strncmp(szFileSystem, "fat", 3) == 0) // if FAT
      FS = new CFatPart(szDevice, fDeviceFile, qwPartSize);
   else if (strcmp(szFileSystem, "hpfs") == 0) // if HPFS
      FS = new CHpfsPart(szDevice, fDeviceFile, qwPartSize);
   else if (strcmp(szFileSystem, "jfs") == 0) // if JFS
      FS = new CJfsPart(szDevice, fDeviceFile, qwPartSize);
   else if (strcmp(szFileSystem, "xfs") == 0) // if XFS
      FS = new CXfsPart(szDevice, fDeviceFile, qwPartSize);
   else if (strcmp(szFileSystem, "ntfs") == 0) // if NTFS
      FS = new CNtfsPart(szDevice, fDeviceFile, qwPartSize);
   else // if UNKNOWN
      FS = new CUnknownPart(szDevice, fDeviceFile, qwPartSize);

   if (!FS)
      return (-1);

   nRes = -1;
   pi_printf ("run STATS: %s,%s\n", szDevice, szFileSystem);
   try 
   { 
      nRes = FS -> getGeneralInformations(info);
      //pi_printf ("getGeneralInformations(info) -> nRes = %d\n", nRes);
   }
   catch (int *a)
   {
      pi_printf ("CATCH: -1\n");
      nRes = -1;
   }
   //pi_printf ("--> stats for %s,%s=%d\n", szDevice, szFileSystem, nRes);

   fclose(fDeviceFile);
   delete FS;

   return nRes;
}

//------------------------------------------------
void getPartDetails(CNewNet *net)
{
   char szPart[1024];
   char szDevice[2048];
   char szFileSystem[512];
   DWORD dwMinor;
   PedDisk *disk;
   PedPartition *part;
   PedDevice *device;
   CLocalHeader info;
   QWORD qwPartSize;

   // init
   memset(&info, 0, sizeof(info));

   // recv infos
   net->recv_string(szDevice);
   dwMinor = net->recv_u32();

   snprintf(szPart, sizeof(szPart), "%s%lu", szDevice, dwMinor);

   // partition list
   device = NULL;
   ped_device_probe_all();
   
   while ((device = ped_device_get_next(device)) != NULL)
   {
      disk = ped_disk_new(device);
      if (!disk)
      {
	 pi_printf("Error in ped_disk_new\n");
	 net->send(&info, sizeof(info));
	 return ;
      }

      part = NULL;
      while ((part = ped_disk_next_partition(disk, part)) != 0)
      {
	 if ((part->num == (int)dwMinor) && (part->type != PED_PARTITION_EXTENDED))
	 {
	    snprintf(szFileSystem, sizeof(szFileSystem), "%s", part->fs_type->name);
	    qwPartSize = (QWORD)512*(QWORD)part->geom.length;
	    getDeviceInfos(&info, szPart, szFileSystem, qwPartSize);
	    net->send(&info, sizeof(info));
	    return;
	 }
      }
   };

}

